PRAGMA foreign_keys = ON;

ATTACH DATABASE '../Cache/cache.db' AS cache;

CREATE TABLE IF NOT EXISTS cache.post_cache (
  post    INTEGER NOT NULL PRIMARY KEY,
  thread  INTEGER NOT NULL,
  board   TEXT    NOT NULL,
  html    TEXT    NOT NULL
);

CREATE TABLE IF NOT EXISTS cache.thread_cache (
  thread       INTEGER NOT NULL,
  overboard    INTEGER NOT NULL,
  board        TEXT    NOT NULL,
  max_post     INTEGER NOT NULL,
  max_date     INTEGER NOT NULL,
  index_html   TEXT    NOT NULL,
  catalog_html TEXT    NOT NULL,
  PRIMARY KEY (thread, overboard)
);

CREATE INDEX IF NOT EXISTS cache.post_cache_by_thread ON post_cache (thread, post);
CREATE INDEX IF NOT EXISTS cache.post_cache_by_board ON post_cache (board, thread, post);
CREATE INDEX IF NOT EXISTS cache.post_cache_by_board_post ON post_cache (board, post);
CREATE INDEX IF NOT EXISTS cache.thread_cache_by_max_post ON thread_cache (max_post, thread);
CREATE INDEX IF NOT EXISTS cache.thread_cache_by_board ON thread_cache (board, max_post, thread);

CREATE TEMP VIEW file_human_sizes (hash, human_size) AS
  WITH
    t1 AS (SELECT hash, MIN(FLOOR(LOG(1024, size)), 2) AS exponent, * FROM files),
    t2 AS (SELECT *, size / POW(1024, exponent) AS converted FROM t1),
    t3 AS (SELECT *, IIF(exponent > 0, MAX(2 - FLOOR(LOG10(converted)), 0), 0) AS digits FROM t2)
  SELECT hash,
    PRINTF(PRINTF('%%.%df', digits), converted)
      || SUBSTR(' MiB KiB B', 9 - 4 * exponent, 4)
  FROM t3;

CREATE TEMP VIEW file_metadata (hash, html) AS
  SELECT f.hash,
    TRIM(IFNULL(file_human_sizes.human_size, '') || ' ' ||
      TRIM(IFNULL(f.width || 'x' || f.height, '') || 'x' ||
        CASE
        WHEN IFNULL(f.duration, 0) = 0 THEN
          ''
        WHEN f.duration < 60 THEN
          PRINTF('%.2fs', f.duration)
        ELSE
          LTRIM(PRINTF(
            '%02d:%02d:%02d',
            FLOOR(f.duration / 3600),
            FLOOR(f.duration / 60) % 60,
            ROUND(f.duration) % 60
          ), '0:')
        END
      , 'x')
    )
  FROM files AS f
  JOIN file_human_sizes ON f.hash = file_human_sizes.hash;

CREATE TEMP VIEW thumbnails (post, position, html) AS
  SELECT r.post, r.position,
    CASE WHEN f.type = 2 /*image*/ AND f.width IS NOT NULL AND f.height IS NOT NULL THEN
      '<label><input class="expand-toggle" type="checkbox" /><span class="expanded" style="background-image: url(''/File/'
        || SUBSTR(f.hash, 1, 2) || '/' || f.hash || f.extension || IFNULL('/' || r.name_uri, '')
        || '''); width: ' || f.width || 'px; padding-top: ' || (100 * f.height / f.width)
        || '%;"></span>'
    WHEN f.type = 4 /*video*/ AND f.width IS NOT NULL AND f.height IS NOT NULL THEN
      '<label class="expanded-video-container"><input class="expand-toggle" type="checkbox" /><video class="expanded" preload="none" controls=""'
        || IFNULL(' poster="/File/' || SUBSTR(f.hash, 1, 2) || '/' || f.thumb || '"', '')
        || ' width="' || f.width || '" height="' || f.height || '" src="/File/'
        || SUBSTR(f.hash, 1, 2) || '/' || f.hash || f.extension || IFNULL('/' || r.name_uri, '')
        || '"></video><span class="expanded-video-close">x</span>'
    ELSE
      '<a target="_blank" href="/File/' || SUBSTR(f.hash, 1, 2) || '/' || f.hash || f.extension || IFNULL('/' || r.name_uri, '') || '">'
    END
    || IIF(r.spoiler IS NOT NULL,
         '<span class="thumb spoiler-thumb" data-spoiler="' || r.spoiler
           || '" style="background-image: url(''/File/' || SUBSTR(f.hash, 1, 2) || '/' || f.thumb || ''');'
           || IFNULL(' width: '  || f.thumb_width  || 'px;', '')
           || IFNULL(' height: ' || f.thumb_height || 'px;', '')
           || '"></span>',
         '<img class="thumb" src="/File/' || SUBSTR(f.hash, 1, 2) || '/' || f.thumb || '"'
           || IFNULL(' width="'  || f.thumb_width  || '"', '')
           || IFNULL(' height="' || f.thumb_height || '"', '')
           || ' alt="thumb" />'
       )
    || CASE WHEN (f.type = 2 /*image*/ OR f.type == 4 /*video*/) AND f.width IS NOT NULL AND f.height IS NOT NULL THEN '</label>'
       ELSE '</a>'
       END
  FROM file_refs AS r
  JOIN files AS f ON r.hash = f.hash;

CREATE TEMP VIEW file_html (post, position, html) AS
  SELECT r.post, r.position,
    '<div class="file" data-file-ref="' || r.file_ref || '" data-hash="' || f.hash
      || '"><a class="file-link" target="_blank" href="/File/'
      || SUBSTR(f.hash, 1, 2) || '/' || f.hash || f.extension || IFNULL('/' || r.name_uri, '') || '">'
      || IFNULL(r.name, f.hash || f.extension)
      || '</a> <span class="file-metadata">'
      || file_metadata.html
      || IFNULL(' (' || r.spoiler || ')', '')
      || IFNULL(' [' || f.status || ']', '')
      || '</span>'
      || IIF(f.thumb IS NOT NULL AND f.status IS NULL, thumbnails.html, '')
      || IIF(f.type = 3 /*audio*/, '<audio preload="none" controls="" src="/File/'
      || SUBSTR(f.hash, 1, 2) || '/' || f.hash || f.extension || IFNULL('/' || r.name_uri, '') || '"></audio>' , '')
      || '</div>'
  FROM file_refs AS r
  JOIN files AS f ON r.hash = f.hash
  JOIN file_metadata ON r.hash = file_metadata.hash
  JOIN thumbnails ON r.post = thumbnails.post AND r.position = thumbnails.position;

CREATE TEMP VIEW link_attr (post, thread, attr_intra, attr_inter) AS
  SELECT p.post, p.thread,
    ' class="postlink intra" href="/' || t.board || '/' || p.thread || '#P' || p.post || '"',
    ' class="postlink inter" href="/' || t.board || '/' || p.thread || '#P' || p.post || '"'
  FROM posts AS p
  JOIN threads AS t ON p.thread = t.thread;

CREATE TEMP VIEW backlink_html (post, src, html) AS
  SELECT l.dest, l.src,
    '<a' || IIF(p.thread = a.thread, a.attr_intra, a.attr_inter) || '>P' || l.src || '</a>'
  FROM links AS l
  JOIN link_attr AS a ON l.src = a.post
  JOIN posts AS p ON l.dest = p.post;

CREATE TEMP VIEW postdata_html (post, html) AS
  SELECT post,
    '<input class="postselect" type="checkbox" name="C' || post || '" id="C' || post
      || '" /><label for="C' || post || '"></label>'
      || '<a class="in-thread" href="/' || board || '/' || p.thread || '"></a> '
      || IFNULL('<span class="moderator">' || moderator || '</span> ', '')
      || '<span class="postnumber">P' || post
      || '</span> <span class="subject">' || subject || '</span> '
      || STRFTIME('<time datetime="%Y-%m-%dT%H:%M:%SZ">', date, 'unixepoch')
      || SUBSTR('Sun Mon Tue Wed Thu Fri Sat', 4 * STRFTIME('%w', date, 'unixepoch') + 1, 3)
      || STRFTIME(' %Y-%m-%d %H:%M:%S', date, 'unixepoch')
      || '</time> <a href="/' || board || '/' || p.thread || '#P' || post
      || '">link</a> <a href="/' || board || '/' || p.thread || '#postform">reply</a>'
  FROM posts AS p JOIN threads AS t ON p.thread = t.thread;

/*
  Order of GROUP_CONCAT is undefined but this works in practice
  because sqlite does not flatten the subquery per:
  https://sqlite.org/optoverview.html#flattening
  Alternatives tried do not perform as well.
*/
CREATE TEMP VIEW insert_links (post, comment) AS
  SELECT p.post,
    IFNULL((
      SELECT
        GROUP_CONCAT(SUBSTR(comment, prev, position - prev) || attr, '')
          || SUBSTR(comment, MAX(position))
      FROM (
        SELECT src, position,
          IFNULL(LAG(position) OVER (PARTITION BY src ORDER BY position), 1) AS prev,
          IIF(p.thread = a.thread, attr_intra, attr_inter) AS attr
        FROM links
        JOIN link_attr AS a ON dest = a.post
        WHERE src = p.post
        ORDER BY position
      )
    ), p.comment)
  FROM posts AS p;

CREATE TEMP VIEW move_notice (post, html) AS
  SELECT post,
    IIF(original_thread = post,
      IFNULL(
        (SELECT (
          IIF(p.thread <> original_thread OR board <> original_board,
            '<div class="move-notice">Moved from <a href="/'
              || original_board || '/">/' || original_board || '/</a></div>'
          , '')
        ) FROM threads AS ot WHERE ot.thread = original_thread)
      ,
        '<div class="move-notice">Moved from thread ' || original_thread || '</div>'
      )
    ,
      IIF(p.thread <> original_thread,
        IFNULL(
          (SELECT (
            '<div class="move-notice">Moved from ' || '<a href="/'
              || ot.board || '/' || original_thread || '#P' || original_thread || '">/'
              || ot.board || '/' || original_thread || '</a></div>'
          ) FROM threads AS ot WHERE ot.thread = original_thread)
        ,
          '<div class="move-notice">Moved from thread ' || original_thread || '</div>'
        )
      , '')
    )
  FROM posts AS p;

CREATE TEMP VIEW post_html (post, thread, board, html) AS
  SELECT post, thread, board,
    '<div class="post" id="P' || post || '" data-thread="' || thread || '" data-board="' || board
      || '"><div class="postbox" tabindex="0"><div class="postdata">' || d_html || '</div>'
      || IFNULL(f_html, '')
      || '<div class="comment">' || i_comment || '</div>'
      || m_html
      || IFNULL('<div class="backlinks">Referenced by: ' || b_html || '</div>', '')
      || '</div></div>' || CHAR(10)
  FROM (
    SELECT *, t.board, d.html AS d_html, i.comment AS i_comment, m.html AS m_html,
      (SELECT
        '<div class="files" data-count="' || COUNT() || '">' || GROUP_CONCAT(html, '') || '</div>'
        FROM (SELECT f.html FROM file_html AS f WHERE f.post = p.post ORDER BY f.position)
      ) AS f_html,
      (SELECT GROUP_CONCAT(html, ' ') FROM (SELECT DISTINCT b.html FROM backlink_html AS b WHERE b.post = p.post ORDER BY b.src)) AS b_html
    FROM posts AS p
    JOIN threads       AS t ON p.thread = t.thread
    JOIN postdata_html AS d ON p.post = d.post
    JOIN insert_links  AS i ON p.post = i.post
    JOIN move_notice   AS m ON p.post = m.post
  );

CREATE TEMP VIEW thread_html (thread, overboard, board, max_post, max_date, index_html, catalog_html) AS
  SELECT
    thread, overboard, board, max_post, max_date,
    (
      SELECT
        '<fieldset class="thread" id="T' || thread || '" data-board="' || board
          || '"><legend><h2><a class="thread-full-line" href="/' || board || '/' || thread
          || '"></a><span><a href="/' || board || '/' || thread
          || '">Thread ' || thread
          || '</a><span class="in-board"> in <a href="/' || board || '/">/' || board
          || '/</a></span></span></h2></legend>' || CHAR(10)
          || IFNULL((SELECT c.html FROM post_cache AS c WHERE post = op), '')
          || IIF(omitted > 0,
               '<a class="omitted" href="/' || board || '/' || thread|| '">'
                 || omitted || ' repl' || IIF(omitted = 1, 'y', 'ies')
                 || ' omitted.</a>' || CHAR(10)
             , '')
          || IFNULL((SELECT GROUP_CONCAT('<hr class="hidden" />' || html, '') FROM (
               SELECT html FROM (
                 SELECT p.post, c.html
                   FROM posts AS p JOIN post_cache AS c ON p.post = c.post
                   WHERE p.thread = t.thread AND p.post <> op AND NOT (overboard AND p.overboard_exclude) ORDER BY p.post DESC LIMIT 5
               ) ORDER BY post
             )), '')
          || '</fieldset>' || CHAR(10)
    ),
    (
      SELECT
        '<div class="catalog-thread" id="T' || t.thread || '" data-board = "' || t.board
          || '" tabindex="0"><div class="catalog-info"><input class="postselect" type="checkbox" name="C' || thread || '" id="C' || thread
          || '" /><label for="C' || thread
          || '"></label><a href="/' || board || '/' || thread || '">Thread ' || thread
          || '</a><span class="in-board"> in <a href="/' || board || '/catalog">/' || board
          || '/</a></span></div><div class="catalog-info2">P: ' || count
          || STRFTIME(', <time datetime="%Y-%m-%dT%H:%M:%SZ" title="', date, 'unixepoch')
          || SUBSTR('Sun Mon Tue Wed Thu Fri Sat', 4 * STRFTIME('%w', date, 'unixepoch') + 1, 3)
          || STRFTIME(' %Y-%m-%d %H:%M:%S"></time></div>', date, 'unixepoch')
          || IFNULL((SELECT IIF(spoiler IS NULL,
               '<a class="catalog-thumb-link" href="/' || board || '/' || thread || '"><img class="catalog-thumb" src="/File/'
                 || SUBSTR(f.hash, 1, 2) || '/' || IFNULL(f.thumb_catalog, f.thumb) || '" alt="thumb" /></a>'
             , '') FROM file_refs AS r JOIN files AS f ON r.hash = f.hash WHERE r.post = p.post AND r.position = 1), '')
          || '<div class="subject">' || subject
          || '</div><div class="comment">' || i.comment
          || '</div></div>'
      FROM posts AS p JOIN insert_links AS i ON p.post = i.post WHERE p.post = op
    )
  FROM (
    SELECT *,
      count - 1 - MIN((SELECT SUM(NOT (overboard AND p.overboard_exclude)) FROM posts AS p WHERE p.thread = t.thread AND p.post <> op), 5) AS omitted,
      (SELECT IFNULL(MIN(max_bump, MAX(post)), MAX(post)) FROM posts AS p WHERE p.thread = t.thread AND (p.post = op OR NOT (overboard AND p.overboard_exclude))) AS max_post
    FROM (
      SELECT posts.thread, board, MAX(date) AS max_date, COUNT() AS count, max_bump,
        IFNULL((SELECT post FROM posts AS p WHERE p.post = posts.thread), MIN(post)) AS op
      FROM posts
      JOIN threads ON posts.thread = threads.thread
      GROUP BY posts.thread
    ) AS t JOIN (SELECT 0 AS overboard UNION SELECT 1 AS overboard)
  ) AS t;

CREATE TEMP VIEW thread_relative_dates (thread, relative_date) AS
  WITH t1 AS (
    SELECT thread,
      STRFTIME('%Y', 'now') - STRFTIME('%Y', max_date, 'unixepoch') AS dy,
      STRFTIME('%m', 'now') - STRFTIME('%m', max_date, 'unixepoch') AS dm,
      STRFTIME('%d', 'now') - STRFTIME('%d', max_date, 'unixepoch') AS dd,
      STRFTIME('%s', 'now') - max_date AS ds FROM thread_cache WHERE overboard = 0
  ), t2 AS (
    SELECT thread, dy - (dm < 0 OR dm = 0 AND dd < 0) AS dy,
      (12 + dm - (dd < 0)) % 12 AS dm,
      ds
      FROM t1
  ), t3 AS (
    SELECT thread,
      CASE WHEN ds < 0 THEN '0 sec'
      WHEN dy > 0 THEN dy || ' year'
      WHEN dm > 0 THEN dm || ' month'
      WHEN ds >= 86400 THEN CAST(ds / 86400 + 0.5 AS INT) || ' day'
      WHEN ds >= 3600 THEN CAST(ds / 3600 + 0.5 AS INT) || ' hour'
      WHEN ds >= 60 THEN CAST(ds / 60 + 0.5 AS INT) || ' minute'
      ELSE ds || ' second'
      END AS relative_date
      FROM t2
  )
    SELECT thread,
      relative_date || IIF(SUBSTR(relative_date, 1, 2) <> '1 ', 's', '')
      FROM t3;

CREATE TEMP VIEW mod_action_html (action, html) AS
  SELECT action,
    '<legend><h2>' || description || ' by ' || moderator
      || ' at ' || STRFTIME('<time datetime="%Y-%m-%dT%H:%M:%SZ">', date, 'unixepoch')
      || SUBSTR('Sun Mon Tue Wed Thu Fri Sat', 4 * STRFTIME('%w', date, 'unixepoch') + 1, 3)
      || STRFTIME(' %Y-%m-%d %H:%M:%S', date, 'unixepoch')
      || '</time>'
      || IFNULL(' because ' || reason, '') || '</h2></legend>' || CHAR(10)
    AS html
  FROM mod_actions;

/* table_id: 1=threads, 2=posts, 3=files, 4=file_refs */
CREATE TEMP VIEW mod_row_action_html (action, table_id, row_key, post, html) AS
  WITH t1 AS (
    SELECT a.action, table_id, row_key,
      CASE
        WHEN table_id IN (1, 2) THEN row_key
        WHEN table_id = 3 THEN (SELECT post FROM file_refs WHERE hash = row_key ORDER BY file_ref LIMIT 1)
        WHEN table_id = 4 THEN (SELECT post FROM file_refs WHERE file_ref = row_key)
      END
      AS post,
      CASE
        WHEN table_id IN (1, 2) THEN NULL
        WHEN table_id = 3 THEN row_key
        WHEN table_id = 4 THEN (SELECT hash FROM file_refs WHERE file_ref = row_key)
      END
      AS hash,
      CASE
        WHEN NOT is_insert AND NOT is_delete THEN ''
        WHEN     is_insert AND NOT is_delete THEN ' (created)'
        WHEN NOT is_insert AND     is_delete THEN ' (deleted)'
        WHEN     is_insert AND     is_delete THEN ' (created and deleted)'
      END
      AS row_action_type
    FROM mod_actions AS a LEFT JOIN mod_row_actions AS r ON a.action = r.action
  ),
  t2 AS (
    SELECT *,
      (SELECT extension FROM files WHERE files.hash = t1.hash) AS ext
    FROM t1
  ),
  t3 AS (
    SELECT *,
      CASE
        WHEN table_id IN (1, 2) THEN '/P' || row_key
        WHEN table_id IN (3, 4) THEN '/File/' || SUBSTR(hash, 1, 2) || '/' || hash || ext
      END
      AS link,
      CASE
        WHEN table_id = 1 THEN 'Thread ' || row_key
        WHEN table_id = 2 THEN 'Post ' || row_key
        WHEN table_id = 3 THEN 'File ' || row_key || IFNULL(ext, '')
        WHEN table_id = 4 THEN 'File link ' || row_key || IFNULL(' (' || hash || ext || ')', '')
      END
      AS description,
      CASE
        WHEN table_id = 3 THEN 'data-hash'
        WHEN table_id = 4 THEN 'data-file-ref'
      END
      AS attribute
    FROM t2
  )
    SELECT action, table_id, row_key, post,
      IFNULL(
        '<div>'
          || IFNULL('<a href="' || link || '">', '')
          || description
          || IIF(link IS NOT NULL, '</a>', '')
          || row_action_type
          || '</div>'
          || IFNULL(
               '<style>#P' || post || '_A' || action
                 || ' .file[' || attribute || '="' || row_key
                 || '"] {outline: 4px solid red;}</style>'
             , '')
          || CHAR(10)
      , '') AS html
    FROM t3;

CREATE TEMP VIEW cooldown_dates (for_threads, for_captcha, date) AS
  SELECT t, c, (
    SELECT MAX(date + (n - burst) * cooldown * wind / (wind - burst)) FROM (
      SELECT date, ROW_NUMBER() OVER (ORDER BY post DESC) AS n
      FROM posts WHERE t = IIF(post = original_thread, 1, (
        SELECT COUNT() = 0 FROM posts AS p2
        WHERE p2.thread = posts.original_thread AND p2.post < posts.post AND posts.date - p2.date <= 30*86400
      ))
      ORDER BY post DESC LIMIT 256
    ) WHERE n <= wind
  )
  FROM (
    SELECT *, burst * burst_window_ratio AS wind FROM (
      SELECT *,
        IIF(t, g.burst_thread, g.burst_post) * IIF(c, g.captcha_block_ratio, 1) AS burst,
        IIF(t, g.cooldown_thread, g.cooldown_post) AS cooldown
      FROM (SELECT 0 AS t UNION SELECT 1 AS t)
      JOIN (SELECT 0 AS c UNION SELECT 1 AS c)
      JOIN global AS g
    )
  );

CREATE TEMP VIEW cooldown_remaining (for_threads, block_time, captcha_time) AS
  SELECT for_threads, MIN(sec), MAX(sec) FROM (
    SELECT for_threads, IFNULL(MAX(0, date - CAST(STRFTIME('%s', 'now') AS INT)), 0) AS sec FROM cooldown_dates
  ) GROUP BY for_threads;

CREATE TEMP TABLE empty (
  id   INTEGER NOT NULL,
  type INTEGER NOT NULL,
  PRIMARY KEY (id, type),
  CHECK (0)
);

CREATE TEMP TABLE rebuild (
  id   INTEGER NOT NULL,
  type INTEGER NOT NULL,
  PRIMARY KEY (id, type) ON CONFLICT REPLACE,
  FOREIGN KEY (id, type) REFERENCES empty DEFERRABLE INITIALLY DEFERRED
);

CREATE TEMP VIEW start_rebuild (x) AS
  SELECT 0 WHERE 0;

CREATE TEMP TRIGGER insert_threads AFTER INSERT ON threads FOR EACH ROW BEGIN
  INSERT INTO rebuild SELECT post, 1
    FROM posts AS p
    WHERE thread <> original_thread AND original_thread = NEW.thread;
END;

CREATE TEMP TRIGGER update_threads_thread AFTER UPDATE OF thread ON threads FOR EACH ROW BEGIN
  SELECT RAISE(ABORT, 'update of thread number not supported');
END;

CREATE TEMP TRIGGER update_threads_original_board AFTER UPDATE OF original_board ON threads FOR EACH ROW BEGIN
  SELECT RAISE(ABORT, 'update of original board not supported');
END;

CREATE TEMP TRIGGER update_threads_board AFTER UPDATE OF board ON threads FOR EACH ROW BEGIN
  INSERT INTO rebuild SELECT post, 1
    FROM posts
    WHERE thread = NEW.thread;
  INSERT INTO rebuild SELECT src, 1
    FROM links
    JOIN posts AS s ON src = s.post
    JOIN posts AS d ON dest = d.post
    WHERE d.thread = NEW.thread;
  INSERT INTO rebuild SELECT dest, 1
    FROM links
    JOIN posts AS s ON src = s.post
    JOIN posts AS d ON dest = d.post
    WHERE s.thread = NEW.thread;
  INSERT INTO rebuild SELECT post, 1
    FROM posts AS p
    WHERE thread <> original_thread AND original_thread = NEW.thread;
END;

CREATE TEMP TRIGGER update_threads_max_bump AFTER UPDATE OF max_bump ON threads FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (NEW.thread, 2);
END;

CREATE TEMP TRIGGER delete_threads AFTER DELETE ON threads FOR EACH ROW BEGIN
  INSERT INTO rebuild SELECT post, 1
    FROM posts AS p
    WHERE thread <> original_thread AND original_thread = OLD.thread;
END;

CREATE TEMP TRIGGER insert_posts AFTER INSERT ON posts FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (NEW.post, 1);
  INSERT INTO rebuild SELECT src, 1 FROM links WHERE dest = NEW.post;
END;

CREATE TEMP TRIGGER update_posts AFTER UPDATE ON posts FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (NEW.post, 1);
END;

CREATE TEMP TRIGGER update_posts_post AFTER UPDATE OF post ON posts FOR EACH ROW BEGIN
  SELECT RAISE(ABORT, 'update of post number not supported');
END;

CREATE TEMP TRIGGER update_posts_original_thread AFTER UPDATE OF original_thread ON posts FOR EACH ROW BEGIN
  SELECT RAISE(ABORT, 'update of original thread not supported');
END;

CREATE TEMP TRIGGER update_posts_thread AFTER UPDATE OF thread ON posts FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (OLD.thread, 2);
  INSERT INTO rebuild SELECT src, 1 FROM links WHERE dest = NEW.post;
  INSERT INTO rebuild SELECT dest, 1 FROM links WHERE src = NEW.post;
END;

CREATE TEMP TRIGGER delete_posts AFTER DELETE ON posts FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (OLD.post, 1);
  INSERT INTO rebuild VALUES (OLD.thread, 2);
  INSERT INTO rebuild SELECT src, 1 FROM links WHERE dest = OLD.post;
END;

CREATE TEMP TRIGGER insert_links AFTER INSERT ON links FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (NEW.src, 1);
  INSERT INTO rebuild VALUES (NEW.dest, 1);
END;

CREATE TEMP TRIGGER update_links AFTER UPDATE ON links FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (OLD.src, 1);
  INSERT INTO rebuild VALUES (OLD.dest, 1);
  INSERT INTO rebuild VALUES (NEW.src, 1);
  INSERT INTO rebuild VALUES (NEW.dest, 1);
END;

CREATE TEMP TRIGGER delete_links AFTER DELETE ON links FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (OLD.src, 1);
  INSERT INTO rebuild VALUES (OLD.dest, 1);
END;

CREATE TEMP TRIGGER update_files AFTER UPDATE ON files FOR EACH ROW BEGIN
  INSERT INTO rebuild SELECT post, 1 FROM file_refs WHERE hash = NEW.hash;
END;

CREATE TEMP TRIGGER update_files_hash AFTER UPDATE OF hash ON files FOR EACH ROW BEGIN
  SELECT RAISE(ABORT, 'update of file hash not supported');
END;

CREATE TEMP TRIGGER insert_file_refs AFTER INSERT ON file_refs FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (NEW.post, 1);
END;

CREATE TEMP TRIGGER update_file_refs AFTER UPDATE ON file_refs FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (NEW.post, 1);
END;

CREATE TEMP TRIGGER update_file_refs_file_ref AFTER UPDATE OF file_ref ON file_refs FOR EACH ROW BEGIN
  SELECT RAISE(ABORT, 'update of file reference ID not supported');
END;

CREATE TEMP TRIGGER update_file_refs_post AFTER UPDATE OF post ON file_refs FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (OLD.post, 1);
END;

CREATE TEMP TRIGGER delete_file_refs AFTER DELETE ON file_refs FOR EACH ROW BEGIN
  INSERT INTO rebuild VALUES (OLD.post, 1);
END;

CREATE TEMP TRIGGER insert_rebuild AFTER INSERT ON rebuild FOR EACH ROW
  WHEN NEW.type = 1
BEGIN
  INSERT INTO rebuild SELECT thread, 2 FROM posts WHERE post = NEW.id;
END;

CREATE TEMP TRIGGER delete_rebuild_post AFTER DELETE ON rebuild FOR EACH ROW
  WHEN OLD.type = 1
BEGIN
  INSERT INTO post_cache
    SELECT * FROM post_html WHERE post = OLD.id
    ON CONFLICT (post) DO
      UPDATE SET (thread, board, html)
        = (SELECT thread, board, html FROM post_html WHERE post = OLD.id);
  DELETE from post_cache
    WHERE post = OLD.id
    AND (SELECT 1 FROM post_html WHERE post = OLD.id) IS NULL;
END;

CREATE TEMP TRIGGER delete_rebuild_thread AFTER DELETE ON rebuild FOR EACH ROW
  WHEN OLD.type = 2
BEGIN
  INSERT INTO thread_cache
    SELECT * FROM thread_html WHERE thread = OLD.id
    ON CONFLICT (thread, overboard) DO
      UPDATE SET (board, max_post, max_date, index_html, catalog_html)
        = (SELECT board, max_post, max_date, index_html, catalog_html FROM thread_html WHERE thread = excluded.thread AND overboard = excluded.overboard);
  DELETE from thread_cache
    WHERE thread = OLD.id
    AND (SELECT 1 FROM thread_html WHERE thread = OLD.id) IS NULL;
END;

CREATE TEMP TRIGGER insert_start_rebuild INSTEAD OF INSERT ON start_rebuild FOR EACH ROW BEGIN
  DELETE FROM rebuild WHERE type = 1;
  DELETE FROM rebuild;
END;
