local captcha = {}
local openssl = require('openssl')
local common = require('common')
local config = require('config')
local db = require('database')('../Data/private.db')
local db2 = require('database')('../Data/board.db', true):read('cache.sql')

local query_count = db:prepare([[
  SELECT COUNT() FROM captchas
    WHERE date IS NULL OR date >= CAST(STRFTIME('%s', 'now') AS INT) - 30*60
]])
local query_challenge = db:prepare([[
  SELECT challenge, date FROM captchas
    WHERE date IS NULL OR date >= CAST(STRFTIME('%s', 'now') AS INT) - 30*60
    LIMIT 1 OFFSET ?
]])
local query_solution = db:prepare([[
  SELECT COUNT() FROM captchas
    WHERE (date IS NULL OR date >= CAST(STRFTIME('%s', 'now') AS INT) - 60*60)
      AND solution = ?
]])
local insert_captcha = db:prepare([[
  INSERT INTO captchas (challenge, solution) VALUES (?, ?)
]])
local clean_captchas = db:prepare([[
  DELETE FROM captchas
    WHERE NOT (date IS NULL OR date >= CAST(STRFTIME('%s', 'now') AS INT) - 60*60)
]])
local update_captcha_date = db:prepare([[
  UPDATE captchas SET date = CAST(STRFTIME('%s', 'now') AS INT)
    WHERE challenge = ?
]])
local delete_captcha = db:prepare([[
  DELETE FROM captchas WHERE solution = ?
]])

local query_cooldown = db2:prepare([[
  SELECT block_time, captcha_time FROM cooldown_remaining WHERE for_threads = ?
]])
local query_spam = db2:prepare([[
  SELECT IIF(last_spam IS NULL, 0, (
    SELECT MAX(0, date + ? - CAST(STRFTIME('%s', 'now') AS INT)) FROM posts WHERE post = last_spam
  ))
  FROM (
    SELECT MAX(c.post) AS last_spam
    FROM post_cache AS c JOIN posts AS p on c.post = p.post JOIN threads AS t ON c.thread = t.thread
    WHERE c.board = ? AND (original_thread <> c.thread OR c.post = c.thread AND original_board <> c.board)
  )
]])

local function random()
  local x, y = string.byte(openssl.random(2), 1, 2)
  return (0x100 * x + y) / 0x10000
end

local function generate()
  local len = config.MAX_LENGTH.captcha
  local captcha_out = common.exec_read("./captcha.sh " .. len)
  if captcha_out:len() <= len then
    error('captcha generation failed')
  end
  local challenge = openssl.base64(captcha_out:sub(1, -(len + 1)))
  local solution = captcha_out:sub(-len)
  return challenge, solution
end

local function new_captcha(immediate)
  local challenge, solution = generate()
  db:transaction(function()
    insert_captcha(challenge, solution)
    if immediate then
      update_captcha_date(challenge)
    end
    clean_captchas()
    return true
  end)
  if immediate then
    return challenge
  end
end

local function get_one_from_db()
  local _, challenge = db:transaction(function()
    local count = query_count()
    if count >= 0 then
      local i = math.floor(random() * count)
      local challenge, date = query_challenge(i)
      if not date then
        update_captcha_date(challenge)
      end
      return true, challenge
    end
    return true
  end, true)
  return challenge
end

function captcha.get_one()
  local challenge = get_one_from_db()
  if not challenge then
    challenge = new_captcha(true)
  end
  return challenge
end

function captcha.verify(solution)
  solution = solution:upper()
  local _, correct = db:transaction(function()
    local correct = (query_solution(solution) > 0)
    if correct then
      delete_captcha(solution)
    end
    return true, correct
  end, true)
  return correct
end

function captcha.fill()
  if query_count() < 30 then
    new_captcha()
  end
end

function captcha.cooldown(is_thread)
  local block_time, captcha_time = query_cooldown(is_thread)
  local captcha_time2 = query_spam(config.SPAM_CAPTCHA_TIME, config.SPAM_BOARD)
  local reason = 'flood'
  if captcha_time2 > captcha_time then
    captcha_time = captcha_time2
    reason = 'recent spam'
  end
  return block_time, captcha_time, reason
end

return captcha
