local sqlite3 = require('lsqlite3')
local common = require('common')
local Database = {}

local function database(filename, readonly)
  local db, _, err  = sqlite3.open(filename, readonly and sqlite3.OPEN_READONLY or nil)
  if not db then
    error(err)
  end
  db:busy_timeout(10000)
  return setmetatable({raw = db}, Database)
end

Database.__index = Database

function Database:check(code)
  if not (code == sqlite3.OK or code == sqlite3.ROW or code == sqlite3.DONE) then
    error(self.raw:errmsg(), 2)
  else
    return code
  end
end

function Database:bind(stmt, ...)
  if ... and type(...) == 'table' then
    return self:check(stmt:bind_names(...))
  else
    return self:check(stmt:bind_values(...))
  end
end

function Database:prepare(query)
  local stmt = self.raw:prepare(query)
  self:check(self.raw:errcode())

  return function(...)
    self:bind(stmt, ...)
    local code = stmt:step()
    if not (code == sqlite3.ROW or code == sqlite3.DONE) then
      stmt:reset()
      error(self.raw:errmsg())
    end
    local results = (code == sqlite3.ROW) and stmt:get_values() or {}
    local n = stmt:columns()
    self:check(stmt:reset())
    return unpack(results, 1, n)
  end
end

function Database:prepare_mult(query)
  local stmt = self.raw:prepare(query)
  self:check(self.raw:errcode())

  return function(results, ...)
    self:bind(stmt, ...)
    local code = stmt:step()
    while code == sqlite3.ROW do
      local val = stmt:get_value(0)
      if val ~= nil then
        table.insert(results, val)
      end
      code = stmt:step()
    end
    if code ~= sqlite3.DONE then
      stmt:reset()
      error(self.raw:errmsg())
    end
    self:check(stmt:reset())
    return results
  end
end

function Database:exec(query)
  self:check(self.raw:exec(query))
  return self
end

function Database:transaction(cb, deferred)
  if deferred then
    self:exec('BEGIN DEFERRED')
  else
    self:exec('BEGIN IMMEDIATE')
  end
  local results = common.pack(xpcall(cb, debug.traceback))
  if results[1] then
    if results[2] then
      if self.raw:exec('COMMIT') ~= sqlite3.OK then
        self.raw:exec('ROLLBACK')
        error(self.raw:errmsg())
      end
    else
      self:exec('ROLLBACK')
    end
    return unpack(results, 2, results.n)
  else
    self.raw:exec('ROLLBACK')
    error(results[2])
  end
end

function Database:read(file)
  return self:exec(common.read_all(file))
end

return database
