local posix = require('posix')
local cjson_safe = require('cjson.safe')
local common = require('common')
local config = require('config')
local gen = {}

local MEDIA_TYPE = {TEXT = 1, IMAGE = 2, AUDIO = 3, VIDEO = 4, APPLICATION = 5}

local escape_table = {
  ['&'] = '&amp;',
  ['<'] = '&lt;',
  ['>'] = '&gt;',
  ['"'] = '&quot;',
  ["'"] = '&#39;'
}

function gen.escape(s)
  return (string.gsub(s, '[&<>"\']', escape_table))
end

local tex_commands = {}
local tex_fonts = {}

local function add_command(char, command)
  char = gen.escape(char)
  command = gen.escape(command)
  local font, arg = command:match('^(\\%a+){(.*)}$')
  if font then
    tex_fonts[font] = tex_fonts[font] or {}
    tex_fonts[font][arg] = tex_fonts[font][arg] or char
  else
    tex_commands[command] = tex_commands[command] or char
  end
end

for line in io.lines('latex.txt') do
  if not line:find('^#') then
    local char, command = line:match('^(.*) (.*)$')
    add_command(char, command)
  end
end
for line in io.lines('../Lib/unimathsymbols.txt') do
  if not line:find('^#') then
    local char, command = line:match('^.-%^(.-)%^(.-)%^')
    if command:find('^\\') then
      add_command(char, command)
    end
    local comment = line:match('[^^]*$')
    for alt_command, source in comment:gmatch('= (\\[^ ,]+)([^,]*),') do
      if source == '' or source:find('ams') then
        add_command(char, alt_command)
      end
    end
  end
end

function gen.encodeURIComponent(s)
  -- single quote ' is safe for URIs but escaped here to be safe for HTML, CSS
  s = string.gsub(s, "[^A-Za-z0-9_.!~*()-]", function(c)
    return string.format('%%%02X', string.byte(c))
  end)
  return s
end

function gen.excerpt(s)
  return s
    :gsub('<br />', ' ')
    :gsub('<[^<>]*>?', '')
    :gsub('%[spoiler: ?[^%[%]<>]*%]?', '[spoiler]')
    :gsub('^%s+', '')
    :gsub('%s+$', '')
end

local function mktemp(template)
  local fd, name, err = posix.stdlib.mkstemp(template)
  if err then error('mkstemp(' .. template .. ') failed with error ' .. err) end
  assert(posix.unistd.close(fd))
  assert(posix.chmod(name, 'rw-r--r--'))
  return name
end

local extensions = {}
local mimetypes = {}
do
  for line in io.lines('types.txt') do
    local ext, typ = string.match(line, '"(.-)".-"([^;"]*).*"')
    extensions[typ] = ext
    mimetypes[ext] = mimetypes[ext] or typ
  end
end

function gen.file_path(name)
  return '../File/' .. name:sub(1, 2) .. '/' .. name
end

local function shred(name)
  local temp = mktemp('../File/shred.tmpXXXXXX')
  assert(os.rename(name, temp))
  if os.execute("shred -u " .. temp) ~= 0 then
    error('shredding failed')
  end
end

function gen.shred_files(hash)
  if not (hash:len() == 64 and hash:find('^[0-9a-f]+$')) then
    error('invalid hash: ' .. hash)
  end
  local files = posix.glob(gen.file_path(hash) .. '*') or {}
  for _, file in ipairs(files) do
    shred(file)
  end
end

function gen.write_file(hash, body)
  if os.execute("mkdir -p ../File/" .. hash:sub(1, 2)) ~= 0 then
    error('failed to create directory')
  end
  local path = mktemp(gen.file_path(hash .. '.tmpXXXXXX'))
  local fd = assert(io.open(path, 'wb'))
  assert(fd:write(body))
  assert(fd:flush())
  assert(fd:close())
  return path
end

local EXEC_LIMITS = 'ulimit -t 10; '
local function exec_read_timeout(command, start_deadline)
  if os.time() > start_deadline then
    return ''
  end
  return common.exec_read(EXEC_LIMITS .. command)
end
local function exec_timeout(command, start_deadline)
  if os.time() > start_deadline then
    return 1
  end
  return os.execute(EXEC_LIMITS .. command)
end

local function dimensions_ok(metadata)
  return (
    metadata.width and metadata.height and
    metadata.width >= 0 and metadata.height >= 0 and
    metadata.width <= config.MAX_THUMBABLE_WIDTH and
    metadata.height <= config.MAX_THUMBABLE_HEIGHT
  )
end

local function thumb_dimensions(width, height)
  local width_scale  = config.THUMB_WIDTH  / width
  local height_scale = config.THUMB_HEIGHT / height
  if width_scale >= 1 and height_scale >= 1 then
    return width, height
  elseif width_scale < height_scale then
    return config.THUMB_WIDTH, math.floor(width_scale * height + 0.5)
  else
    return math.floor(height_scale * width + 0.5), config.THUMB_HEIGHT
  end
end

local get_metadata = {}

function get_metadata.image(file, start_time)
  local metadata = {}

  if file.extension == '.jxl' then
    local jxlinfo = exec_read_timeout(
      "jxlinfo " .. file.temp.file,
      start_time + config.DEADLINE_METADATA
    )
    jxlinfo = '\n' .. jxlinfo .. '\n'
    local img_or_anim, width, height = jxlinfo:match('\nJPEG XL (%l+), (%d+)x(%d+),')
    local not_animated = (img_or_anim == 'image')
    if not img_or_anim then
      -- old format
      width, height = jxlinfo:match('\ndimensions: (%d+)x(%d+)\n')
      not_animated = jxlinfo:find('\nhave_animation: 0\n') and true or false
    end
    if width then
      if jxlinfo:find('\n[Oo]rientation: [4-7]') then
        width, height = height, width
      end
      metadata.width = tonumber(width)
      metadata.height = tonumber(height)
    end
    if dimensions_ok(metadata) and not_animated then
      local temp = gen.file_path(file.hash .. '.tmp.png')
      if exec_timeout(
        "djxl --quiet -- " .. file.temp.file .. " " .. temp,
        start_time + config.DEADLINE_THUMB
      ) == 0 then
        metadata.temp = temp
        metadata.magick_input = 'png:' .. temp
        metadata.magick_pipe = ''
      else
        shred(temp)
      end
    end
    return metadata
  end

  metadata.magick_input = string.gsub(file.extension, '^%.', '') .. ':' .. file.temp.file
  metadata.magick_pipe = ''

  -- width and height
  local width_height = exec_read_timeout(
    "identify -ping -format '%w %h' " .. metadata.magick_input .. "[0]",
    start_time + config.DEADLINE_METADATA
  )
  local width, height = string.match(width_height, '(%d+) (%d+)')
  if width then
    metadata.width = tonumber(width)
    metadata.height = tonumber(height)
  end

  -- duration
  if file.extension == '.gif' then
    local delays = exec_read_timeout(
      "identify -ping -format '%T ' " .. metadata.magick_input,
      start_time + config.DEADLINE_METADATA
    )
    local duration = 0
    local frames = 0
    for token in string.gmatch(delays, '%d+') do
      frames = frames + 1
      local delay = tonumber(token)
      -- Animated GIFs with delay = 0cs are unfortunately common.
      -- Most browsers display these as if the delay was 10cs.
      if delay < 2 then
        delay = 10
      end
      duration = duration + delay
    end
    if frames > 1 then
      metadata.duration = duration / 100
    elseif frames == 1 then
      metadata.duration = 0
    end
  end

  return metadata
end

function get_metadata.video(file, start_time)
  local metadata = {}
  local format = mimetypes[file.extension]:match('/(.*)')

  local ffprobe_json = exec_read_timeout(
    "ffprobe -v 24 -of json -show_format -show_streams -f " .. format .. " -- " .. file.temp.file,
    start_time + config.DEADLINE_METADATA
  )
  local ffprobe_out = cjson_safe.decode(ffprobe_json)

  if type(ffprobe_out) == 'table' then
    if type(ffprobe_out.streams) == 'table' then
      for _, stream in ipairs(ffprobe_out.streams) do
        if type(stream) == 'table' and stream.codec_type == 'video' then
          metadata.codec = tostring(stream.codec_name)
	  metadata.width = tonumber(stream.width)
	  metadata.height = tonumber(stream.height)
          break
        end
      end
    end
    if type(ffprobe_out.format) == 'table' then
      metadata.duration = tonumber(ffprobe_out.format.duration)
    end
  end

  local decoder = config.VIDEO_CODEC_WHITELIST[metadata.codec]
  if decoder then
    metadata.magick_input = 'pnm:-'
    metadata.magick_pipe =
      "ffmpeg -v warning -nostdin -strict normal -f "
        .. format .. " -c:v " .. decoder .. " -i " .. file.temp.file
        .. " -map 0:v:0 -vf thumbnail -vframes 1 -f rawvideo -c:v pam - | "
  end

  return metadata
end

function get_metadata.audio(file, start_time)
  local metadata = {}
  local format = string.gsub(file.extension, '^%.', '')

  local ffprobe_json = exec_read_timeout(
    "ffprobe -v 24 -of json -show_format -f " .. format .. " -- " .. file.temp.file,
    start_time + config.DEADLINE_METADATA
  )
  local ffprobe_out = cjson_safe.decode(ffprobe_json)

  if type(ffprobe_out) == 'table' and type(ffprobe_out.format) == 'table' then
    metadata.duration = tonumber(ffprobe_out.format.duration)
  end

  return metadata
end

function get_metadata.application(file, start_time)
  local metadata = {}

  if file.extension == '.pdf' then
    local pdfinfo_out = exec_read_timeout(
      "pdfinfo -- " .. file.temp.file,
      start_time + config.DEADLINE_METADATA
    )
    local width, height = ('\n' .. pdfinfo_out):match('\nPage size:%s*([%d%.]+) x ([%d%.]+)')
    if width then
      width = tonumber(width)
      height = tonumber(height)
      if width and height then
        metadata.width = math.floor(width + 0.5)
        metadata.height = math.floor(height + 0.5)
      end
    end

    if dimensions_ok(metadata) then
      local thumb_width, thumb_height = thumb_dimensions(width, height)
      if thumb_width ~= 0 and thumb_height ~= 0 then
        local temp = gen.file_path(file.hash .. '.tmp.ppm')
        if exec_timeout(
          "pdftoppm -singlefile -scale-to-x " .. thumb_width .. " -scale-to-y " .. thumb_height
            .. " -- " .. file.temp.file .. " " .. temp:gsub('%.ppm$', ''),
          start_time + config.DEADLINE_THUMB
        ) == 0 then
          metadata.temp = temp
          metadata.magick_input = 'ppm:' .. temp
          metadata.magick_pipe = ''
        else
          shred(temp)
        end
      end
    end
  end

  return metadata
end

function gen.handle_file(hash, path, size, is_reply, start_time)
  local file = {
    hash = hash,
    size = size,
    temp = {file = path}
  }

  -- determine type
  local type_line = exec_read_timeout(
    "file -bi -P bytes=" .. file.size .. " " .. path,
    start_time + config.DEADLINE_METADATA
  )
  local supertype = string.match(type_line, '[^/;\n]*')
  local subtype = string.match(type_line, '[^;\n]*')
  local charset = string.match(type_line, '; charset=([^\n]*)')

  if extensions[subtype] then
    file.extension = extensions[subtype]
    file.type = MEDIA_TYPE[string.upper(supertype)]
  elseif supertype == 'text' or (charset and charset ~= 'binary') then
    file.extension = '.txt'
    file.type = MEDIA_TYPE.TEXT
  elseif type_line:find('^%l+/') then
    file.extension = ''
  end
  file.file = file.hash .. (file.extension or '')

  -- metadata
  if file.extension == '' or not get_metadata[supertype] then
    return file
  end
  local metadata = get_metadata[supertype](file, start_time)
  for _, key in ipairs({'width', 'height', 'duration'}) do
    file[key] = metadata[key]
  end

  if not metadata.magick_input or not metadata.magick_pipe or not dimensions_ok(file) then
    if metadata.temp then shred(metadata.temp) end
    return file
  end

  -- thumbnail dimensions
  file.thumb_width, file.thumb_height = thumb_dimensions(file.width, file.height)

  if file.thumb_width == 0 or file.thumb_height == 0 then
    file.thumb_width  = nil
    file.thumb_height = nil
    if metadata.temp then shred(metadata.temp) end
    return file
  end

  -- thumbnail
  local temp = mktemp(gen.file_path(file.hash .. '.t.tmpXXXXXX'))
  if exec_timeout(
    metadata.magick_pipe .. "convert -define jpeg:size="
      .. (2*file.thumb_width) .. "x" .. (2*file.thumb_height)
      .. " " .. metadata.magick_input
      .. "[0] -thumbnail '" .. file.thumb_width .. "x" .. file.thumb_height
      .. "!' webp:" .. temp,
    start_time + config.DEADLINE_THUMB
  ) == 0 then
    file.thumb = file.hash .. '.t.webp'
    file.temp.thumb = temp
  else
    shred(temp)
    file.thumb_width  = nil
    file.thumb_height = nil
    if metadata.temp then shred(metadata.temp) end
    return file
  end

  if not is_reply then
    temp = mktemp(gen.file_path(file.hash .. '.t2.tmpXXXXXX'))
    if exec_timeout(
      metadata.magick_pipe .. "convert -define jpeg:size="
        .. (2*file.thumb_width) .. "x" .. (2*file.thumb_height)
        .. " " .. metadata.magick_input
        .. "[0] -thumbnail '" .. config.THUMB_WIDTH_CATALOG .. "x" .. config.THUMB_HEIGHT_CATALOG
        .. "' webp:" .. temp,
      start_time + config.DEADLINE_THUMB
    ) == 0 then
      file.thumb_catalog = file.hash .. '.t2.webp'
      file.temp.thumb_catalog = temp
    else
      shred(temp)
    end
  end

  if metadata.temp then shred(metadata.temp) end
  return file
end

function gen.commit_file(file)
  for key, src in pairs(file.temp) do
    local dest = gen.file_path(file[key])
    local state = posix.stat(dest)
    if state then
      assert(type(state) == 'table')
      assert(posix.chmod(src, state.mode) == 0)
    end
    assert(os.rename(src, dest))
  end
end

function gen.cancel_file(file)
  for _, val in pairs(file.temp) do
    shred(val)
  end
end

function gen.redact_file(hash, redacted)
  redacted = (redacted == nil) and true or redacted
  local mode = redacted and '-w-------' or 'rw-r--r--'
  local files = posix.glob(gen.file_path(hash) .. '*') or {}
  assert(type(files) == 'table')
  for _, file in ipairs(files) do
    assert(posix.chmod(file, mode) == 0)
  end
end

local parser_methods = {
  matches = function(self, regexp)
    self.result = select(2, self.text:find('^' .. regexp, self.pos))
    return self.result
  end,
  eat = function(self, pos2)
    pos2 = pos2 or self.result
    local s = self.text:sub(self.pos, pos2)
    self.pos = pos2 + 1
    return s
  end
}

local function make_parser(text)
  return setmetatable({text = text, pos = 1, len = text:len()}, {__index = parser_methods})
end

local function hide(s)
  return '<span class="markup">' .. s .. '</span>'
end

local function replace(s, replacement)
  return '<span data-r="' .. replacement .. '"></span>' .. hide(s)
end

local tags = {
  spoiler = {},
  bold = {element = 'b'},
  tex = {},
}

local function handle_tags_rec(text, depth)
  depth = depth or 0
  if depth >= config.MAX_MARKUP_DEPTH then
    return text
  end
  local p = make_parser(text)
  local text2 = ''
  while p.pos <= p.len do
    local i, j, tag_name = text:find('%[(%l+): ?', p.pos)
    local tag = tags[tag_name]
    if tag then
      local element = tag.element or 'span'
      local parse_body = tag.parse_body or handle_tags_rec
      local _, k = text:find('^%b[]', i)
      text2 = text2 .. p:eat(i - 1)
      text2 = text2 .. hide(p:eat(j)) .. '<' .. element .. ' class="' .. tag_name .. '">'
      if k then
        text2 = text2 .. parse_body(p:eat(k - 1), depth + 1)
        text2 = text2 .. '</' .. element .. '>' .. hide(p:eat(k))
      else
        text2 = text2 .. parse_body(p:eat(p.len), depth + 1) .. '</' .. element .. '>'
      end
    else
      text2 = text2 .. p:eat(j or p.len)
    end
  end
  return text2
end

local function handle_tags(text)
  local text2 = text:gsub('[^<>]+', handle_tags_rec)
  local _, newlines = string.gsub(text, '\n', '')
  local _, fractions = string.gsub(text2, '<span class="fraction">', '')
  local _, columns = string.gsub(text2, '<span class="column">', '')
  if newlines + fractions + columns > config.MAX_NEWLINES.comment then
    return text
  else
    return text2
  end
end

local function tex_next(p, depth, font, hide_braces)
  if p:matches('\\%a+') or p:matches('\\&.-;') or p:matches('\\[^&]') then
    local command = p:eat()
    if command:find('^\\d?frac$') and depth < config.MAX_MARKUP_DEPTH then
      local arg1 = tex_next(p, depth + 1, font, true)
      local arg2 = tex_next(p, depth + 1, font, true)
      return (
        hide(command)
        .. '<span class="fraction"><span class="num">' .. arg1
        .. '</span><span class="den">' .. arg2 .. '</span></span>'
      )
    elseif command == '\\binom' and depth < config.MAX_MARKUP_DEPTH then
      local arg1 = tex_next(p, depth + 1, font, true)
      local arg2 = tex_next(p, depth + 1, font, true)
      return (
        hide(command)
        .. '<span class="binom"><span class="column"><span>' .. arg1
        .. '</span><span>' .. arg2 .. '</span></span></span>'
      )
    elseif tex_fonts[command] and depth < config.MAX_MARKUP_DEPTH then
      local arg = tex_next(p, depth + 1, tex_fonts[command], true)
      return hide(command) .. arg
    elseif font and font[command] then
      return replace(command, font[command])
    elseif tex_commands[command] then
      return replace(command, tex_commands[command])
    else
      return command
    end
  elseif p:matches('{') then
    local text2 = p:eat()
    local level = 1
    while p.pos <= p.len do
      if p:matches('{') then
        level = level + 1
        text2 = text2 .. p:eat()
      elseif p:matches('}') then
        level = level - 1
        text2 = text2 .. p:eat()
        if level == 0 then
          if hide_braces then
            text2 = text2:gsub('^{', hide):gsub('}$', hide)
          end
          return text2
        end
      else
        text2 = text2 .. tex_next(p, depth, font)
      end
    end
    return text2
  elseif p:matches('[_%^]') and depth < config.MAX_MARKUP_DEPTH then
    local text2 = ''
    local n = {sub = 0, sup = 0}
    while (n.sub == 0 and p:matches('_')) or (n.sup == 0 and p:matches('%^')) do
      local op = p:eat()
      local class = (op == '_') and 'sub' or 'sup'
      local arg = tex_next(p, depth + 1, font, true)
      text2 = text2 .. '<span class="' .. class .. '">' .. hide(op) .. arg .. '</span>'
      n[class] = n[class] + 1
    end
    if n.sub == 1 and n.sup == 1 then
      return '<span class="supsub">' .. text2 .. '</span>'
    else
      return text2
    end
  elseif p:matches('&.-;') or p:matches('.') then
    local char = p:eat()
    if font and font[char] then
      return replace(char, font[char])
    else
      return char
    end
  else
    return ''
  end
end

tags.tex.parse_body = function(text, depth)
  local p = make_parser(text)
  local text2 = ''
  while p.pos <= p.len do
    text2 = text2 .. tex_next(p, depth)
  end
  return text2
end

local function handle_brackets(s)
  if s:find('<a>') then
    return s:sub(1, -2) .. '</a>' .. s:sub(-1)
  end
end

local function handle_links(text)
  return (
    text
      :gsub('https?://[^%c "\'<>\\%^`{|}\127]*', '<a>%0</a>')
      :gsub('(%W)(P%d+)', '%1<a>%2</a>')
      :gsub('^(P%d+)', '<a>%1</a>')
      :gsub('<a>(P%d+)</a>(%w)', '%1%2')
      :gsub('(<a>http[^<>]*)<a>(P%d+)</a>', '%1%2')
      :gsub('(&[gl]t;[^<>]*)</a>', '</a>%1')
      :gsub('(&quot;[^<>]*)</a>', '</a>%1')
      :gsub('([!,%.%?])</a>', '</a>%1')
      :gsub('&#39;</a>', '</a>&#39;')
      :gsub('(%b())</a>', handle_brackets)
      :gsub('(%b[])</a>', handle_brackets)
      :gsub('<a>(http[^<>"]*)</a>', '<a href="%1" rel="noreferrer noopener">%1</a>')
      :gsub('\n', '<br />')
  )
end

function gen.markup(text)
  text = (
    gen.escape(text)
      :gsub('^(&gt;[^\n]*)', '<span class="quote">%1</span>')
      :gsub('\n(&gt;[^\n]*)', '\n<span class="quote">%1</span>')
      :gsub('</span>\n<span class="quote">', '\n')
  )
  text = handle_tags(text)
  for _, item in ipairs({'MESSAGE', 'PUBLIC KEY BLOCK', 'SIGNATURE'}) do
    text = text:gsub(
      '(%-? ?%-%-%-%-%-BEGIN PGP ' .. item .. '%-%-%-%-%-)(\n[^<>]-\n%-? ?%-%-%-%-%-END PGP ' .. item .. '%-%-%-%-%-)',
      '<label class="pgp"><input type="checkbox">%1<span>%2</span></label>'
    )
  end
  text = handle_links(text)
  text = text:gsub('\n', '<br />')
  return text
end

function gen.linkify(text)
  return handle_links(gen.escape(text)):gsub('<a>(P%d+)</a>', '<a class="postlink" href="/%1">%1</a>')
end

local unescape_table = {}
for k, v in pairs(escape_table) do
  unescape_table[v] = k
end

function gen.unmarkup(text)
  return text
    :gsub('<br />', '\n')
    :gsub('<[^<>]->', '')
    :gsub('&[%l#%d]-;', unescape_table)
end

function gen.find_links(text)
  local links = {}
  local i = 0
  repeat
    local j, n
    j, i, n = text:find('<a>P(%d+)</a>', i + 1)
    if i then
      table.insert(links, {position = common.utf8len(text:sub(1, j - 1)) + 3, dest = tonumber(n)})
    end
  until not i
  return links
end

return gen
