local sqlite3 = require('lsqlite3')
local common = require('common')

local LOGGED_TABLES = {'threads', 'posts', 'files', 'file_refs'}
local OPERATIONS = {'INSERT', 'UPDATE', 'DELETE'}

local PRELUDE = [[
CREATE TEMP TABLE current_action (
  id     INTEGER NOT NULL DEFAULT 1    PRIMARY KEY CHECK (id IS 1),
  action INTEGER          DEFAULT NULL
);

INSERT INTO current_action DEFAULT VALUES;

CREATE TEMP TRIGGER set_current_action AFTER INSERT ON mod_actions FOR EACH ROW BEGIN
  UPDATE current_action SET action = NEW.action;
END;
]]

local LOG_ROW_ACTION = [[

CREATE TEMP TRIGGER log_row_action_{suffix} AFTER {operation} ON {sqltable} FOR EACH ROW BEGIN
  INSERT INTO mod_row_actions
    (action, table_id, row_key, is_insert, is_delete)
    VALUES (
      (SELECT action FROM current_action),
      {isqltable}, {key}, {insert}, {delete}
    )
    ON CONFLICT DO UPDATE SET is_delete = {delete};
END;
]]

local LOG_CHANGE = [[

CREATE TEMP TRIGGER log_change_{suffix} AFTER {operation}{updateof} ON {sqltable} FOR EACH ROW BEGIN
  INSERT INTO mod_changes
    (action, table_id, row_key, column_id, old_value, new_value)
    VALUES (
      (SELECT action FROM current_action),
      {isqltable}, {key}, '{column}', {oldvalue}, {newvalue}
    )
    ON CONFLICT DO UPDATE SET new_value = {newvalue};
END;
]]

local function log_triggers()
  local db, _, err = sqlite3.open_memory()
  if not db then
    error(err)
  end
  local schema = common.read_all('board.sql')
  if db:exec(schema) ~= sqlite3.OK then
    error(db:errmsg())
  end

  local table_info = {}
  for i, sqltable in ipairs(LOGGED_TABLES) do
    local t = {columns = {}}
    for _, column, _, _, _, pk in db:urows('PRAGMA TABLE_INFO(' .. sqltable .. ')') do
      if pk == 1 then
        t.key = column
      end
      table.insert(t.columns, column)
    end
    table_info[i] = t
  end

  local sql = {PRELUDE}

  for i, sqltable in ipairs(LOGGED_TABLES) do
    local t = table_info[i]
    for _, operation in ipairs(OPERATIONS) do

      local s = {
        suffix = sqltable .. '_' .. operation:lower(),
        operation = operation,
        sqltable = sqltable,
        isqltable = i,
        key = ((operation ~= 'DELETE') and 'NEW.' or 'OLD.') .. t.key,
        insert = (operation == 'INSERT') and 1 or 0,
        delete = (operation == 'DELETE') and 1 or 0,
      }
      local trigger = LOG_ROW_ACTION:gsub('{(%w+)}', s)
      table.insert(sql, trigger)

      for _, column in ipairs(t.columns) do
        s.suffix = sqltable .. '_' .. operation:lower() .. '_' .. column
        s.updateof = (operation == 'UPDATE') and (' OF ' .. column) or ''
        s.column = column
        s.oldvalue = (operation ~= 'INSERT') and ('OLD.' .. column) or 'NULL'
        s.newvalue = (operation ~= 'DELETE') and ('NEW.' .. column) or 'NULL'
        trigger = LOG_CHANGE:gsub('{(%w+)}', s)
        table.insert(sql, trigger)
      end

    end
  end

  if db:close() ~= sqlite3.OK then
    error(db:errmsg())
  end

  return table.concat(sql)
end

return log_triggers
