local cjson = require('cjson')
local openssl = require('openssl')
local config = require('config')
local common = require('common')
local log_triggers = require('log_triggers')
local db = require('database')('../Data/board.db')
  :read('cache.sql')
  :exec(log_triggers())
  :exec("ATTACH DATABASE '../Data/private.db' AS private;")
local db_readonly = require('database')('../Data/board.db', true):read('cache.sql')
local gen = require('gen')
local login = require('login')
local mod = {}

local insert_action = db:prepare([[
  INSERT INTO mod_actions
    (action, date, moderator, function, arguments, description, reason, status)
    VALUES (
      (SELECT last_action + 1 FROM global), STRFTIME('%s', 'now'),
      :mod, :func, :args, 'pending', :reason, :status
    )
]])

local set_description = db:prepare([[
  UPDATE mod_actions SET description = ?
    WHERE action = (SELECT action FROM current_action)
]])

local copy_changes_to_private = db:prepare([[
  INSERT INTO redacted_changes
    SELECT * FROM mod_changes WHERE action = (SELECT action FROM current_action)
]])
local redact_changes = db:prepare([[
  DELETE FROM mod_changes WHERE action = (SELECT action FROM current_action)
]])

local action_complete = db:prepare('UPDATE current_action SET action = NULL')
local trigger_rebuild = db:prepare('INSERT INTO start_rebuild VALUES (1)')

local move_post_to_thread   = db:prepare('UPDATE posts SET thread = ? WHERE post = ?')
local move_thread_to_thread = db:prepare('UPDATE posts SET thread = ? WHERE thread = ?')
local move_thread_to_board = db:prepare('UPDATE threads SET board = ? WHERE thread = ?')

local create_thread = db:prepare([[
  INSERT INTO threads (thread, board, original_board) VALUES (?1, ?2, ?2)
]])

local delete_file = db:prepare("UPDATE files SET status = 'Deleted', thumb = NULL, thumb_catalog = NULL WHERE hash = ?")
local ban_file = db:prepare("UPDATE files SET status = 'Banned', thumb = NULL, thumb_catalog = NULL WHERE hash = ?")
local unban_file = db:prepare("UPDATE files SET status = 'Deleted' WHERE hash = ? AND status = 'Banned'")
local redact_file = db:prepare("UPDATE files SET status = ? WHERE hash = ? AND (status IS NULL OR status = 'Redacted')")

local set_spoiler = db:prepare("UPDATE file_refs SET spoiler = ? WHERE file_ref = ?")

local delete_links = db:prepare("DELETE FROM links WHERE src = ?")
local insert_link = db:prepare("INSERT INTO links (src, position, dest) VALUES (:src, :position, :dest)")

local insert_board = db:prepare("INSERT OR IGNORE INTO boards (board) VALUES (?)")

local set_board_prop = {}
do
  local keys = {'overboard_exclude', 'mod_ids', 'new_threads_disabled', 'text_only'}
  for _, key in ipairs(keys) do
    set_board_prop[key] = db:prepare("UPDATE boards SET " .. key .. " = ? WHERE board = ?")
  end
end

local set_overboard_exclude_thread = db:prepare("UPDATE threads SET overboard_exclude = ? WHERE thread = ?")
local set_overboard_exclude_post = db:prepare("UPDATE posts SET overboard_exclude = ? WHERE post = ?")

local insert_login = db:prepare("INSERT INTO logins (moderator) VALUES (?)")
local set_password = db:prepare("UPDATE logins SET password = ? WHERE moderator = ?")
local set_mod_status = db:prepare("UPDATE logins SET status = ? WHERE moderator = ?")

local query_thread = db:prepare('SELECT thread FROM posts WHERE post = ?')
local query_board  = db:prepare('SELECT board FROM threads WHERE thread = ?')

-- When multiple posts are moved by move_to_board, try to collect them by original thread.
local query_move_to_board_collect = db:prepare([[
  SELECT MIN(post) FROM posts AS p
    WHERE post = thread AND thread <> original_thread AND original_thread = :thread
    AND (SELECT board FROM threads AS t WHERE p.thread = t.thread) = :dest
    AND (SELECT COUNT() FROM posts WHERE thread = p.thread) < :limit
]])

local query_file_refs_of_post = db:prepare_mult('SELECT file_ref FROM file_refs WHERE post = ?')
local query_file_ref_at_post_position = db:prepare([[
  SELECT file_ref FROM (
    SELECT file_ref, ROW_NUMBER() OVER (ORDER BY position) AS pos FROM file_refs WHERE post = ?
  ) WHERE pos = ?
]])
local query_hash_of_file_ref = db:prepare('SELECT hash FROM file_refs WHERE file_ref = ?')

local query_comment = db:prepare('SELECT comment FROM posts WHERE post = ?')
local query_subject = db:prepare('SELECT subject FROM posts WHERE post = ?')
local set_comment = db:prepare('UPDATE posts SET comment = ? WHERE post = ?')

local function redact(ctx, passage, text)
  local hash = openssl.hex(openssl.digest.get('sha256'):digest(text))
  local message = '&lt;redacted, sha256=' .. hash .. '&gt;'
  local styled = '<span class="redacted">' .. message .. '</span>'
  local i, j = passage:find(text, 1, true)
  while i do
    local part1 = passage:sub(1, i - 1)
    local part2 = passage:sub(j + 1)
    if part1:find('<[^>]*$') then
      passage = part1 .. message .. part2
    else
      passage = part1 .. styled .. part2
    end
    i, j = passage:find(text, passage:len() - part2:len() + 1, true)
  end
  ctx:result_text(passage)
end

db.raw:create_function('redact', 2, redact)
db_readonly.raw:create_function('redact', 2, redact)

local redact_comment_update = db:prepare("UPDATE posts SET comment = redact(comment, ?) WHERE post = ?")
local redact_subject_update = db:prepare("UPDATE posts SET subject = redact(subject, ?) WHERE post = ?")

local function fix_links(post)
  local comment = query_comment(post)
  local links = gen.find_links(comment)
  delete_links(post)
  for _, link in ipairs(links) do
    link.src = post
    insert_link(link)
  end
end

local function redact_text(post, texts)
  for _, text in ipairs(texts) do
    redact_comment_update(text, post)
    redact_subject_update(text, post)
  end
  fix_links(post)
end

local function thread_link(board, thread)
  return '<a href="/' .. board .. '/' .. thread .. '#P' .. thread .. '">/' .. board .. '/' .. thread .. '</a>'
end

local function board_link(board)
  return '<a href="/' .. board .. '/">/' .. board .. '/</a>'
end

local function log_actions(name, cbs, args, ...)
  local args2 = common.pack(...)
  return db:transaction(function()
    insert_action({
      mod = args.mod and gen.escape(args.mod),
      func = name,
      args = cjson.encode(args),
      reason = args.reason and gen.linkify(args.reason),
      status = 1
    })
    local description = ''
    local any_needs_redact = false
    for i, cb in ipairs(cbs) do
      local subdescription, err, needs_redact = cb(args, unpack(args2, 1, args2.n))
      if not subdescription then
        return nil, err
      end
      if i == 1 then
        description = subdescription
      else
        if #cbs > 2 then
          description = description .. ','
        end
        if i == #cbs then
          description = description .. ' and'
        end
        description = description .. ' ' .. subdescription:sub(1,1):lower() .. subdescription:sub(2)
      end
      any_needs_redact = any_needs_redact or needs_redact
    end
    if any_needs_redact then
      copy_changes_to_private()
      redact_changes()
    end
    set_description(description)
    action_complete()
    trigger_rebuild()
    return true
  end)
end

local action_callbacks = {}

local function action(name, cb)
  action_callbacks[name] = cb
  mod[name] = function(args, ...)
    log_actions(name, {cb}, args, ...)
  end
end

function mod.multistep(args, ...)
  local cbs = {}
  for _, name in ipairs(args.actions) do
    table.insert(cbs, action_callbacks[name])
  end
  log_actions('multistep', cbs, args, ...)
end

action('move_to_board', function(args)
  for _, post in ipairs(args.posts) do
    local thread = query_thread(post)
    local board = query_board(thread)
    if thread and args.dest ~= board and post == thread then
      move_thread_to_board(args.dest, thread)
    end
  end
  for _, post in ipairs(args.posts) do
    local thread = query_thread(post)
    local board = query_board(thread)
    if thread and args.dest ~= board and post ~= thread then
      local collect = query_move_to_board_collect{
        post = post,
        thread = thread,
        dest = args.dest,
        limit = config.THREAD_MAX_POSTS
      }
      if collect then
        move_post_to_thread(collect, post)
      else
        create_thread(post, args.dest)
        move_post_to_thread(post, post)
      end
    end
  end
  return 'Moved to ' .. board_link(args.dest)
end)

action('move_to_thread', function(args)
  local dest_thread = query_thread(args.dest)
  if not dest_thread then
    return nil, 'destination ' .. args.dest .. ' does not exist'
  end
  local board = query_board(dest_thread)
  if args.dest ~= dest_thread then
    local current_board = query_board(args.dest)
    if not current_board then
      create_thread(args.dest, board)
    elseif current_board ~= board then
      move_thread_to_board(board, args.dest)
    end
    move_post_to_thread(args.dest, args.dest)
  end
  for _, post in ipairs(args.posts) do
    local thread = query_thread(post)
    if thread then
      if post == thread and post ~= args.dest then
        move_thread_to_thread(args.dest, thread)
      else
        move_post_to_thread(args.dest, post)
      end
    end
  end
  return 'Moved to ' .. thread_link(board, args.dest)
end)

action('delete_files', function(args)
  local hashes = mod.hashes(args)
  for _, hash in ipairs(hashes) do
    delete_file(hash)
    gen.shred_files(hash)
  end
  return ((#hashes == 1) and 'File' or 'Files') .. ' deleted'
end)

action('ban_files', function(args)
  local hashes = mod.hashes(args)
  for _, hash in ipairs(hashes) do
    ban_file(hash)
    gen.shred_files(hash)
  end
  return ((#hashes == 1) and 'File' or 'Files') .. ' banned'
end)

action('unban_files', function(args)
  local hashes = mod.hashes(args)
  for _, hash in ipairs(hashes) do
    unban_file(hash)
  end
  return ((#hashes == 1) and 'File' or 'Files') .. ' unbanned'
end)

action('spoiler_files', function(args)
  local file_refs = mod.file_refs(args)
  for _, file_ref in ipairs(file_refs) do
    set_spoiler(args.spoiler, file_ref)
  end
  local desc
  if args.spoiler == 'Spoiler' then
    desc = 'spoilered'
  elseif not args.spoiler then
    desc = 'unspoilered'
  else
    desc = 'marked ' .. args.spoiler
  end
  return ((#file_refs == 1) and 'File' or 'Files') .. ' ' .. desc
end)

action('redact_files', function(args)
  local hashes = mod.hashes(args)
  local redacted = (args.redacted == nil) and true or args.redacted
  for _, hash in ipairs(hashes) do
    redact_file(redacted and 'Redacted' or nil, hash)
    gen.redact_file(hash, redacted)
  end
  return ((#hashes == 1) and 'File' or 'Files') .. (redacted and ' redacted' or ' unredacted')
end)

action('redact_text', function(args, ...)
  for _, post in ipairs(args.posts) do
    redact_text(post, {...})
  end
  return 'Text redacted', nil, true
end)

action('redact_links', function(args)
  local nlinks = 0
  for _, post in ipairs(args.posts) do
    local links = {}
    for _, text in ipairs{query_comment(post), query_subject(post)} do
      for link in text:gmatch('href="https?://([^"]+)"') do
        if not link:find('&lt;redacted') then
          nlinks = nlinks + 1
          link = link:gsub('^https?://', '')
          table.insert(links, link)
        end
      end
    end
    table.sort(links, function(a, b) return a > b end)
    redact_text(post, links)
  end
  return ((nlinks == 1) and 'Link' or 'Links') .. ' redacted', nil, true
end)

action('rerun_markup', function(args)
  for _, post in ipairs(args.posts) do
    local comment = query_comment(post)
    comment = gen.markup(gen.unmarkup(comment))
    set_comment(comment, post)
    fix_links(post)
  end
  return 'Markup rerun'
end)

action('set_overboard_exclude', function(args)
  for _, post in ipairs(args.posts) do
    local thread = query_thread(post)
    if thread and post == thread then
      set_overboard_exclude_thread(args.exclude, thread)
    else
      set_overboard_exclude_post(args.exclude, post)
    end
  end
  return 'Overboard exclude set to ' ..  (({[true] = 'true', [false] = 'false'})[args.exclude] or 'NULL')
end)

action('config_board', function(args)
  insert_board(args.board)
  local desc = ''
  for key, val in pairs(args) do
    if set_board_prop[key] then
      set_board_prop[key](val and 1 or 0, args.board)
      if desc ~= '' then
        desc = desc .. ' and '
      end
      desc = desc .. key .. ' to ' .. (val and 'true' or 'false')
    end
  end
  return 'Board /' .. args.board .. '/ configured setting ' .. desc
end)

action('create_mod', function(args)
  local password = login.generate()
  local hash = login.get_hash(password)
  insert_login(gen.escape(args.account))
  set_password(hash, gen.escape(args.account))
  set_mod_status(1, gen.escape(args.account))
  print('Password: ' .. password)
  return 'New moderator account ' .. gen.escape(args.account) .. ' created'
end)

action('reset_mod_password', function(args)
  local password = login.generate()
  local hash = login.get_hash(password)
  set_password(hash, gen.escape(args.account))
  print('Password: ' .. password)
  return 'Password reset for moderator account ' .. gen.escape(args.account)
end)

action('set_mod_password', function(args)
  io.stdout:write('Password: ')
  local password = io.stdin:read()
  local hash = login.get_hash(password)
  set_password(hash, gen.escape(args.account))
  return 'Password changed for moderator account ' .. gen.escape(args.account)
end)

action('set_mod_status', function(args)
  set_mod_status(args.enabled, gen.escape(args.account))
  return 'Moderator account ' .. gen.escape(args.account) .. ' set to ' .. (args.enabled and 'enabled' or 'disabled')
end)

action('exec', function(args)
  db:exec(args.query)
  return 'SQL query executed: "' .. gen.escape(args.query) .. '"'
end)

action('exec_redacted', function(args, ...)
  local stmt = db.raw:prepare(args.query)
  db:check(db.raw:errcode())
  db:bind(stmt, ...)
  db:check(stmt:step())
  db:check(stmt:finalize())
  return 'Redacted SQL query executed: "' .. gen.escape(args.query) .. '"', nil, true
end)

function mod.query(query, do_print)
  local results = {}
  local stmt = db_readonly.raw:prepare(query)
  db_readonly:check(db_readonly.raw:errcode())
  local n = stmt:columns()
  for row in stmt:rows() do
    table.insert(results, row[1])
    if do_print then
      print(unpack(row, 1, n))
    end
  end
  db_readonly:check(stmt:finalize())
  return results
end

function mod.prepare(...)
  return db_readonly:prepare(...)
end

function mod.stdin()
  local posts = {}
  local post
  repeat
    post = io.read('*n')
    if post then
      table.insert(posts, post)
    end
  until not post
  return posts
end

function mod.file_refs(args)
  local file_refs = {}
  for _, file_ref in ipairs(args.file_refs or {}) do
    table.insert(file_refs, file_ref)
  end
  for _, post in ipairs(args.posts or {}) do
    query_file_refs_of_post(file_refs, post)
  end
  for _, post_position in ipairs(args.post_positions or {}) do
    table.insert(file_refs, query_file_ref_at_post_position(unpack(post_position)))
  end
  return file_refs
end

function mod.hashes(args)
  local hashes = {}
  for _, hash in ipairs(args.hashes or {}) do
    table.insert(hashes, hash)
  end
  for _, file_ref in ipairs(mod.file_refs(args)) do
    table.insert(hashes, query_hash_of_file_ref(file_ref))
  end
  return hashes
end

function mod.range(start, finish)
  local results = {}
  for i = start, finish do
    table.insert(results, i)
  end
  return results
end

mod.escape = gen.escape

return mod
