local multipart = {}

local function rtrim(s)
  return (string.gsub(s, '[ \t]*$', ''))
end

-- make case insensitive pattern
local function ci(s)
  local pattern = ''
  for i = 1, string.len(s) do
    local c = string.sub(s, i, i)
    if string.upper(c) == string.lower(c) then
      pattern = pattern .. c
    else
      pattern = pattern .. '[' .. string.upper(c) .. string.lower(c) .. ']'
    end
  end
  return pattern
end

local DISPOSITION_PATTERN = '\r\n' .. ci('content%-disposition') .. '[ \t]*:[^\r\n]*\r\n'
local FILENAME_PATTERN    = ';[ \t]*' .. ci('filename') .. '[ \t]*=[ \t]*'
local NAME_PATTERN        = ';[ \t]*' .. ci('name')     .. '[ \t]*=[ \t]*'
local BOUNDARY_PATTERN =
  '^' .. ci('multipart/form%-data') .. '.*;[ \t]*' .. ci('boundary') .. '[ \t]*=[ \t]*"?([^"]*)'

-- assumes " absent or escaped as %22, \ absent or not escaped
-- https://github.com/whatwg/html/pull/6282
local function param_value(headers, i)
  if string.sub(headers, i, i) == '"' then
    return string.match(headers, '^[^"\r\n]*', i + 1), true
  else
    return rtrim(string.match(headers, '^[^;\r\n]*', i)), false
  end
end

local function parse_disposition(disposition)
  local name, filename

  -- filename
  local _, end_match = string.find(disposition, FILENAME_PATTERN)
  if end_match then
    local quoted
    filename, quoted = param_value(disposition, end_match + 1)
    if quoted then
      local end_filename = end_match + 1 + string.len(filename)
      -- support old Firefox escaping of " as \"
      if string.sub(filename, -1) == '\\' then
        local filename2 = string.match(disposition, '([^\r\n]*)"\r\n', end_match + 2)
        if filename2 and not string.find(filename2, '[^\\]"') then
          end_filename = end_match + 1 + string.len(filename2)
          filename = string.gsub(filename2, '\\"', '"')
        end
      end
      -- remove filename to prevent it from breaking other pattern matches
      disposition = string.sub(disposition, 1, end_match + 1) .. string.sub(disposition, end_filename + 1)
    end
    -- restore quotation marks (leave \r and \n escaped)
    filename = string.gsub(filename, '%%22', '"')
    -- remove path if present
    filename = string.gsub(filename, '^.*[/\\]', '')
  end

  -- field name
  _, end_match = string.find(disposition, NAME_PATTERN)
  if end_match then
    name = param_value(disposition, end_match + 1)
  end

  return name, filename
end

function multipart.parse(data, boundary, file_fields, max_fields, max_files)
  file_fields = file_fields or {}

  local results = {}
  for _, name in ipairs(file_fields) do
    results[name] = {}
  end

  local count_fields = 0
  local count_files  = 0
  local _, prev_boundary = string.find(data, '--' .. boundary, 1, true)
  while prev_boundary do
    local res = {}
    local name = nil

    -- closing boundary
    if string.sub(data, prev_boundary + 1, prev_boundary + 3) == '--' then
      break
    end

    -- limit iterations
    count_fields = count_fields + 1
    if count_fields > max_fields then
      return nil, 'maximum of ' .. max_fields .. ' form fields exceeded'
    end

    -- headers
    local _, end_headers =
      string.find(data, '\r\n\r\n', prev_boundary + 1, true)
    if not end_headers then
      break
    end
    local headers = string.sub(data, prev_boundary + 1, end_headers)
    -- unfolding
    headers = string.gsub(headers, '\r\n([ \t])', '%1')

    -- Content-Disposition header
    local disposition = string.match(headers, DISPOSITION_PATTERN)
    if disposition then
      name, res.filename = parse_disposition(disposition)
    end

    -- next boundary
    local beg_boundary, end_boundary =
      string.find(data, '\r\n--' .. boundary, end_headers + 1, true)
    if beg_boundary and name then
      local body = string.sub(data, end_headers + 1, beg_boundary - 1)
      if type(results[name]) == 'table' then
        -- for file fields return array of one table for each file
        -- empty file indicates no file selected
        if string.len(body) > 0 then
          res.body = body
          table.insert(results[name], res)
          count_files = count_files + 1
          if count_files > max_files then
            return nil, 'maximum of ' .. max_files .. ' files exceeded'
          end
        end
      else
        -- convert \r\n back to \n in text fields
        body = string.gsub(body, '\r\n', '\n')
        -- for text fields just return body
        results[name] = body
      end
    end
    prev_boundary = end_boundary
  end
  return results
end

function multipart.boundary(content_type)
  return string.match(content_type, BOUNDARY_PATTERN)
end

return multipart
