#!/usr/bin/luajit
local fcgi = require('fcgi')
local openssl = require('openssl')
local config = require('config')
local common = require('common')
local db = require('database')('../Data/board.db'):read('cache.sql')
local multipart = require('multipart')
local gen = require('gen')
local captcha = require('captcha')
local login = require('login')
local mod = require('mod')

local POST_FIELDS = {}
for _, key in ipairs({'parent', 'subject', 'comment', 'captcha', 'spoiler', 'modpassword'}) do
  POST_FIELDS[key] = {
    max_length   = config.MAX_LENGTH[key],
    max_newlines = config.MAX_NEWLINES[key] or 0
  }
end

local FILENAME_RULES = {
  max_length   = config.MAX_LENGTH.filename,
  max_newlines = 0
}

local REASON_RULES = {
  max_length   = config.MAX_LENGTH.reason,
  max_newlines = 0
}

local insert_thread = db:prepare([[
  INSERT INTO threads (thread, board, original_board)
    VALUES (:thread, :board, :board);
]])

local insert_post = db:prepare([[
  INSERT INTO posts (post, thread, original_thread, moderator, subject, date, comment)
    VALUES (:post, :thread, :thread, :moderator, :subject, STRFTIME('%s', 'now'), :comment);
]])

local insert_link = db:prepare([[
  INSERT INTO links (src, position, dest)
    VALUES (:src, :position, :dest);
]])

local insert_file = db:prepare([[
  INSERT OR IGNORE INTO files (hash) VALUES (:hash);
]])

local file_fields = 'extension, thumb, thumb_catalog, size, width, height, duration, thumb_width, thumb_height, type'
local update_file = db:prepare([[
  UPDATE files SET ]] .. string.gsub(file_fields, '[%w_]+', '%0 = IFNULL(:%0, %0)') .. [[,
      status = IIF(status = 'Redacted', 'Redacted', NULL)
    WHERE hash = :hash;
]])

local insert_file_ref = db:prepare([[
  INSERT INTO file_refs (file_ref, post, position, hash, name, name_uri, spoiler)
    VALUES (:file_ref, :post, :position, :hash, :name, :name_uri, :spoiler);
]])

local trigger_rebuild = db:prepare([[
  INSERT INTO start_rebuild VALUES (1);
]])

local query_thread_length = db:prepare([[
  SELECT COUNT() FROM posts WHERE thread = ?;
]])

local query_thread_staleness = db:prepare([[
  SELECT STRFTIME('%s', 'now') - MAX(date) FROM posts WHERE thread = ?;
]])

local query_last = db:prepare([[
  SELECT last_post, last_file_ref FROM global;
]])

local query_link_attr = db:prepare([[
  SELECT thread, attr_intra, attr_inter FROM link_attr WHERE post = ?;
]])

local query_board = db:prepare([[
  SELECT board FROM threads WHERE thread = :thread;
]])

local query_file_banned = db:prepare([[
  SELECT status = 'Banned' FROM files WHERE hash = ?
]])

local query_new_threads_disabled = db:prepare([[
  SELECT COUNT() FROM boards WHERE board = ? AND new_threads_disabled = 1
]])

local query_text_only = db:prepare([[
  SELECT COUNT() FROM boards WHERE board = ? AND text_only = 1
]])

local invalid_utf8 = {
  -- always invalid
  '[\192\193\245-\255]',
  -- too few or too many continuation bytes
  '^[\128-\191]',
  '[%z\1-\127][\128-\191]',
  '[\194-\244][^\128-\191]',
  '[\194-\244]$',
  '[\194-\223][\128-\191][\128-\191]',
  '[\224-\244][\128-\191][^\128-\191]',
  '[\224-\244][\128-\191]$',
  '[\224-\239][\128-\191][\128-\191][\128-\191]',
  '[\240-\244][\128-\191][\128-\191][^\128-\191]',
  '[\240-\244][\128-\191][\128-\191]$',
  '[\240-\244][\128-\191][\128-\191][\128-\191][\128-\191]',
  -- overlong
  '\224[\128-\159]',
  '\240[\128-\143]',
  -- beyond U+10FFFF
  '\244[\144-\191]',
  -- surrogate halves
  '\237[\160-\191]',
  -- invalid characters
  '\239\191[\190\191]',
  '\239\183[\144-\175]',
  '[\240-\244][\143\159\175\191]\191[\190\191]'
}

local function valid_utf8(s)
  for _, regexp in ipairs(invalid_utf8) do
    if s:find(regexp) then
      return false
    end
  end
  return true
end

local function validate(text, rules)
  if string.len(text) > 3 * rules.max_length then
    return false, 'has more than ' .. rules.max_length .. ' characters (as UTF-16)'
  end
  if not valid_utf8(text) then
    return false, 'is not valid UTF-8'
  end
  if common.utf16len(text) > rules.max_length then
    return false, 'has more than ' .. rules.max_length .. ' characters (as UTF-16)'
  end
  if string.find(text, '[%z\1-\8\11\12\14-\31\127]') or string.find(text, '\194[\128-\159]') then
    return false, 'contains control characters'
  end
  local _, newlines = string.gsub(text, '\n', '')
  if newlines > rules.max_newlines then
    return false, 'has more than ' .. rules.max_newlines .. ' newlines'
  end
  return true
end

local function files_allowed(hashes)
  for _, hash in ipairs(hashes) do
    if query_file_banned(hash) == 1 then
      return false, 'banned file'
    end
  end
  return true
end

local function posting_allowed(args)
  if args.thread then
    local count = query_thread_length(args.thread)
    if count == 0 then
      return false, 'thread does not exist'
    elseif count >= config.THREAD_MAX_POSTS then
      return false, 'thread has reached ' .. config.THREAD_MAX_POSTS .. ' posts and is locked'
    end
  end
  if args.board and query_new_threads_disabled(args.board) == 1 then
    return false, 'new threads disabled on /' .. args.board .. '/'
  end
  local board = args.board or query_board(args.thread)
  if args.nfiles > 0 and query_text_only(board) == 1 then
    return false, '/' .. board .. '/ is text only'
  end
  local is_necro = (query_thread_staleness(args.thread) or 0) > 30*86400
  local block_time, captcha_time, captcha_reason = captcha.cooldown(not args.is_reply or is_necro)
  local captcha_verified
  if block_time > 0 then
    return false,
      'posting ' .. (args.is_reply and 'replies' or 'threads')
        .. ' temporarily disabled for ' .. block_time .. ' seconds'
  elseif captcha_time > 0 then
    if captcha_reason == 'flood' and args.nchars < config.MIN_LENGTH.comment_flood then
      return false, 'wait a bit or post at least ' .. config.MIN_LENGTH.comment_flood .. ' characters'
    elseif args.captcha_verified then
      captcha_verified = true
    elseif args.captcha:find('%S') then
      captcha_verified = captcha.verify(args.captcha)
      if not captcha_verified then
        return false, 'incorrect captcha'
      end
    else
      local desc = args.is_reply and (is_necro and 'bumping old threads' or 'replies') or 'threads'
      return false, 'captcha temporarily required for ' .. desc
    end
  end
  if args.hashes then
    local allowed, reason = files_allowed(args.hashes)
    if not allowed then
      return false, reason
    end
  end
  return true, '', captcha_verified
end

local function static_links(text, thread)
  local cache = {}
  return text:gsub('<a>P(%d+)</a>', function(n)
    n = tonumber(n)
    local attr
    if cache[n] then
      attr = cache[n]
    else
      local thread2, attr_intra, attr_inter = query_link_attr(n)
      attr = ((thread == thread2) and attr_intra or attr_inter) or ''
      cache[n] = attr
    end
    return '<a' .. attr .. '>P' .. n .. '</a>'
  end)
end

local function rejection(reason, status)
  status = status or '400 Bad Request'
  return 'Status: ' .. status
    .. '\r\nContent-Type: text/plain; charset=utf-8\r\n\r\nError: '
    .. reason .. '\n'
end

local function handle_post(write)
  local start_time = os.time()

  -- parse post
  local content_type = fcgi.getenv('CONTENT_TYPE')
  local boundary = multipart.boundary(content_type or '')
  if not boundary then
    return write(rejection('post sent with wrong content type'))
  end

  local data = fcgi.post()
  local form, err = multipart.parse(data, boundary, {'file'}, config.MAX_FILES + config.MAX_FIELDS, config.MAX_FILES)
  if not form then
    return write(rejection(err))
  end

  -- validate fields
  for name, rules in pairs(POST_FIELDS) do
    if not form[name] then
      form[name] = ''
    end
    local success, reason = validate(form[name], rules)
    if not success then
      return write(rejection(name .. ' ' .. reason))
    end
  end

  local is_reply = form.parent:match('^[0-9]+$')
  if not is_reply and not (
    form.parent:len() >= config.MIN_LENGTH.board and
    form.parent:len() <= config.MAX_LENGTH.board and
    form.parent:match('^[a-z]+$')
  ) then
    return write(rejection(
      'board name must consist of '
        .. config.MIN_LENGTH.board .. '-'
        .. config.MAX_LENGTH.board .. ' lowercase letters'
    ))
  end

  if form.comment == '' and # form.file == 0 then
    return write(rejection('nothing posted'))
  end

  if not is_reply and common.utf16len(form.comment) < config.MIN_LENGTH.op then
    return write(rejection(
      'To start a new thread, the comment must contain at least ' .. config.MIN_LENGTH.op .. ' characters (as UTF-16)'
    ))
  end

  if form.keepfilenames then
    for _, field in ipairs(form.file) do
      local success, reason = validate(field.filename, FILENAME_RULES)
      if not success then
        return write(rejection('filename ' .. reason))
      end
    end
  end

  -- process post
  local post = {}
  if is_reply then
    post.thread = tonumber(form.parent)
  else
    post.board = form.parent
  end
  if (form.modpassword or '') ~= '' then
    post.moderator = login.verify(form.modpassword)
    if not post.moderator then
      return write(rejection('incorrect password', '403 Forbidden'))
    end
  end
  local nchars = common.utf8len(form.comment)
  local allowed, reason, captcha_verified = posting_allowed{
    is_reply = is_reply,
    board = post.board,
    thread = post.thread,
    nchars = nchars,
    nfiles = # form.file,
    captcha = form.captcha,
    moderator = post.moderator
  }
  if not allowed then
    return write(rejection(reason, '403 Forbidden'))
  end
  post.subject = gen.escape(form.subject)
  post.comment = gen.markup(form.comment)
  local links = gen.find_links(post.comment)
  if # links > config.MAX_DYNAMIC_LINKS then
    post.comment = static_links(post.comment, post.thread)
    links = {}
  end

  local files = {}
  local hashes = {}
  for i, field in ipairs(form.file) do
    hashes[i] = openssl.hex(openssl.digest.get('sha256'):digest(field.body))
  end
  allowed, reason = files_allowed(hashes)
  if not allowed then
    return write(rejection(reason, '403 Forbidden'))
  end
  for i, field in ipairs(form.file) do
    local path = gen.write_file(hashes[i], field.body)
    local file = gen.handle_file(hashes[i], path, string.len(field.body), is_reply, start_time)
    file.position = i
    if form.keepfilenames and string.len(field.filename or '') > 0 then
      file.name = gen.escape(field.filename)
      file.name_uri = gen.encodeURIComponent(field.filename)
    end
    if form.spoiler:len() > 0 then
      file.spoiler = gen.escape(form.spoiler)
    end
    table.insert(files, file)
  end

  -- insert into database
  allowed, reason = db:transaction(function()
    -- whether posting allowed may have changed since prior check
    allowed, reason = posting_allowed{
      is_reply = is_reply,
      board = post.board,
      thread = post.thread,
      nchars = nchars,
      nfiles = # form.file,
      hashes = hashes,
      captcha = form.captcha,
      captcha_verified = captcha_verified,
      moderator = post.moderator
    }
    if not allowed then
      for _, file in ipairs(files) do
        gen.cancel_file(file)
      end
      return false, reason
    end
    local last_post, last_file_ref = query_last()
    post.post = last_post + 1
    if not is_reply then
      post.thread = post.post
      insert_thread(post)
    end
    insert_post(post)
    for _, link in ipairs(links) do
      link.src = post.post
      insert_link(link)
    end
    for i, file in ipairs(files) do
      gen.commit_file(file)
      file.post = post.post
      file.file_ref = last_file_ref + i
      insert_file(file)
      update_file(file)
      insert_file_ref(file)
    end
    trigger_rebuild()
    return true
  end)
  if not allowed then
    return write(rejection(reason, '403 Forbidden'))
  end

  -- redirect to post
  if not post.board then
    post.board = query_board(post)
  end

  write(
    'Status: 303 See Other\r\nLocation: /'
      .. post.board .. '/' .. post.thread .. '#P' .. post.post .. '\r\n\r\n'
  )
end

local query_post_status = db:prepare([[
  SELECT board, posts.thread, STRFTIME('%s', 'now') - date, threads.overboard_exclude, posts.overboard_exclude
    FROM posts JOIN threads ON posts.thread = threads.thread WHERE post = ?;
]])

local function contains(haystack, needle)
  for _, val in ipairs(haystack) do
    if val == needle then
      return true
    end
  end
  return false
end

local function list_posts(heading, posts)
  local output = heading .. ':'
  for _, post in ipairs(posts) do
    output = output .. ' <a href="/P' .. post .. '">P' .. post .. '</a>'
  end
  return output .. '<br />\n'
end

local function percent_decode(text)
  return text:gsub('%+', ' '):gsub('%%[0-9A-F][0-9A-F]', function(p)
    return string.char(tonumber(p:sub(2), 16))
  end)
end

local function mod_action(write)
  local data = fcgi.post()
  if data:len() > config.MAX_MOD_REQUEST then
    return write('request too large')
  end
  data = '&' .. data

  -- parse and validate password
  local password = select(3, data:find('&modpassword=([^&]+)'))
  if not password or password:len() == 0 then
    return write(rejection('missing password'))
  end
  password = percent_decode(password)
  if common.utf16len(password) > config.MAX_LENGTH.modpassword then
    return write(rejection('password has too many characters'))
  end
  local moderator, status = login.verify(password)
  if not moderator then
    return write(rejection('incorrect password', '403 Forbidden'))
  end
  if status ~= 1 then
    return write(rejection('account disabled', '403 Forbidden'))
  end

  -- parse and validate action
  local action, param = select(3, data:find('&action_([a-z]+)_?([a-z]*)='))
  if not action then
    return write(rejection('missing action'))
  end
  if not (action == 'move' or action == 'cp') then
    return write(rejection('unrecognized action'))
  end
  if action == 'move' then
    if param == '' then
      return write(rejection('missing board name'))
    end
    if not contains(config.MOVE_OPTIONS, param) then
      return write(rejection('cannot move to /' .. param .. '/'))
    end
  end

  -- parse and validate reason
  local reason = percent_decode(select(3, data:find('&reason=([^&]+)')) or '')
  local success, reject_reason = validate(reason, REASON_RULES)
  if not success then
    return write(rejection('reason ' .. reject_reason))
  end
  if not reason:find('%S') then
    if action == 'cp' then
      reason = 'appears to be child nudity/pornography'
    else
      reason = nil
    end
  end

  -- parse post numbers
  local posts = {}
  for post in data:gmatch('&C([0-9]+)=') do
    table.insert(posts, tonumber(post))
  end
  table.sort(posts)

  -- URL to return to
  local ref = select(3, data:find('&ref=([0-9A-Za-z%%]+)'))
  if ref then
    ref = ref:gsub('%%2F', '/')
  end
  if not ref or ref:find('%%') then
    ref = '/All/'
  end

  -- decide action on each post
  local not_found, too_old, moved, excluded, unmoved, unmoved_board = {}, {}, {}, {}, {}, {}
  for _, board in ipairs(config.MOVE_OPTIONS) do
    unmoved_board[board] = {}
  end
  for _, post in ipairs(posts) do
    local board, thread, age, excluded_thread, excluded_post = query_post_status(post)
    if not board then
      table.insert(not_found, post)
    elseif age > config.MOD_MAX_POST_AGE then
      table.insert(too_old, post)
    else
      if contains(config.MOVE_OPTIONS, board) then
        if excluded_thread == 0 and post ~= thread and excluded_post ~= 1 then
          table.insert(excluded, post)
        else
          table.insert(unmoved, post)
          table.insert(unmoved_board[board], post)
        end
      else
        table.insert(moved, post)
      end
    end
  end

  -- carry out actions
  local output = ''
  local moved_prefix = (action == 'cp') and 'redacted and ' or ''
  local function add_actions(actions)
    if action == 'cp' then
      table.insert(actions, 1, 'redact_files')
      table.insert(actions, 2, 'redact_links')
    end
    return actions
  end
  if # moved > 0 then
    local actions = add_actions{'move_to_board'}
    local dest = (action == 'cp') and config.SPAM_BOARD or param
    mod.multistep{mod = moderator, actions = actions, posts = moved, dest = dest, reason = reason}
    output = output .. list_posts(moved_prefix .. 'moved to /' .. dest .. '/', moved)
  end
  if # excluded > 0 then
    local actions = add_actions{'set_overboard_exclude'}
    mod.multistep{mod = moderator, actions = actions, posts = excluded, exclude = true, reason = reason}
    output = output .. list_posts(moved_prefix .. 'excluded from overboard', excluded)
  end
  if # unmoved > 0 then
    if action == 'cp' then
      mod.multistep{mod = moderator, actions = add_actions{}, posts = unmoved, reason = reason}
    end
    for _, board in ipairs(config.MOVE_OPTIONS) do
      if # unmoved_board[board] > 0 then
        output = output .. list_posts(moved_prefix .. 'already on /' .. board .. '/', unmoved_board[board])
      end
    end
  end
  if # too_old > 0 then
    output = output .. list_posts('too old', too_old)
  end
  if # not_found > 0 then
    output = output .. list_posts('not found', not_found)
  end

  write(
    'Status: 200 OK\r\nContent-Type: text/html; charset=utf-8\r\n\r\n'
      .. '<!DOCTYPE html><html><head><title>Mod action results</title></head><body style="font-family: monospace;">\n'
      .. '<div style="float: right;">[<a href="' .. gen.escape(ref) .. '">Back</a>]</div>\n'
      .. output .. '</body></html>'
  )
end

local function route(write)
  if fcgi.getenv('REQUEST_METHOD') ~= 'POST' then
    return write(rejection("You're supposed to POST to this URL.", '405 Method Not Allowed'))
  end
  local uri = fcgi.getenv('REQUEST_URI')
  if uri == '/Post' then
    return handle_post(write)
  elseif uri == '/Mod' then
    return mod_action(write)
  else
    return write(rejection('404 Not Found', '404 Not Found'))
  end
end

while fcgi.accept() do
  common.handle_errors(route, fcgi.print)
end
