#include "sqlite_prepared_statement.h"

#include <stdexcept>
#include <utility>

namespace database {

SqlitePreparedStatement::SqlitePreparedStatement(sqlite3* dbHandle, std::string_view sql) : dbHandle(dbHandle) {
  if (sqlite3_prepare(dbHandle, sql.data(), sql.length(), &handle, nullptr) != SQLITE_OK) {
    throw std::runtime_error(sqlite3_errmsg(dbHandle));
  }
}

SqlitePreparedStatement::~SqlitePreparedStatement() { sqlite3_finalize(handle); }

SqlitePreparedStatement::SqlitePreparedStatement(SqlitePreparedStatement&& other) noexcept {
  dbHandle = std::exchange(other.dbHandle, nullptr);
  handle = std::exchange(other.handle, nullptr);
}

SqlitePreparedStatement& SqlitePreparedStatement::operator=(SqlitePreparedStatement&& other) noexcept {
  std::swap(dbHandle, other.dbHandle);
  std::swap(handle, other.handle);
  return *this;
}

void SqlitePreparedStatement::bind(int column, bool data) {
  if (sqlite3_bind_int64(handle, column, data) != SQLITE_OK) {
    throw std::runtime_error(sqlite3_errmsg(dbHandle));
  }
}

void SqlitePreparedStatement::bind(int column, int64_t data) {
  if (sqlite3_bind_int64(handle, column, data) != SQLITE_OK) {
    throw std::runtime_error(sqlite3_errmsg(dbHandle));
  }
}

void SqlitePreparedStatement::bind(int column, std::string_view data) {
  if (sqlite3_bind_text(handle, column, data.data(), data.length(), SQLITE_STATIC) != SQLITE_OK) {
    throw std::runtime_error(sqlite3_errmsg(dbHandle));
  }
}

std::span<const uint8_t> SqlitePreparedStatement::getBlob(int column) {
  return {static_cast<const uint8_t*>(sqlite3_column_blob(handle, column)),
          static_cast<size_t>(sqlite3_column_bytes(handle, column))};
}

bool SqlitePreparedStatement::getBool(int column) { return getInt(column); }

int64_t SqlitePreparedStatement::getInt(int column) { return sqlite3_column_int64(handle, column); }

std::string SqlitePreparedStatement::getString(int column) {
  return reinterpret_cast<const char*>(sqlite3_column_text(handle, column));
}

QueryResult SqlitePreparedStatement::step() {
  switch (sqlite3_step(handle)) {
    case SQLITE_ROW:
      return QueryResult::row;
    case SQLITE_DONE:
      return QueryResult::done;
    default:
      throw std::runtime_error(sqlite3_errmsg(dbHandle));
  }
}

}  // namespace database