#include "captcha_system.h"

#include <absl/strings/str_cat.h>
#include <sys/random.h>

#include <iostream>
#include <memory>

namespace imageboard {

constexpr std::unordered_map<std::string, std::span<uint8_t>>::size_type cacheSize = 30;
constexpr size_t captchaWidth = 310;
constexpr size_t captchaHeight = 100;
constexpr size_t captchaLength = 6;
constexpr time_t captchaMaxAge = 60 * 60;
constexpr time_t captchaDeathAge = 90 * 60;
constexpr std::string_view easyAlphabet = "0123456789ABCDEFGHIJKLNPQRSTUVXYZacefghijknorstuvxyz";
constexpr std::string_view hardAlphabet = "0123456789dx+-^/";

static_assert(captchaWidth > 0);
static_assert(captchaHeight > 0);
static_assert(captchaLength > 0);

static std::string toLower(std::string_view string) {
  std::string ret;
  for (char c : string) {
    if (c >= 'A' && c <= 'Z') {
      ret += c - ('A' - 'a');
    } else {
      ret += c;
    }
  }
  return ret;
}

static std::vector<size_t> pickEasySolution() {
  std::vector<size_t> solution;
  uint64_t random[captchaLength * sizeof(uint64_t)];
  while (getrandom(&random, sizeof(random), 0) == -1)
    ;
  for (size_t i = 0; i < captchaLength; i++) {
    size_t c = random[i] % easyAlphabet.length();
    solution.push_back(c);
  }
  return solution;
}

static std::vector<size_t> pickHardSolution() {
  std::vector<size_t> solution;
  uint64_t random[4 * sizeof(uint64_t)];
  while (getrandom(&random, sizeof(random), 0) == -1)
    ;
  solution.push_back(hardAlphabet.find('d'));
  solution.push_back(hardAlphabet.find('/'));
  solution.push_back(hardAlphabet.find('d'));
  solution.push_back(hardAlphabet.find('x'));
  solution.push_back(random[0] % 9 + 1);
  solution.push_back(hardAlphabet.find('x'));
  solution.push_back(hardAlphabet.find('^'));
  solution.push_back(random[1] % 7 + 3);
  solution.push_back(hardAlphabet.find('+'));
  solution.push_back(random[2] % 9 + 1);
  solution.push_back(hardAlphabet.find('x'));
  solution.push_back(hardAlphabet.find('^'));
  solution.push_back(random[3] % 7 + 3);
  return solution;
}

static std::vector<size_t> pickSolution(bool hard) {
  return hard ? pickHardSolution() : pickEasySolution();
}

CaptchaSystem::CaptchaSystem() {
  font::FontFace fontFace = freetype.open("DejaVu Sans Mono");
  fontFace.setSize(80);
  for (char c : std::string_view(easyAlphabet)) {
    easyGlyphs.emplace_back(fontFace.rasterize(c));
  }
  fontFace.setSize(40);
  for (char c : std::string_view(hardAlphabet)) {
    hardGlyphs.emplace_back(fontFace.rasterize(c));
  }
}

CaptchaSystem::CaptchaSystem(CaptchaSystem &&other) noexcept {
  compressor = std::move(other.compressor);
  easyGlyphs = std::move(other.easyGlyphs);
  hardGlyphs = std::move(other.hardGlyphs);
  captchas = std::move(other.captchas);
  oldCaptchas = std::move(other.oldCaptchas);
}

CaptchaSystem &CaptchaSystem::operator=(CaptchaSystem &&other) noexcept {
  compressor = std::move(other.compressor);
  easyGlyphs = std::move(other.easyGlyphs);
  hardGlyphs = std::move(other.hardGlyphs);
  captchas = std::move(other.captchas);
  oldCaptchas = std::move(other.oldCaptchas);
  return *this;
}

void CaptchaSystem::generate() {
  std::vector<size_t> solution = pickSolution(hard);
  std::string solutionString;
  // Find solution string
  if (!hard) {
    for (size_t i : solution) {
      solutionString += easyAlphabet[i];
    }
  } else {
    uint64_t a = solution[4] * solution[7];
    uint64_t b = solution[7] - 1;
    uint64_t c = solution[9] * solution[12];
    uint64_t d = solution[12] - 1;
    solutionString = absl::StrCat(a, "x^", b, "+", c, "x^", d);
  }
  uint64_t random[3 * solution.size() * sizeof(uint64_t)];
  while (getrandom(&random, sizeof(random), 0) == -1)
    ;
  uint8_t buffer[captchaWidth * captchaHeight];
  // Set the background to white
  memset(buffer, 0xFF, captchaWidth * captchaHeight);
  size_t offsets[solution.size() * 2];
  size_t offsetX = 0;
  // Draw each character
  for (size_t i = 0; i < solution.size(); i++) {
    // Draw character
    FT_Bitmap bitmap = (hard ? hardGlyphs : easyGlyphs)[solution[i]].bitmap();
    size_t offsetY = random[3 * i + 0] % std::max(captchaHeight - bitmap.rows, 1ul);
    if (hard && false) {
      auto jitter = std::make_unique<uint64_t[]>(bitmap.rows);
      while (getrandom(jitter.get(), bitmap.rows * sizeof(jitter[0]), 0) == -1)
        ;
      for (unsigned int y = 0; y < bitmap.rows; y++) {
        size_t rowOffsetX = std::min(offsetX + jitter[y] % 5, captchaWidth - 1);
        memcpy(buffer + (y + offsetY) * captchaWidth + rowOffsetX, bitmap.buffer + y * bitmap.width, bitmap.width);
      }
    } else {
      // Flip the character 50% of the time
      bool flipped = !hard && random[3 * i + 1] % 2 == 0;
      for (unsigned int y = 0; y < bitmap.rows; y++) {
        size_t rowOffsetX = std::min(offsetX, captchaWidth - 1);
        unsigned int row = y;
        if (flipped) {
          //row = bitmap.rows - y - 1;
        }
        memcpy(buffer + (y + offsetY) * captchaWidth + rowOffsetX, bitmap.buffer + row * bitmap.width, bitmap.width);
      }
    }
    // Save the center of the character
    offsets[i * 2 + 0] = offsetX + bitmap.width / 2;
    offsets[i * 2 + 1] = offsetY + bitmap.rows / 2;
    // Bump the offset forward
    offsetX += bitmap.width - 5 + random[3 * i + 2] % 10;
  }
  if (hard && false) {
    // Draw boxes between character centers
    for (size_t i = 0; i < solution.size() - 1; i++) {
      size_t sx = offsets[i * 2 + 0];
      size_t sy = offsets[i * 2 + 1];
      size_t ex = offsets[i * 2 + 2];
      size_t ey = offsets[i * 2 + 3];
      // ey should be bigger than sy
      if (ey < sy) {
        std::swap(sy, ey);
      }
      // If the box is too short max it bigger
      if (ey - sy < 5) {
        ey = std::min(ey + 10, captchaHeight - 1);
      }
      // Draw the box
      for (size_t y = sy; y <= ey; y++) {
        memset(buffer + y * captchaWidth + sx, 0xFF, ex - sx);
      }
    }
  }
  // Diagonal invert
  uint64_t points[4 * sizeof(int64_t)];
  while (getrandom(&points, sizeof(points), 0) == -1)
    ;
  int64_t x1 = points[0] % (captchaWidth / 2);
  int64_t y1 = 0;
  int64_t x2 = points[2] % (captchaWidth / 2) + (captchaWidth / 2);
  int64_t y2 = captchaHeight - 1;
  for (int64_t x = 0; x < static_cast<int64_t>(captchaWidth); x++) {
    for (int64_t y = 0; y < static_cast<int64_t>(captchaHeight); y++) {
      if (y > y1 + ((y2 - y1) * (x - x1)) / (x2 - x1)) {
        buffer[y * captchaWidth + x] ^= 0xFF;
      }
    }
  }
  media::JPEG image = compressor.compress(buffer, captchaWidth, 0, captchaHeight, 1, true);
  captchas.emplace(toLower(solutionString), Captcha{time(nullptr), std::move(image)});
}

void CaptchaSystem::update() {
  // Generate will not increase the size of captchas if it generates a duplicate
  while (captchas.size() < cacheSize) {
    generate();
  }
  // Remove old captchas to limit stockpiling solutions
  for (auto it = captchas.begin(); it != captchas.end();) {
    if (it->second.date + captchaMaxAge < time(nullptr)) {
      oldCaptchas.emplace(std::move(*it));
      it = captchas.erase(it);
    } else {
      it++;
    }
  }
  for (auto it = oldCaptchas.begin(); it != oldCaptchas.end();) {
    if (it->second.date + captchaDeathAge < time(nullptr)) {
      it = oldCaptchas.erase(it);
    } else {
      it++;
    }
  }
}

std::span<uint8_t> CaptchaSystem::image() {
  if (captchas.empty()) {
    generate();
  }
  uint64_t random;
  getrandom(&random, sizeof(random), 0);
  return std::next(captchas.begin(), random % captchas.size())->second.image.buffer;
}

void CaptchaSystem::setHard(bool hard) {
  if (this->hard == hard) {
    return;
  }
  this->hard = hard;
  // Get rid of captchas using the old difficulty
  captchas.clear();
  oldCaptchas.clear();
}

bool CaptchaSystem::validate(const std::string &attempt) {
  std::string lowerAttempt = toLower(attempt);
  auto captcha = captchas.find(lowerAttempt);
  if (captcha != captchas.end()) {
    captchas.erase(captcha);
    return true;
  }
  captcha = oldCaptchas.find(lowerAttempt);
  if (captcha != oldCaptchas.end()) {
    oldCaptchas.erase(captcha);
    return true;
  }
  return false;
}

}  // namespace imageboard
