#ifndef IMAGEBOARD_CAPTCHA_SYSTEM_H
#define IMAGEBOARD_CAPTCHA_SYSTEM_H

#include <turbojpeg.h>

#include <span>
#include <string>
#include <unordered_map>
#include <vector>

#include "../font/font_glyph.h"
#include "../font/freetype.h"
#include "../media/jpeg_compressor.h"

namespace imageboard {

/**
 * @brief A captcha challenge
 */
struct Captcha {
  /// The date the captcha was created
  time_t date;
  /// A JEPG encoded image of the challenge
  media::JPEG image;
};

/**
 * @brief A system for managing active captchas
 */
class CaptchaSystem {
 private:
  font::Freetype freetype;
  /**
   * @brief A handle for the TurboJPEG library
   */
  media::JPEGCompressor compressor;
  bool hard = false;
  /**
   * @brief A cache of glyphs for characters used in easy captchas
   */
  std::vector<font::FontGlyph> easyGlyphs;
  /**
   * @brief A cache of glyphs for characters used in hard captchas
   */
  std::vector<font::FontGlyph> hardGlyphs;
  /**
   * @brief A cache of captcha solution and challenges
   */
  std::unordered_map<std::string, Captcha> captchas;
  /**
   * @brief A cache of old captcha solution and challenges
   */
  std::unordered_map<std::string, Captcha> oldCaptchas;
  /**
   * @brief Generates a captcha challenge
   *
   * @warning If a duplicate captcha is generated the captcha cache will not
   * grow
   */
  void generate();
  void updateGlyphCache();

 public:
  CaptchaSystem();
  /**
   * @brief Destructor
   */
  ~CaptchaSystem() = default;
  CaptchaSystem(const CaptchaSystem& other) = delete;
  CaptchaSystem& operator=(const CaptchaSystem& other) = delete;
  /**
   * @brief Move constructor
   *
   * @param other Other request
   */
  CaptchaSystem(CaptchaSystem&& other) noexcept;
  /**
   * @brief Move assignment
   *
   * @param other Other request
   */
  CaptchaSystem& operator=(CaptchaSystem&& other) noexcept;
  /**
   * @brief Updates the captcha cache
   */
  void update();
  /**
   * @brief Get an image of a captcha challenge
   */
  std::span<uint8_t> image();
  void setHard(bool hard);
  /**
   * @brief Check if a captcha solution exists in the cache
   *
   * @param attempt A potential captcha solution
   * @return If the solution was valid
   */
  bool validate(const std::string& attempt);
};

}  // namespace imageboard

#endif
