#include "imageboard.h"

#include <absl/strings/str_format.h>
#include <sys/random.h>
#include <sys/stat.h>

#include <ctime>
#include <iostream>
#include <regex>
#include <unordered_map>
#include <unordered_set>

#include "absl/container/flat_hash_map.h"

namespace imageboard {

ImageBoard::ImageBoard(const char *database) : database(std::make_unique<database::SqliteDatabase>(database)) {}

static std::string formatSize(const std::string &filename) {
  struct stat statBuffer;
  if (stat(filename.c_str(), &statBuffer) == -1) {
    return "";
  }
  auto size = static_cast<float>(statBuffer.st_size);
  std::string suffix;
  if (statBuffer.st_size >= 1024 * 1024) {
    size /= 1024 * 1024;
    suffix = "MiB";
  } else if (statBuffer.st_size >= 1024) {
    size /= 1024;
    suffix = "KiB";
  } else {
    suffix = "B";
  }
  return absl::StrFormat("%.2f %s", size, suffix);
}

std::string escapeHTML(std::string_view string) {
  std::string escaped;
  for (char c : string) {
    switch (c) {
      case '&':
        absl::StrAppend(&escaped, "&amp;");
        break;
      case '<':
        absl::StrAppend(&escaped, "&lt;");
        break;
      case '>':
        absl::StrAppend(&escaped, "&gt;");
        break;
      case '"':
        absl::StrAppend(&escaped, "&quot;");
        break;
      case '\'':
        absl::StrAppend(&escaped, "&#39;");
        break;
      default:
        escaped += c;
        break;
    }
  }
  return escaped;
}

// TODO: Come back when std::regex::multiline is added
const std::regex newLineRegex("\r*\n", std::regex::optimize);
const std::regex quoteRegex("&gt;&gt;(\\d+)", std::regex::optimize);
const std::regex boardReferenceRegex("&gt;&gt;&gt;/([\\d[:lower:]]+?)/($|\\D)", std::regex::optimize);
const std::regex postReferenceRegex("&gt;&gt;&gt;/([\\d[:lower:]]+?)/(\\d+)", std::regex::optimize);
const std::regex greenTextRegex("(^|\r*\n)(&gt;.*)", std::regex::optimize);
const std::regex pinkTextRegex("(^|\r*\n)(&lt;.*)", std::regex::optimize);
const std::regex kikeTextRegex("\\(\\(\\(.*?\\)\\)\\)", std::regex::optimize);
const std::regex redTextRegex("==(.*?)==", std::regex::optimize);
const std::regex spoilerTextRegex("\\*\\*(.*?)\\*\\*", std::regex::optimize);
const std::regex strikethroughTextRegex("~~(.*?)~~", std::regex::optimize);
const std::regex underlineTextRegex("__(.*?)__", std::regex::optimize);
const std::regex boldTextRegex("&#39;&#39;&#39;(.*?)&#39;&#39;&#39;", std::regex::optimize);
const std::regex italicsTextRegex("&#39;&#39;(.*?)&#39;&#39;", std::regex::optimize);
const std::regex urlRegex("[a-zA-Z][a-zA-Z0-9+-.]*://[a-zA-Z0-9-._~%!$&'()*+,;=:@/?#]+", std::regex::optimize);

std::string ImageBoard::formatComment(const std::string &comment, const bool style) const {
  if (!style) {
    return std::regex_replace(comment, newLineRegex, "<br>");
  }
  std::string formatted = std::regex_replace(comment, quoteRegex, "<a class='reference' href='#post$1'>$&</a>");
  formatted =
      std::regex_replace(formatted, boardReferenceRegex, "<a class='reference' href='/$1'>&gt;&gt;&gt;/$1/</a>$2");
  std::string news;
  auto it = formatted.cbegin();
  std::smatch match;
  while (std::regex_search(it, formatted.cend(), match, postReferenceRegex)) {
    it = match.suffix().first;
    absl::StrAppend(&news, match.prefix().str());
    std::string board = match.str(1);
    std::string number = match.str(2);
    std::optional<int64_t> parentNumber = parent(board, std::stoi(number));
    if (!parentNumber) {
      absl::StrAppend(&news, match.str(0));
      continue;
    }
    if (!*parentNumber) {
      absl::StrAppend(&news, "<a class='reference' href='/", board, "/", number, "'>&gt;&gt;&gt;/", board, "/", number,
                      "</a>");
    } else {
      absl::StrAppend(&news, "<a class='reference' href='/", board, "/", *parentNumber, "#post", number,
                      "'>&gt;&gt;&gt;/", board, "/", number, "</a>");
    }
  }
  if (!news.empty()) {
    formatted = absl::StrCat(news, formatted.substr(formatted.length() - match.position()));
  }
  formatted = std::regex_replace(formatted, greenTextRegex, "$1<span class='greentext'>$2</span>");
  formatted = std::regex_replace(formatted, pinkTextRegex, "$1<span class='pinktext'>$2</span>");
  formatted = std::regex_replace(formatted, kikeTextRegex, "<span class='kiketext'>$&</span>");
  formatted = std::regex_replace(formatted, redTextRegex, "<span class='redtext'>$1</span>");
  formatted = std::regex_replace(formatted, spoilerTextRegex, "<span class='spoiler'>$1</span>");
  formatted = std::regex_replace(formatted, strikethroughTextRegex, "<s>$1</s>");
  formatted = std::regex_replace(formatted, underlineTextRegex, "<u>$1</u>");
  formatted = std::regex_replace(formatted, boldTextRegex, "<b>$1</b>");
  formatted = std::regex_replace(formatted, italicsTextRegex, "<i>$1</i>");
  formatted = std::regex_replace(formatted, urlRegex, "<a rel='noreferrer' href='$&'>$&</a>");
  return std::regex_replace(formatted, newLineRegex, "<br>");
}

void ImageBoard::addFile(const std::string_view filename, const int64_t thumbWidth, int64_t thumbHeight) const {
  auto statement = database->prepareStatement(
      "INSERT INTO File (Name, ThumbWidth, ThumbHeight) "
      "VALUES (?, ?, ?);");
  statement->bind(1, filename);
  statement->bind(2, thumbWidth);
  statement->bind(3, thumbHeight);
  statement->step();
}

std::optional<int64_t> ImageBoard::addPost(const Board &board, const int64_t parent,
                                           const std::string_view nameUnescaped, const std::string_view emailUnescaped,
                                           const std::string_view subjectUnescaped,
                                           const std::string_view commentUnescaped, const std::string_view file,
                                           const int64_t bumpLimit, const int64_t replies) const {
  time_t date = time(nullptr);
  std::string name = escapeHTML(nameUnescaped);
  std::string email = escapeHTML(emailUnescaped);
  std::string subject = escapeHTML(subjectUnescaped);
  std::string comment = escapeHTML(commentUnescaped);
  bool sage = email == "sage";
  if (name == "") {
    name = "Nanonymous";
  }
  std::unordered_set<int64_t> quotes;
  std::string::const_iterator it = comment.cbegin();
  std::smatch match;
  while (std::regex_search(it, comment.cend(), match, quoteRegex)) {
    it = match.suffix().first;
    quotes.emplace(std::stoi(match[1]));
  }
  comment = formatComment(comment, email != "nofo" && email != "noformatting");
  int64_t number;
  try {
    database->transactionBegin();
    auto incrementStatement = database->prepareStatement(
        "UPDATE Boards "
        "SET MaxPostNumber = MaxPostNumber + 1 "
        "WHERE Name = ?;");
    incrementStatement->bind(1, board.name);
    incrementStatement->step();
    auto numberStatement = database->prepareStatement(
        "SELECT MaxPostNumber "
        "FROM Boards "
        "WHERE Name = ?;");
    numberStatement->bind(1, board.name);
    if (numberStatement->step() != database::QueryResult::row) {
      database->transactionRollback();
      return std::nullopt;
    }
    number = numberStatement->getInt(0);
    database->transactionEnd();
  } catch (const std::runtime_error &exception) {
    database->transactionRollback();
    throw;
  }
  auto postStatement = database->prepareStatement(
      "INSERT INTO Posts (Board, Number, Parent, Date, LastBumpDate, Name, "
      "Email, Subject, Comment, File, Sticky, Cycle, Autosage, Lock, "
      "tvolDeleteName, tvolDeleteDate) "
      "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, 0, 0, ?, 0, '', 0);");
  postStatement->bind(1, board.name);
  postStatement->bind(2, number);
  postStatement->bind(3, parent);
  postStatement->bind(4, date);
  postStatement->bind(5, date);
  postStatement->bind(6, name);
  postStatement->bind(7, email);
  postStatement->bind(8, subject);
  postStatement->bind(9, comment);
  postStatement->bind(10, file);
  postStatement->bind(11, sage);
  postStatement->step();
  for (int64_t quote : quotes) {
    auto referenceStatement = database->prepareStatement(
        "INSERT INTO Refs (Board, Referee, Referrer) "
        "VALUES (?, ?, ?);");
    referenceStatement->bind(1, board.name);
    referenceStatement->bind(2, quote);
    referenceStatement->bind(3, number);
    referenceStatement->step();
  }
  if (parent) {
    std::optional<Thread> parentThread = thread(board, parent);
    if (!parentThread) {
      return std::nullopt;
    }
    if ((!parentThread->autosage && !sage) && (parentThread->cycle || replies < bumpLimit)) {
      auto bumpStatement = database->prepareStatement(
          "UPDATE Posts "
          "SET LastBumpDate = ?"
          "WHERE Board = ? AND Number = ?;");
      bumpStatement->bind(1, date);
      bumpStatement->bind(2, board.name);
      bumpStatement->bind(3, parent);
      bumpStatement->step();
    }
  }
  if (board.captchaTriggerPPH > 0 && !board.captcha) {
    std::optional<int64_t> posts = pph(board.name);
    if (*posts > board.captchaTriggerPPH) {
      setCaptcha(board.name, true);
      createLog(std::nullopt, board.name, "Automatically enabled captcha due to excessive PPH");
    }
  }
  return number;
}

std::string ImageBoard::announcement() const {
  auto statement = database->prepareStatement(
      "SELECT Value "
      "FROM Global "
      "WHERE Name = 'announce';");
  if (statement->step() != database::QueryResult::row) {
    throw std::runtime_error("An announcement has not been set");
  }
  return statement->getString(0);
}

std::string ImageBoard::banner(std::string_view board) const {
  auto statement = database->prepareStatement(
      "SELECT File "
      "FROM Banners "
      "Where Board = ?;");
  statement->bind(1, board);
  std::vector<std::string> banners;
  while (statement->step() == database::QueryResult::row) {
    banners.emplace_back(statement->getString(0));
  }
  if (banners.empty()) {
    return "";
  }
  uint64_t random;
  getrandom(&random, sizeof(random), 0);
  return banners.at(random % banners.size());
}

std::optional<Board> ImageBoard::board(const std::string_view board) const {
  auto statement = database->prepareStatement(
      "SELECT Title, Subtitle, Lock, DisplayOverboard, MaxThreadsPerHour, "
      "MinThreadChars, BumpLimit, PostLimit, ThreadLimit, RequireCaptcha, "
      "CaptchaTriggerPPH "
      "FROM Boards "
      "WHERE Name = ?;");
  statement->bind(1, board);
  if (statement->step() != database::QueryResult::row) {
    return std::nullopt;
  }
  std::string title = statement->getString(0);
  std::string subtitle = statement->getString(1);
  bool lock = statement->getBool(2);
  bool displayOverboard = statement->getBool(3);
  int64_t maximumTPD = statement->getInt(4);
  int64_t minimumThreadCharacters = statement->getInt(5);
  int64_t bumpLimit = statement->getInt(6);
  int64_t postLimit = statement->getInt(7);
  int64_t threadLimit = statement->getInt(8);
  bool requireCaptcha = statement->getBool(9);
  int64_t captchaTriggerPPH = statement->getInt(10);
  Board b = {std::string(board),      title,     subtitle,  lock,        displayOverboard, maximumTPD,
             minimumThreadCharacters, bumpLimit, postLimit, threadLimit, requireCaptcha,   captchaTriggerPPH};
  return b;
}

std::vector<Board> ImageBoard::boards() const {
  std::vector<Board> boards;
  auto statement = database->prepareStatement(
      "SELECT Name, Title, Subtitle, Lock, DisplayOverboard, "
      "MaxThreadsPerHour, MinThreadChars, BumpLimit, PostLimit, "
      "ThreadLimit, RequireCaptcha, CaptchaTriggerPPH "
      "FROM Boards "
      "ORDER BY MaxPostNumber DESC;");
  while (statement->step() == database::QueryResult::row) {
    std::string name = statement->getString(0);
    std::string title = statement->getString(1);
    std::string subtitle = statement->getString(2);
    bool lock = statement->getBool(3);
    bool displayOverboard = statement->getBool(4);
    int64_t maximumTPD = statement->getInt(5);
    int64_t minimumThreadCharacters = statement->getInt(6);
    int64_t bumpLimit = statement->getInt(7);
    int64_t postLimit = statement->getInt(8);
    int64_t threadLimit = statement->getInt(9);
    bool requireCaptcha = statement->getBool(10);
    int64_t captchaTriggerPPH = statement->getInt(11);
    boards.push_back({name, title, subtitle, lock, displayOverboard, maximumTPD, minimumThreadCharacters, bumpLimit,
                      postLimit, threadLimit, requireCaptcha, captchaTriggerPPH});
  }
  return boards;
}

void ImageBoard::createLog(const std::optional<std::string> account, const std::optional<std::string> boardRaw,
                           const std::string &commentRaw) const {
  time_t date = time(nullptr);
  std::string name = "<i>System</i>";
  if (account) {
    name = *account;
  }
  std::string board = "<i>Global</i>";
  if (boardRaw) {
    board = absl::StrCat("<a href='/", *boardRaw, "'>/", *boardRaw, "/</a>");
  }
  std::string comment = formatComment(commentRaw, true);
  auto statement = database->prepareStatement(
      "INSERT INTO Logs (Name, Board, Date, Description) "
      "VALUES (?, ?, ?, ?);");
  statement->bind(1, name);
  statement->bind(2, board);
  statement->bind(3, date);
  statement->bind(4, comment);
  statement->step();
}

bool ImageBoard::cycle(const std::string_view board, const int64_t number) const {
  auto statement = database->prepareStatement(
      "SELECT Cycle "
      "FROM Posts "
      "WHERE Board = ? AND Number = ?;");
  statement->bind(1, board);
  statement->bind(2, number);
  if (statement->step() != database::QueryResult::row) {
    throw std::runtime_error("Could not check if cyclical");
  }
  return statement->getBool(0);
}

bool ImageBoard::flag(const std::string_view flag) const {
  auto statement = database->prepareStatement(
      "SELECT Value "
      "FROM Global "
      "WHERE Name = ?;");
  statement->bind(1, flag);
  if (statement->step() != database::QueryResult::row) {
    throw std::runtime_error("The flag has not been set");
  }
  return statement->getBool(0);
}

std::vector<std::vector<uint8_t>> ImageBoard::loadBanned() const {
  std::vector<std::vector<uint8_t>> banned;
  auto statement = database->prepareStatement(
      "SELECT Hash "
      "FROM Banned;");
  while (statement->step() == database::QueryResult::row) {
    std::span<const uint8_t> hash = statement->getBlob(0);
    banned.emplace_back(hash.data(), hash.data() + hash.size());
  }
  return banned;
}

std::vector<LogEntry> ImageBoard::logs(const int64_t page) const {
  std::vector<LogEntry> entries;
  auto statement = database->prepareStatement(
      "SELECT Name, Board, Date, Description "
      "FROM Logs "
      "ORDER BY Date DESC "
      "LIMIT 128 OFFSET ?;");
  statement->bind(1, (page - 1) * 128);
  while (statement->step() == database::QueryResult::row) {
    std::string account = statement->getString(0);
    std::string board = statement->getString(1);
    time_t timestamp = statement->getInt(2);
    std::string description = statement->getString(3);
    std::string time;
    time.resize(32);
    time.resize(strftime(time.data(), 32, "%F %T", gmtime(&timestamp)));
    entries.push_back({account, board, time, description});
  }
  return entries;
}

std::optional<int64_t> ImageBoard::parent(const std::string_view board, const int64_t number) const {
  auto statement = database->prepareStatement(
      "SELECT Parent "
      "FROM Posts "
      "WHERE Board = ? AND Number = ?;");
  statement->bind(1, board);
  statement->bind(2, number);
  if (statement->step() != database::QueryResult::row) {
    return std::nullopt;
  }
  return statement->getInt(0);
}

std::vector<Post> ImageBoard::posts(const std::string_view board, const int64_t parent, const int64_t postLimit) const {
  std::vector<Post> posts;
  std::unordered_map<int64_t, std::optional<File>> files;
  std::unordered_map<int64_t, std::vector<std::string>> references;
  auto filesStatement = database->prepareStatement(
      "SELECT p.ROWID, File, ThumbWidth, ThumbHeight "
      "FROM (SELECT ROWID, File FROM Posts WHERE Board = ? AND "
      "(Number = ? OR Parent = ?) AND File <> '') p "
      "LEFT JOIN File f "
      "ON File = f.Name;");
  filesStatement->bind(1, board);
  filesStatement->bind(2, parent);
  filesStatement->bind(3, parent);
  while (filesStatement->step() == database::QueryResult::row) {
    int64_t rowId = filesStatement->getInt(0);
    std::string filename = filesStatement->getString(1);
    int64_t thumbWidth = filesStatement->getInt(2);
    int64_t thumbHeight = filesStatement->getInt(3);
    files.emplace(rowId, File{filename, thumbWidth, thumbHeight, formatSize(absl::StrCat(".media/", filename))});
  }
  auto referencesStatement = database->prepareStatement(
      "SELECT p.ROWID, Referrer "
      "FROM (SELECT ROWID, Board, Number FROM Posts WHERE "
      "Board = ? AND (Number = ? OR Parent = ?)) p "
      "INNER JOIN Refs r "
      "ON p.Board = r.Board AND Number = Referee;");
  referencesStatement->bind(1, board);
  referencesStatement->bind(2, parent);
  referencesStatement->bind(3, parent);
  while (referencesStatement->step() == database::QueryResult::row) {
    int64_t rowId = referencesStatement->getInt(0);
    int64_t referrer = referencesStatement->getInt(1);
    references[rowId].emplace_back(std::to_string(referrer));
  }
  auto postsStatement = database->prepareStatement(
      "SELECT * FROM (SELECT ROWID, Number, Parent, Date, Name, Email, "
      "Subject, Comment, Sticky, Lock, Autosage, Cycle "
      "FROM Posts "
      "WHERE Board = ? AND Parent = ? AND tvolDeleteName = '' "
      "ORDER BY Date DESC "
      "LIMIT ?) "
      "UNION "
      "SELECT ROWID, Number, Parent, Date, Name, Email, Subject, Comment, "
      "Sticky, Lock, Autosage, Cycle "
      "FROM Posts "
      "WHERE Board = ? AND Number = ? AND tvolDeleteName = '' "
      "ORDER BY Date;");
  postsStatement->bind(1, board);
  postsStatement->bind(2, parent);
  postsStatement->bind(3, postLimit);
  postsStatement->bind(4, board);
  postsStatement->bind(5, parent);
  while (postsStatement->step() == database::QueryResult::row) {
    int64_t rowId = postsStatement->getInt(0);
    int64_t number = postsStatement->getInt(1);
    int64_t parent = postsStatement->getInt(2);
    time_t timestamp = postsStatement->getInt(3);
    std::string name = postsStatement->getString(4);
    std::string email = postsStatement->getString(5);
    std::string subject = postsStatement->getString(6);
    std::string comment = postsStatement->getString(7);
    bool sticky = postsStatement->getBool(8);
    bool lock = postsStatement->getBool(9);
    bool autosage = postsStatement->getBool(10);
    bool cycle = postsStatement->getBool(11);
    std::string time;
    time.resize(32);
    time.resize(strftime(time.data(), 32, "%F %T", gmtime(&timestamp)));
    std::string flags;
    if (sticky) {
      absl::StrAppend(&flags, "(S)");
    }
    if (lock) {
      absl::StrAppend(&flags, "(L)");
    }
    if (autosage) {
      absl::StrAppend(&flags, "(A)");
    }
    if (cycle) {
      absl::StrAppend(&flags, "(C)");
    }
    posts.push_back({std::string(board), number, parent, time, name, email, subject, comment, flags,
                     std::move(files[rowId]), std::move(references[rowId])});
  }
  return posts;
}

int64_t ImageBoard::pph(const std::string_view board) const {
  auto statement = database->prepareStatement(
      "SELECT COUNT(*) "
      "FROM Posts "
      "WHERE Date > CAST(STRFTIME('%s', 'now', '-1 hour') AS "
      "INTEGER) AND Board = ?;");
  statement->bind(1, board);
  if (statement->step() != database::QueryResult::row) {
    throw std::runtime_error("Could not calculate pph");
  }
  return statement->getInt(0);
}

std::vector<Post> ImageBoard::recent(const int64_t page, const bool sage, const bool file, const bool parent,
                                     const bool customName, const bool tvolDelete, const int64_t postLimit) const {
  std::vector<Post> posts;
  std::unordered_map<int64_t, std::optional<File>> files;
  std::unordered_map<int64_t, std::vector<std::string>> references;
  std::string filter = "WHERE ";
  if (sage) {
    absl::StrAppend(&filter, "Email = 'sage' AND ");
  }
  if (file) {
    absl::StrAppend(&filter, "File <> '' AND ");
  }
  if (parent) {
    absl::StrAppend(&filter, "Parent = 0 AND ");
  }
  if (customName) {
    absl::StrAppend(&filter, "Name <> 'Nanonymous' AND ");
  }
  if (tvolDelete) {
    absl::StrAppend(&filter, "tvolDeleteName <> ''");
  } else {
    absl::StrAppend(&filter, "tvolDeleteName = ''");
  }
  std::string sql1 = absl::StrCat(
      "SELECT p.ROWID, File, ThumbWidth, ThumbHeight "
      "FROM (SELECT ROWID, File FROM Posts ",
      filter,
      " ORDER BY Date DESC LIMIT ? OFFSET ?) p "
      "LEFT JOIN File f "
      "ON File = f.Name "
      "WHERE File <> '';");
  auto filesStatement = database->prepareStatement(sql1);
  filesStatement->bind(1, postLimit);
  filesStatement->bind(2, (page - 1) * postLimit);
  while (filesStatement->step() == database::QueryResult::row) {
    int64_t rowId = filesStatement->getInt(0);
    std::string filename = filesStatement->getString(1);
    int64_t thumbWidth = filesStatement->getInt(2);
    int64_t thumbHeight = filesStatement->getInt(3);
    files.emplace(rowId, File{filename, thumbWidth, thumbHeight, formatSize(absl::StrCat(".media/", filename))});
  }
  std::string sql2 = absl::StrCat(
      "SELECT p.ROWID, Referrer "
      "FROM (SELECT ROWID, Board, Number FROM Posts ",
      filter,
      " ORDER BY Date DESC LIMIT ? OFFSET ?) p "
      "INNER JOIN Refs r "
      "ON p.Board = r.Board AND Number = Referee;");
  auto referencesStatement = database->prepareStatement(sql2);
  referencesStatement->bind(1, postLimit);
  referencesStatement->bind(2, (page - 1) * postLimit);
  while (referencesStatement->step() == database::QueryResult::row) {
    int64_t rowId = referencesStatement->getInt(0);
    int64_t referrer = referencesStatement->getInt(1);
    references[rowId].emplace_back(std::to_string(referrer));
  }
  std::string sql3 = absl::StrCat(
      "SELECT ROWID, Board, Number, Parent, Date, Name, Email, Subject, "
      "Comment, Sticky, Lock, Autosage, Cycle "
      "FROM Posts ",
      filter,
      " ORDER BY Date DESC "
      "LIMIT ? OFFSET ?;");
  auto postsStatement = database->prepareStatement(sql3);
  postsStatement->bind(1, postLimit);
  postsStatement->bind(2, (page - 1) * postLimit);
  while (postsStatement->step() == database::QueryResult::row) {
    int64_t rowId = postsStatement->getInt(0);
    std::string board = postsStatement->getString(1);
    int64_t number = postsStatement->getInt(2);
    int64_t parent = postsStatement->getInt(3);
    time_t timestamp = postsStatement->getInt(4);
    std::string name = postsStatement->getString(5);
    std::string email = postsStatement->getString(6);
    std::string subject = postsStatement->getString(7);
    std::string comment = postsStatement->getString(8);
    bool sticky = postsStatement->getBool(9);
    bool lock = postsStatement->getBool(10);
    bool autosage = postsStatement->getBool(11);
    bool cycle = postsStatement->getBool(12);
    std::string time;
    time.resize(32);
    time.resize(strftime(time.data(), 32, "%F %T", gmtime(&timestamp)));
    std::string flags;
    if (sticky) {
      absl::StrAppend(&flags, "(S)");
    }
    if (lock) {
      absl::StrAppend(&flags, "(L)");
    }
    if (autosage) {
      absl::StrAppend(&flags, "(A)");
    }
    if (cycle) {
      absl::StrAppend(&flags, "(C)");
    }
    posts.push_back({board, number, parent, time, name, email, subject, comment, flags, std::move(files[rowId]),
                     std::move(references[rowId])});
  }
  return posts;
}

void ImageBoard::setCaptcha(const std::string_view board, const bool requireCaptcha) const {
  auto statement = database->prepareStatement(
      "UPDATE Boards "
      "SET RequireCaptcha = ? "
      "WHERE Name = ?;");
  statement->bind(1, requireCaptcha);
  statement->bind(2, board);
  statement->step();
}

std::vector<StatsEntry> ImageBoard::stats() const {
  std::vector<StatsEntry> stats;
  auto statement = database->prepareStatement(
      "SELECT b.Name, "
      "       (SELECT COUNT(*) FROM Posts WHERE Date > CAST(STRFTIME('%s', "
      "'now', '-1 day') AS INTEGER) AND Parent = 0 AND Board = b.Name), "
      "       (SELECT COUNT(*) FROM Posts WHERE Date > CAST(STRFTIME('%s', "
      "'now', '-7 days') AS INTEGER) AND Parent = 0 AND Board = b.Name), "
      "       (SELECT COUNT(*) FROM Posts WHERE Date > CAST(STRFTIME('%s', "
      "'now', '-1 hour') AS INTEGER) AND Board = b.Name), "
      "       (SELECT COUNT(*) FROM Posts WHERE Date > CAST(STRFTIME('%s', "
      "'now', '-1 day') AS INTEGER) AND Board = b.Name), "
      "       (SELECT COUNT(*) FROM Posts WHERE Date > CAST(STRFTIME('%s', "
      "'now', '-7 days') AS INTEGER) AND Board = b.Name), "
      "       b.MaxPostNumber "
      "FROM Boards b "
      "ORDER BY MaxPostNumber DESC;");
  int64_t totalTPD = 0;
  int64_t totalTPW = 0;
  int64_t totalPPH = 0;
  int64_t totalPPD = 0;
  int64_t totalPPW = 0;
  int64_t totalPosts = 0;
  while (statement->step() == database::QueryResult::row) {
    std::string name = statement->getString(0);
    int64_t tpd = statement->getInt(1);
    int64_t tpw = statement->getInt(2);
    int64_t pph = statement->getInt(3);
    int64_t ppd = statement->getInt(4);
    int64_t ppw = statement->getInt(5);
    int64_t posts = statement->getInt(6);
    std::string ppdh = absl::StrFormat("%.1f", static_cast<float>(ppd) / 24);
    std::string ppwd = absl::StrFormat("%.1f", static_cast<float>(ppw) / 7);
    stats.push_back({name, std::to_string(tpd), std::to_string(tpw), std::to_string(pph), ppdh, std::to_string(ppd),
                     ppwd, std::to_string(posts)});
    totalTPD += tpd;
    totalTPW += tpw;
    totalPPH += pph;
    totalPPD += ppd;
    totalPPW += ppw;
    totalPosts += posts;
  }
  std::string totalPPDH = absl::StrFormat("%.1f", static_cast<float>(totalPPD) / 24);
  std::string totalPPWD = absl::StrFormat("%.1f", static_cast<float>(totalPPW) / 7);
  stats.push_back({"total", std::to_string(totalTPD), std::to_string(totalTPW), std::to_string(totalPPH), totalPPDH,
                   std::to_string(totalPPD), totalPPWD, std::to_string(totalPosts)});
  return stats;
}

std::optional<Thread> ImageBoard::thread(const Board &board, const int64_t number) const {
  auto repliesStatement = database->prepareStatement(
      "SELECT COUNT(*) "
      "FROM Posts "
      "WHERE Board = ? AND Parent = ?"
      "GROUP BY Parent;");
  repliesStatement->bind(1, board.name);
  repliesStatement->bind(2, number);
  int64_t replies = 0;
  if (repliesStatement->step() == database::QueryResult::row) {
    replies = repliesStatement->getInt(0);
  }
  auto fileStatement = database->prepareStatement(
      "SELECT File, ThumbWidth, ThumbHeight "
      "FROM Posts "
      "LEFT JOIN File f "
      "ON File = f.Name "
      "WHERE Board = ? AND Number = ? AND File <> '';");
  fileStatement->bind(1, board.name);
  fileStatement->bind(2, number);
  std::optional<File> file;
  if (fileStatement->step() == database::QueryResult::row) {
    std::string filename = fileStatement->getString(0);
    int64_t thumbWidth = fileStatement->getInt(1);
    int64_t thumbHeight = fileStatement->getInt(2);
    file = {filename, thumbWidth, thumbHeight, formatSize(absl::StrCat(".media/", filename))};
  }
  auto threadStatement = database->prepareStatement(
      "SELECT LastBumpDate, Subject, Comment, Sticky, Lock, "
      "Autosage, Cycle "
      "FROM Posts "
      "WHERE Board = ? AND Number = ? AND Parent = 0;");
  threadStatement->bind(1, board.name);
  threadStatement->bind(2, number);
  if (threadStatement->step() != database::QueryResult::row) {
    return std::nullopt;
  }
  time_t lastBumpTimestamp = threadStatement->getInt(0);
  std::string subject = threadStatement->getString(1);
  std::string comment = threadStatement->getString(2);
  bool sticky = threadStatement->getBool(3);
  bool lock = threadStatement->getBool(4);
  bool autosage = threadStatement->getBool(5);
  bool cycle = threadStatement->getBool(6);
  std::string lastBumpTime;
  lastBumpTime.resize(32);
  lastBumpTime.resize(strftime(lastBumpTime.data(), 32, "%F %T", gmtime(&lastBumpTimestamp)));
  std::string flags;
  if (sticky) {
    absl::StrAppend(&flags, "(S)");
  }
  if (lock) {
    absl::StrAppend(&flags, "(L)");
  }
  if (autosage) {
    absl::StrAppend(&flags, "(A)");
  }
  if (cycle) {
    absl::StrAppend(&flags, "(C)");
  }
  Thread thread = {board.name,
                   std::to_string(number),
                   lastBumpTime,
                   subject,
                   comment,
                   file,
                   lock,
                   cycle,
                   autosage,
                   flags,
                   cycle ? std::min(replies, board.postLimit) : replies};
  return thread;
}

std::vector<Thread> ImageBoard::threads(const std::unordered_map<std::string, Board> &boards, const bool sticky,
                                        const bool autosage, const int64_t threadLimit) const {
  std::vector<Thread> threads;
  absl::flat_hash_map<int64_t, std::optional<File>> files;
  absl::flat_hash_map<std::pair<std::string, int64_t>, int64_t> threadReplies;
  std::string sql1 =
      "SELECT Board, Parent, COUNT(*) "
      "FROM Posts ";
  if (!boards.empty()) {
    absl::StrAppend(&sql1, "WHERE Board IN (?");
    for (std::vector<const std::string *>::size_type i = 1; i < boards.size(); i++) {
      absl::StrAppend(&sql1, ", ?");
    }
    absl::StrAppend(&sql1, ") ");
  }
  absl::StrAppend(&sql1, "GROUP BY Board, Parent;");
  auto repliesStatement = database->prepareStatement(sql1);
  auto it = boards.begin();
  for (std::unordered_map<std::string, Board>::size_type i = 0; i < boards.size(); i++) {
    repliesStatement->bind(i + 1, it->first);
    it++;
  }
  while (repliesStatement->step() == database::QueryResult::row) {
    std::string board = repliesStatement->getString(0);
    int64_t parent = repliesStatement->getInt(1);
    int64_t count = repliesStatement->getInt(2);
    threadReplies.try_emplace(std::make_pair(board, parent), count);
  }
  std::string sql2 =
      "SELECT p.ROWID, File, ThumbWidth, ThumbHeight "
      "FROM (SELECT ROWID, File FROM Posts WHERE ";
  if (!boards.empty()) {
    absl::StrAppend(&sql2, "Board IN (?");
    for (std::unordered_map<std::string, Board>::size_type i = 1; i < boards.size(); i++) {
      absl::StrAppend(&sql2, ", ?");
    }
    absl::StrAppend(&sql2, ") AND ");
  }
  absl::StrAppend(&sql2, "Parent = 0 ");
  if (!autosage) {
    absl::StrAppend(&sql2, " AND Autosage = 0 ");
  }
  absl::StrAppend(&sql2, "AND File <> '' ORDER BY ");
  if (sticky) {
    absl::StrAppend(&sql2, "Sticky DESC, ");
  }
  absl::StrAppend(&sql2,
                  "LastBumpDate DESC LIMIT ?) p "
                  "LEFT JOIN File f "
                  "ON File = f.Name;");
  auto filesStatement = database->prepareStatement(sql2);
  it = boards.begin();
  for (std::unordered_map<std::string, Board>::size_type i = 0; i < boards.size(); i++) {
    filesStatement->bind(i + 1, it->first);
    it++;
  }
  filesStatement->bind(boards.size() + 1, threadLimit);
  while (filesStatement->step() == database::QueryResult::row) {
    int64_t rowId = filesStatement->getInt(0);
    std::string filename = filesStatement->getString(1);
    int64_t thumbWidth = filesStatement->getInt(2);
    int64_t thumbHeight = filesStatement->getInt(3);
    files.try_emplace(rowId, std::in_place, filename, thumbWidth, thumbHeight,
                      formatSize(absl::StrCat(".media/", filename)));
  }
  std::string sql3 =
      "SELECT ROWID, Board, Number, LastBumpDate, Subject, Comment, Sticky, "
      "Lock, Autosage, Cycle "
      "FROM Posts "
      "WHERE ";
  if (!boards.empty()) {
    absl::StrAppend(&sql3, "Board IN (?");
    for (std::unordered_map<std::string, Board>::size_type i = 1; i < boards.size(); i++) {
      absl::StrAppend(&sql3, ", ?");
    }
    absl::StrAppend(&sql3, ") AND ");
  }
  absl::StrAppend(&sql3, "Parent = 0 ");
  if (!autosage) {
    absl::StrAppend(&sql3, "AND Autosage = 0 ");
  }
  absl::StrAppend(&sql3, "AND tvolDeleteName = '' ORDER BY ");
  if (sticky) {
    absl::StrAppend(&sql3, "Sticky DESC, ");
  }
  absl::StrAppend(&sql3,
                  "LastBumpDate DESC "
                  "LIMIT ?;");
  auto threadsStatement = database->prepareStatement(sql3);
  it = boards.begin();
  for (std::unordered_map<std::string, Board>::size_type i = 0; i < boards.size(); i++) {
    threadsStatement->bind(i + 1, it->first);
    it++;
  }
  threadsStatement->bind(boards.size() + 1, threadLimit);
  while (threadsStatement->step() == database::QueryResult::row) {
    int64_t rowId = threadsStatement->getInt(0);
    std::string boardName = threadsStatement->getString(1);
    int64_t number = threadsStatement->getInt(2);
    time_t lastBumpTimestamp = threadsStatement->getInt(3);
    std::string subject = threadsStatement->getString(4);
    std::string comment = threadsStatement->getString(5);
    bool stickyFlag = threadsStatement->getBool(6);
    bool lockFlag = threadsStatement->getBool(7);
    bool autosageFlag = threadsStatement->getBool(8);
    bool cycleFlag = threadsStatement->getBool(9);
    std::string lastBumpTime;
    lastBumpTime.resize(32);
    lastBumpTime.resize(strftime(lastBumpTime.data(), 32, "%F %T", gmtime(&lastBumpTimestamp)));
    std::string flags;
    if (stickyFlag) {
      absl::StrAppend(&flags, "(S)");
    }
    if (lockFlag) {
      absl::StrAppend(&flags, "(L)");
    }
    if (autosageFlag) {
      absl::StrAppend(&flags, "(A)");
    }
    if (cycleFlag) {
      absl::StrAppend(&flags, "(C)");
    }
    int64_t replies = 0;
    std::optional<File> file;
    auto board = boards.find(boardName);
    if (board != boards.end()) {
      auto replyCount = threadReplies.find(std::make_pair(boardName, number));
      if (replyCount != threadReplies.end()) {
        replies = cycleFlag ? std::min(replyCount->second, board->second.postLimit) : replyCount->second;
      }
      auto fileInfo = files.find(rowId);
      if (fileInfo != files.end()) {
        file = std::move(fileInfo->second);
      }
    }
    threads.push_back({boardName, std::to_string(number), lastBumpTime, subject, comment, std::move(file), lockFlag,
                       cycleFlag, autosageFlag, flags, replies});
  }
  return threads;
}

int64_t ImageBoard::tpd(const std::string_view board) const {
  auto statement = database->prepareStatement(
      "SELECT COUNT(*) "
      "FROM Posts "
      "WHERE Date > CAST(STRFTIME('%s', 'now', '-1 day') AS "
      "INTEGER) AND Parent = 0 AND Board = ?;");
  statement->bind(1, board);
  if (statement->step() != database::QueryResult::row) {
    throw std::runtime_error("Could not calculate tpd");
  }
  return statement->getInt(0);
}

}  // namespace imageboard
