#ifndef IMAGEBOARD_IMAGEBOARD_H
#define IMAGEBOARD_IMAGEBOARD_H

#include <sqlite3.h>

#include <optional>
#include <string>
#include <unordered_map>
#include <vector>

#include "../database/database.h"
#include "../database/sqlite/sqlite_database.h"

namespace imageboard {

/**
 * @brief An image board board
 */
struct Board {
  /// The name of the board
  std::string name;
  /// The title of the board
  std::string title;
  /// The subtitle of the board
  std::string subtitle;
  /// Whether the board is locked
  bool lock;
  /// Whether this board should display on the overboard
  bool displayOverboard;
  /// The maximum TPD allowed for a board
  int64_t maximumTPD;
  /// The minimum amount of characters required for an OP
  int64_t minimumThreadCharacters;
  /// The number of posts before bumps stop working
  int64_t bumpLimit;
  /// The maximum number of posts in a thread
  int64_t postLimit;
  /// The maximum number of threads on a board
  int64_t threadLimit;
  /// Whether captcha is enabled
  bool captcha;
  /// The number of posts in an hour required to activate captcha
  int64_t captchaTriggerPPH;
};

/**
 * @brief A filename
 */
struct File {
  /// The name of the file
  std::string filename;
  /// The width of the thumbnail
  int64_t thumbWidth;
  /// The height of the thumbnail
  int64_t thumbHeight;
  /// The size of the file
  std::string size;
};

/**
 * @brief A log entry
 */
struct LogEntry {
  /// The name of who caused the log entry
  std::string name;
  /// The board this log entry applies to
  std::string board;
  /// A formatted timestamp
  std::string time;
  /// The log message
  std::string description;
};

/**
 * @brief A post
 */
struct Post {
  /// The board this post was posted to
  std::string board;
  /// The number of the post
  int64_t number;
  /// The number of the thread's OP or 0 if it is an OP
  int64_t parent;
  /// A formatted timestamp
  std::string time;
  /// The name of the poster
  std::string name;
  /// The email of the poster
  std::string email;
  /// The subject of the post
  std::string subject;
  /// The comment of the post
  std::string comment;
  /// The flags applied to the post
  std::string flags;
  /// The file attatched to the post
  std::optional<File> file;
  /// The posts referencing this post
  std::vector<std::string> references;
};

/**
 * @brief A thread
 */
struct Thread {
  /// The board this thread was posted to
  std::string board;
  /// The number of the thread
  std::string number;
  /// A formatted timestamp of the last bump
  std::string lastBumpTime;
  /// The subject of the thread
  std::string subject;
  /// The comment of the thread
  std::string comment;
  /// The file attatched to the thread
  std::optional<File> file;
  /// Whether the thread is locked
  bool lock;
  /// Whether the thread is set to cycle
  bool cycle;
  /// Whether the thread is set to autosage
  bool autosage;
  /// The flags applied to the thread
  std::string flags;
  /// The number of replies to the thread
  std::int64_t replies;
};

/**
 * @brief A Stats entry
 */
struct StatsEntry {
  /// The board the stats are for
  std::string board;
  /// The amount of threads posted within the last 24 hours
  std::string tpd;
  /// The amount of threads posted within the last week
  std::string tpw;
  /// The amount of posts posted within the last hour
  std::string pph;
  /// The amount of posts posted within the last day divided by 24
  std::string ppdh;
  /// The amount of posts posted within the last day
  std::string ppd;
  /// The amount of posts posted within the last week divided by 7
  std::string ppwd;
  /// The total amount of posts posted to this board
  std::string posts;
};

/**
 * @param An image board
 */
class ImageBoard {
 private:
  /**
   * @brief The database used for storing the image board
   */
  std::string formatComment(const std::string& comment, bool style) const;

 public:
  std::unique_ptr<database::Database> database;
  explicit ImageBoard(const char* database);
  void addFile(std::string_view filename, int64_t thumbWidth, int64_t thumbHeight) const;
  /**
   * @brief Adds a new post
   *
   * @param board The board the new post will be on
   * @param parent The thread the post belongs to or 0 if it is a new thread
   * @param nameUnescaped An unescaped name
   * @param emailUnescaped An unescaped email
   * @param subjectUnescaped An unescaped subject
   * @param commentUnescaped An unescaped comment
   * @param file The filename of the attached file or an empty string for no
   * file
   * @param bumpLimit The number of posts allowed before bumps stop working
   * @param replies The current number of replies to a thread or 0 if it is a
   * new thread
   * @return The number of the new post
   */
  std::optional<int64_t> addPost(const Board& board, int64_t parent, std::string_view nameUnescaped,
                                 std::string_view emailUnescaped, std::string_view subjectUnescaped,
                                 std::string_view commentUnescaped, std::string_view file, int64_t bumpLimit,
                                 int64_t replies) const;
  /**
   * @brief Retrieves the site wide announcement
   *
   * @return The site wide announcement
   */
  std::string announcement() const;
  /**
   * @brief Retrieves a banner for a board
   *
   * @param board The board the banners should be for
   * @return An image name or an empty string if there is not banner for a board
   */
  std::string banner(std::string_view board) const;
  /**
   * @brief Retrieves a board
   *
   * @param board The name of the board that is being retrieved
   * @return The board
   */
  std::optional<Board> board(std::string_view board) const;
  /**
   * @brief Retrieves all of the boards
   *
   * Boards are sorted by the number of posts that have been made on it.
   * @return The boards
   */
  std::vector<Board> boards() const;
  /**
   * @brief Creates a log entry
   *
   * @param account The account that created this log entry
   * @param boardRaw A board that will be turned into a link
   * @param commentRaw A comment that will be styled
   */
  void createLog(std::optional<std::string> account, std::optional<std::string> boardRaw,
                 const std::string& commentRaw) const;
  bool cycle(std::string_view board, int64_t number) const;
  /**
   * @brief Retrieves a flag
   *
   * @param flag The flag to retrieve
   * @return The status of the flag
   */
  bool flag(std::string_view flag) const;
  std::vector<std::vector<uint8_t>> loadBanned() const;
  /**
   * @brief Retrieves a page of logs
   *
   * @param page The page to retrieve
   * @return A page of logs
   */
  std::vector<LogEntry> logs(int64_t page) const;
  /**
   * @brief Retrieves a thread's posts
   *
   * @param board The name of the board the thread is on
   * @param parent The post number of the OP
   * @param postLimit The amount of posts to include
   * @return The thread's posts
   */
  std::vector<Post> posts(std::string_view board, int64_t parent, int64_t postLimit) const;
  /**
   * @brief Retrieves a boards posts per hour
   *
   * @param board The board to retrieve for
   * @return The number of posts made in the last hour
   */
  int64_t pph(std::string_view board) const;
  /**
   * @brief Retrieves a page of recent posts
   *
   * Post limit can at max retrieve 128 posts
   * @param page The page to retrieve
   * @param sage Only show saged posts
   * @param file Only show posts with files
   * @param parent Only show threads
   * @param customName Only show posts with custom names
   * @param tvolDelete Only show posts deleted by tvols.
   * @param postLimit The amount of posts to include
   * @return A page of recent posts
   */
  std::vector<Post> recent(int64_t page, bool sage, bool file, bool parent, bool customName, bool tvolDelete,
                           int64_t postLimit) const;
  /**
   * @brief Retrieve a post's parent
   *
   * @param board The board the post is on
   * @param number The number of the post to retrieve the parent for
   * @return A post's parents
   */
  std::optional<int64_t> parent(std::string_view board, int64_t number) const;
  /**
   * @brief Sets whether a board should require a captcha
   *
   * @param board The board that the captcha is being set for
   * @param requireCaptcha Whether the board should require a captcha
   */
  void setCaptcha(std::string_view board, bool requireCaptcha) const;
  /**
   * @brief Retrieves stats for the site
   *
   * Boards are sorted by the number of posts that have been made on it.
   * The final element in the log entries with be a total of all the previous
   * stats.
   * @return A list of stats
   */
  std::vector<StatsEntry> stats() const;
  /**
   * @brief Retrieves a thread from a board
   *
   * @param board The board the thread was posted to
   * @param number The number of the thread
   * @return The thread
   */
  std::optional<Thread> thread(const Board& board, int64_t number) const;
  /**
   * @brief Retrieves a boards threads
   *
   * @param boards The boards to retrieve for
   * @param sticky Whether stickied posts should be sticky
   * @param autosage Whether autosage posts should be filtered out
   * @param threadLimit The maximum amount of threads to return
   * @return A list of threads
   */
  std::vector<Thread> threads(const std::unordered_map<std::string, Board>& boards, bool sticky, bool autosage,
                              int64_t threadLimit) const;
  /**
   * @brief Retrieves a boards threads per day
   *
   * @param board The board to retrieve for
   * @return The number of threads made in the last day
   */
  int64_t tpd(std::string_view board) const;
};

}  // namespace imageboard

#endif
