#include "thumbnailer_system.h"

#include <absl/strings/str_cat.h>
#include <absl/strings/str_format.h>
#include <openssl/sha.h>
#include <sys/stat.h>

#include <bit>
#include <iostream>

#include "../media/jpeg_compressor.h"
#include "../media/thumbnailer.h"
#include "../media/thumbnailer_factory.h"
#include "../util/util.h"

namespace imageboard {

static bool sha256(const std::string_view contents, std::array<uint8_t, SHA256_DIGEST_LENGTH>& data) {
  SHA256_CTX context;
  if (!SHA256_Init(&context)) {
    return false;
  }
  if (!SHA256_Update(&context, contents.data(), contents.length())) {
    return false;
  }
  if (!SHA256_Final(data.data(), &context)) {
    return false;
  }
  return true;
}

static int distance(const std::vector<uint8_t>& v1, const std::vector<uint8_t>& v2) {
  int count = 0;
  for (size_t i = 0; i < v1.size() && i < v2.size(); i++) {
    if (v1[i] != v2[i]) {
      count += std::popcount(static_cast<uint8_t>(v1[i] ^ v2[i]));
    }
  }
  return count;
}

ThumbnailerSystem::ThumbnailerSystem(const ImageBoard& imageboard, util::Filesystem& filesystem)
    : imageboard(imageboard), filesystem(filesystem), banned(imageboard.loadBanned()) {}

std::optional<File> ThumbnailerSystem::saveMedia(const std::string_view originalFilename,
                                                 const std::string_view fileContents, const bool generateIcon) {
  std::optional<std::string> extension = util::detectExtension(originalFilename, fileContents);
  if (!extension) {
    return std::nullopt;
  }
  std::array<uint8_t, SHA256_DIGEST_LENGTH> hash;
  if (!sha256(fileContents, hash)) {
    return std::nullopt;
  }
  static_assert(SHA256_DIGEST_LENGTH == 32);
  std::string filename = absl::StrFormat(
      "%02x%02x%02x%02x%02x%02x%02x%02x"
      "%02x%02x%02x%02x%02x%02x%02x%02x"
      "%02x%02x%02x%02x%02x%02x%02x%02x"
      "%02x%02x%02x%02x%02x%02x%02x%02x.%s",
      hash[0], hash[1], hash[2], hash[3], hash[4], hash[5], hash[6], hash[7], hash[8], hash[9], hash[10], hash[11],
      hash[12], hash[13], hash[14], hash[15], hash[16], hash[17], hash[18], hash[19], hash[20], hash[21], hash[22],
      hash[23], hash[24], hash[25], hash[26], hash[27], hash[28], hash[29], hash[30], hash[31], extension->data());
  int64_t thumbWidth = -1;
  int64_t thumbHeight = -1;
  struct stat statBuffer;
  media::ThumbnailerFactory thumbnailerFactory;
  if (thumbnailerFactory.hasThumbnail(*extension)) {
    // Create thumbnailer
    std::unique_ptr<media::Thumbnailer> thumbnailer = thumbnailerFactory.create(filename, fileContents);
    if (!thumbnailer) {
      return std::nullopt;
    }
    // Check if banned
    std::vector<uint8_t> dhash = thumbnailer->dhash();
    for (const std::vector<uint8_t>& bad : banned) {
      if (distance(bad, dhash) <= 50) {
        return std::nullopt;
      }
    }
    // Check if thumbnail already exists
    std::string thumbnailPath = absl::StrCat(".media/thumb/", filename);
    bool thumbnailExists = !stat(thumbnailPath.c_str(), &statBuffer);
    if (errno != 0 && errno != ENOENT) {
      return std::nullopt;
    }
    // Check if icon already exists
    std::string iconPath = absl::StrCat(".media/icon/", filename);
    bool iconExists = true;
    if (generateIcon) {
      iconExists = !stat(iconPath.c_str(), &statBuffer);
      if (errno != 0 && errno != ENOENT) {
        return std::nullopt;
      }
    }
    // Check if we need to make any thumbnails
    if (thumbnailExists && (!generateIcon || iconExists)) {
      return File{filename, -1, -1, ""};
    }
    // Save icon
    if (generateIcon) {
      if (!iconExists) {
        std::optional<media::JPEG> icon = thumbnailer->scale(100, 70);
        if (!icon) {
          return std::nullopt;
        }
        if (!util::writeBinaryFile(iconPath, icon->buffer)) {
          return std::nullopt;
        }
      }
    }
    // Save thumbnail
    if (!thumbnailExists) {
      std::optional<media::JPEG> thumbnail = thumbnailer->scale(200, 200);
      if (!thumbnail) {
        return std::nullopt;
      }
      if (!util::writeBinaryFile(thumbnailPath, thumbnail->buffer)) {
        return std::nullopt;
      }
      // Return thumbnail dimensions if a thumbnail is generated
      thumbWidth = thumbnail->width;
      thumbHeight = thumbnail->height;
    }
  }
  // Save file
  std::string filePath = absl::StrCat(".media/", filename);
  bool fileExists = !stat(filePath.c_str(), &statBuffer);
  if (errno != 0 && errno != ENOENT) {
    return std::nullopt;
  }
  if (!fileExists) {
    if (!filesystem.writeFile(filePath, fileContents)) {
      return std::nullopt;
    }
  }
  return File{filename, thumbWidth, thumbHeight, ""};
}

}  // namespace imageboard
