#include "codec_context.h"

#include <stdexcept>
#include <utility>

namespace media {

CodecContext::CodecContext(const FormatContext& formatContext) {
  const AVCodec* codec;
  videoStreamIndex = av_find_best_stream(formatContext.context, AVMEDIA_TYPE_VIDEO, -1, -1, &codec, 0);
  if (videoStreamIndex < 0) {
    throw std::runtime_error("Could not find video stream");
  }
  context = avcodec_alloc_context3(codec);
  if (!context) {
    throw std::runtime_error("Could not allocate codec context");
  }
  AVStream* stream = formatContext.context->streams[videoStreamIndex];
  if (avcodec_parameters_to_context(context, stream->codecpar) < 0) {
    throw std::runtime_error("Could not set codec context parameters");
  }
  if (avcodec_open2(context, codec, nullptr) < 0) {
    throw std::runtime_error("Could not initialize codec context");
  }
}

CodecContext::~CodecContext() {
  avcodec_free_context(&context);
}

CodecContext::CodecContext(CodecContext&& other) noexcept {
  context = std::exchange(other.context, nullptr);
  videoStreamIndex = other.videoStreamIndex;
}

CodecContext& CodecContext::operator=(CodecContext&& other) noexcept {
  std::swap(context, other.context);
  videoStreamIndex = other.videoStreamIndex;
  return *this;
}

bool CodecContext::sendPacket(const Packet& packet) const {
  int ret = avcodec_send_packet(context, packet.handle);
  return !ret || ret == AVERROR(EAGAIN);
}

bool CodecContext::receiveFrame(const Frame& frame) const {
  int ret = avcodec_receive_frame(context, frame.handle);
  return !ret || ret == AVERROR(EAGAIN);
}

bool CodecContext::isVideoPacket(const Packet& packet) const { return packet.handle->stream_index == videoStreamIndex; }

}  // namespace media
