#include "io_context.h"

#include <stdexcept>
#include <string_view>
#include <utility>

namespace media {

static int myread(void *opaque, uint8_t *buffer, int bufferSize) {
  auto *seekBuffer = static_cast<SeekBuffer *>(opaque);
  std::string_view contents = seekBuffer->buffer.substr(seekBuffer->offset);
  int size = std::min(static_cast<int>(contents.length()), bufferSize);
  if (!size) {
    return AVERROR_EOF;
  }
  memcpy(buffer, contents.data(), size);
  seekBuffer->offset += size;
  return size;
}

static int64_t myseek(void *opaque, int64_t offset, int whence) {
  auto *seekBuffer = static_cast<SeekBuffer *>(opaque);
  if (whence == SEEK_SET) {
    seekBuffer->offset = offset;
  } else if (whence == SEEK_CUR) {
    seekBuffer->offset += offset;
  } else if (whence == SEEK_END) {
    seekBuffer->offset = seekBuffer->buffer.length() + offset;
  } else if (whence == AVSEEK_SIZE) {
    return seekBuffer->buffer.length();
  } else {
    return -1;
  }
  seekBuffer->offset = std::max(seekBuffer->offset, static_cast<size_t>(0));
  seekBuffer->offset = std::min(seekBuffer->offset, seekBuffer->buffer.length());
  return 0;
}

IOContext::IOContext(std::string_view _buffer) : buffer({_buffer, 0}) {
  auto ioBuffer = static_cast<unsigned char *>(av_malloc(4096));
  if (!ioBuffer) {
    throw std::runtime_error("Could not allocate io buffer");
  }
  context = avio_alloc_context(ioBuffer, 4096, 0, &buffer, myread, nullptr, myseek);
  if (!context) {
    av_freep(ioBuffer);
    throw std::runtime_error("Could not allocate io context");
  }
}

IOContext::~IOContext() {
  if (context) {
    av_freep(&context->buffer);
  }
  avio_context_free(&context);
}

IOContext::IOContext(IOContext &&other) noexcept {
  context = std::exchange(other.context, nullptr);
  buffer = other.buffer;
}

IOContext &IOContext::operator=(IOContext &&other) noexcept {
  std::swap(context, other.context);
  buffer = other.buffer;
  return *this;
}

}  // namespace media
