#include "sws_context.h"

namespace media {

std::optional<Image> SWSContext::scale(const Frame& frame, const uint64_t width, const uint64_t height,
                                       const bool grayscale) {
  AVPixelFormat format;
  bool needsScaling = false;
  switch (frame.handle->format) {
    case -1:
      return std::nullopt;
    case AV_PIX_FMT_YUVJ420P:
      format = AV_PIX_FMT_YUV420P;
      needsScaling = true;
      break;
    case AV_PIX_FMT_YUVJ422P:
      format = AV_PIX_FMT_YUV422P;
      needsScaling = true;
      break;
    case AV_PIX_FMT_YUVJ444P:
      format = AV_PIX_FMT_YUV444P;
      needsScaling = true;
      break;
    case AV_PIX_FMT_YUVJ440P:
      format = AV_PIX_FMT_YUV440P;
      needsScaling = true;
      break;
    default:
      format = static_cast<enum AVPixelFormat>(frame.handle->format);
      break;
  }
  SwsContext* context =
      sws_getContext(frame.handle->width, frame.handle->height, format, width, height,
                     grayscale ? AV_PIX_FMT_GRAY8 : AV_PIX_FMT_RGBA, SWS_BILINEAR, nullptr, nullptr, nullptr);
  if (needsScaling) {
    int tmp[4];
    int* invTable = tmp;
    int srcRange;
    int* table = tmp;
    int dstRange;
    int brightness;
    int contrast;
    int saturation;
    sws_getColorspaceDetails(context, &invTable, &srcRange, &table, &dstRange, &brightness, &contrast, &saturation);
    const int* coefficients = sws_getCoefficients(SWS_CS_DEFAULT);
    srcRange = 1;
    sws_setColorspaceDetails(context, coefficients, srcRange, coefficients, dstRange, brightness, contrast, saturation);
  }
  Image image(width, height);
  sws_scale(context, frame.handle->data, frame.handle->linesize, 0, frame.handle->height, image.planes, image.strides);
  if (!grayscale) {
    // Alpha blend
    for (size_t y = 0; y < height; y++) {
      for (size_t x = 0; x < width; x++) {
        double alpha = (double)image.planes[0][y * image.strides[0] + x * 4 + 3] / 0xFF;
        image.planes[0][y * image.strides[0] + x * 4 + 0] *= alpha;
        image.planes[0][y * image.strides[0] + x * 4 + 1] *= alpha;
        image.planes[0][y * image.strides[0] + x * 4 + 2] *= alpha;
        image.planes[0][y * image.strides[0] + x * 4 + 0] += 0xBB * (1 - alpha);
        image.planes[0][y * image.strides[0] + x * 4 + 1] += 0xDD * (1 - alpha);
        image.planes[0][y * image.strides[0] + x * 4 + 2] += 0xCC * (1 - alpha);
      }
    }
    // Convert RGBA to RGB
    for (size_t y = 0; y < height; y++) {
      // Make sure memcpy regions do not overlap
      if (width > 1) {
        image.planes[0][y * image.strides[0] + 3] = image.planes[0][y * image.strides[0] + 4];
        image.planes[0][y * image.strides[0] + 4] = image.planes[0][y * image.strides[0] + 5];
        image.planes[0][y * image.strides[0] + 5] = image.planes[0][y * image.strides[0] + 6];
      }
      if (width > 2) {
        image.planes[0][y * image.strides[0] + 6] = image.planes[0][y * image.strides[0] + 8];
        image.planes[0][y * image.strides[0] + 7] = image.planes[0][y * image.strides[0] + 9];
        image.planes[0][y * image.strides[0] + 8] = image.planes[0][y * image.strides[0] + 10];
      }
      for (size_t x = 3; x < width; x++) {
        memcpy(&image.planes[0][y * image.strides[0] + x * 3], &image.planes[0][y * image.strides[0] + x * 4], 3);
      }
    }
  }
  sws_freeContext(context);
  return image;
}

}  // namespace media
