#include "video_thumbnailer.h"

#include <stdexcept>

#include "codec_context.h"
#include "format_context.h"
#include "image.h"
#include "io_context.h"
#include "jpeg_compressor.h"
#include "sws_context.h"

#include <iostream>

namespace media {

VideoThumbnailer::VideoThumbnailer(const char *filename,
                                   const std::string_view contents, bool firstFrame) {
  IOContext ioContext(contents);
  FormatContext formatContext(filename, std::move(ioContext));
  CodecContext codecContext(formatContext);
  Packet packet;
  for (;;) {
    if (!formatContext.readFrame(packet)) {
      break;
    }
    if (!codecContext.isVideoPacket(packet)) {
      continue;
    }
    if (!codecContext.sendPacket(packet)) {
      throw std::runtime_error("Could not send video packet to decoder");
    }
    if (!codecContext.receiveFrame(frame)) {
      throw std::runtime_error("Could not decode frame");
    }
    if (formatContext.frameTimestamp(packet, frame) >= 1) {
      break;
    }
    if (firstFrame) {
      break;
    }
  }
  if (!frame.applyCropping()) {
    throw std::runtime_error("Could not apply cropping");
  }
}

std::optional<JPEG> VideoThumbnailer::scale(const int maxWidth, const int maxHeight) {
  int width = frame.handle->width;
  int height = frame.handle->height;
  if (width > maxWidth || height > maxHeight) {
    double ratioX = static_cast<double>(frame.handle->width) / maxWidth;
    double ratioY = static_cast<double>(frame.handle->height) / maxHeight;
    if (ratioX >= ratioY) {
      width = maxWidth;
      height /= ratioX;
    } else {
      width /= ratioY;
      height = maxHeight;
    }
  }
  std::optional<Image> image = SWSContext::scale(frame, width, height);
  if (!image) {
    return std::nullopt;
  }
  JPEGCompressor compressor;
  return compressor.compress(image->planes[0], width, image->strides[0], height, 60);
}

std::vector<uint8_t> VideoThumbnailer::dhash() {
  std::optional<Image> image = SWSContext::scale(frame, 9, 8, true);
  std::vector<uint8_t> hash;
  hash.reserve(9 * 8);
  for (int y = 0; y < 8; y++)
  {
    for (int x = 0; x < 8; x++)
    {
      uint8_t p1 = image->planes[0][y * image->strides[0] + x];
      uint8_t p2 = image->planes[0][y * image->strides[0] + x + 1];
      hash.push_back(p2 - p1);
    }
  }
  return hash;
}

}  // namespace media
