#include "nanochan.h"

#include <absl/strings/str_cat.h>

#include <charconv>

#include "404.h"
#include "500.h"
#include "catalog.h"
#include "error.h"
#include "index.h"
#include "log.h"
#include "overboard.h"
#include "recent.h"
#include "redirect.h"
#include "stats.h"
#include "thread.h"

static size_t maxFileSize = 16 * 1024 * 1024;

namespace nanochan {

Nanochan::Nanochan(const char* database) : imageboard(database), thumbnailerSystem(imageboard, filesystem) {
  routes.emplace("/", &Nanochan::index);
  routes.emplace("/.nano/captcha.jpg", &Nanochan::captcha);
  routes.emplace("/.nano/log", &Nanochan::log);
  routes.emplace("/.nano/overboard", &Nanochan::overboard);
  routes.emplace("/.nano/overboarda", &Nanochan::overboarda);
  routes.emplace("/.nano/post", &Nanochan::post);
  routes.emplace("/.nano/recent", &Nanochan::recent);
  routes.emplace("/.nano/stats", &Nanochan::stats);
  // Legacy
  routes.emplace("/index.html", &Nanochan::index);
  routes.emplace("/overboard.html", &Nanochan::overboard);
  routes.emplace("/overboarda.html", &Nanochan::overboarda);
}

void Nanochan::update() {
  // Check if the cooldown is over
  if (time(nullptr) - harderCaptchaCooldown > 0) {
    // Set the cooldown to 10 seconds
    harderCaptchaCooldown = time(nullptr) + 10;
    bool hard = imageboard.flag("HarderCaptchas");
    captchaSystem.setHard(hard);
  }
  captchaSystem.update();
}

std::string Nanochan::handle(const scgi::Request& request) {
  try {
    auto route = routes.find(request.path);
    if (route != routes.end()) {
      return (this->*route->second)(request);
    }
    if (request.parts.size() == 1) {
      return catalog(request.parts[0]);
    }
    if (request.parts.size() == 2) {
      return thread(request.parts[0], request.parts[1]);
    }
    return __404();
  } catch (const std::exception& exception) {
    return __500(exception.what());
  }
}

std::string Nanochan::__404() {
  std::vector<imageboard::Board> boards = imageboard.boards();
  return _404(boards).render();
}

std::string Nanochan::__500(const std::string_view reason) { return _500(reason).render(); }

std::string Nanochan::captcha([[maybe_unused]] const scgi::Request& request) {
  std::span<uint8_t> image = captchaSystem.image();
  return std::string(reinterpret_cast<char*>(image.data()), image.size());
}

std::string Nanochan::catalog(const std::string& boardName) {
  std::string announcement = imageboard.announcement();
  std::vector<imageboard::Board> boards = imageboard.boards();
  std::string banner = imageboard.banner(boardName);
  for (imageboard::Board& board : boards) {
    if (board.name == boardName) {
      std::vector<imageboard::Thread> threads = imageboard.threads({{boardName, board}}, true, true, board.threadLimit);
      bool threadCaptcha = imageboard.flag("ThreadCaptcha");
      return Catalog(announcement, boards, banner, board, threads, threadCaptcha).render();
    }
  }
  return __500("Could not find board");
}

std::string Nanochan::error(const std::string_view reason, const std::string_view description) {
  std::vector<imageboard::Board> boards = imageboard.boards();
  return Error(boards, reason, description).render();
}

std::string Nanochan::index([[maybe_unused]] const scgi::Request& request) {
  std::string announcement = imageboard.announcement();
  std::vector<imageboard::Board> boards = imageboard.boards();
  return Index(announcement, boards).render();
}

std::string Nanochan::log(const scgi::Request& request) {
  int page = 1;
  std::string_view pageString = request.getURLParameter("page");
  std::from_chars(pageString.data(), pageString.data() + pageString.length(), page);
  std::vector<imageboard::Board> boards = imageboard.boards();
  std::vector<imageboard::LogEntry> entries = imageboard.logs(page);
  return Log(boards, entries, page).render();
}

std::string Nanochan::overboard([[maybe_unused]] const scgi::Request& request) {
  std::string announcement = imageboard.announcement();
  std::vector<imageboard::Board> boards = imageboard.boards();
  std::unordered_map<std::string, imageboard::Board> boardMap;
  for (const imageboard::Board& board : boards) {
    if (board.displayOverboard) {
      boardMap.emplace(board.name, board);
    }
  }
  std::vector<imageboard::Thread> threads = imageboard.threads(boardMap, false, false, 100);
  return Overboard(announcement, boards, threads).render();
}

std::string Nanochan::overboarda([[maybe_unused]] const scgi::Request& request) {
  std::string announcement = imageboard.announcement();
  std::vector<imageboard::Board> boards = imageboard.boards();
  std::unordered_map<std::string, imageboard::Board> boardMap;
  for (const imageboard::Board& board : boards) {
    boardMap.emplace(board.name, board);
  }
  std::vector<imageboard::Thread> threads = imageboard.threads(boardMap, false, true, 100);
  return Overboard(announcement, boards, threads).render();
}

std::string Nanochan::post(const scgi::Request& request) {
  std::string_view boardString = request.getContentsParameter("board");
  int64_t parent = 0;
  std::string_view parentString = request.getContentsParameter("parent");
  std::from_chars(parentString.data(), parentString.data() + parentString.length(), parent);
  std::string_view name = request.getContentsParameter("name");
  std::string_view email = request.getContentsParameter("email");
  std::string_view subject = request.getContentsParameter("subject");
  std::string_view comment = request.getContentsParameter("comment");
  std::string_view file = request.getContentsParameter("file");
  std::string_view captcha = request.getContentsParameter("captcha");
  if (comment.empty() && file.empty()) {
    return error("Blank post",
                 "You must either upload a file or write something in the "
                 "comment field.");
  }
  if (comment.length() > 32768) {
    return error("Post too long", "Your post was over 32 KiB. Please reduce its length.");
  }
  if (name.length() > 64) {
    return error("Name too long",
                 "The text in the name field was over 64 bytes. Please reduce "
                 "its length.");
  }
  if (email.length() > 64) {
    return error("Email too long",
                 "The text in the email field was over 64 bytes. Please reduce "
                 "its length.");
  }
  if (subject.length() > 64) {
    return error("Subject too long",
                 "The text in the subject field was over 64 bytes. Please "
                 "reduce its length.");
  }
  size_t newlines = 0;
  for (char c : comment) {
    if (c == '\n') {
      newlines++;
    }
  }
  if (newlines > 512) {
    return error("Too many newlines", "Your post contained over 512 newlines. Please reduce its length.");
  }
  std::optional<imageboard::Board> board = imageboard.board(boardString);
  if (!board) {
    return error("Invalid board", "The board you tried to post to does not exist.");
  }
  if (board->lock) {
    return error("Board locked", "The board you tried to post in is currently locked.");
  }
  int64_t replies = 0;
  bool requireCaptcha = board->captcha;
  if (parent) {
    std::optional<imageboard::Thread> thread = imageboard.thread(*board, parent);
    if (!thread) {
      return error("Invalid thread", "The thread you tried to post in does not exist. Perhaps it has been deleted.");
    }
    replies = thread->replies;
    if (thread->lock) {
      return error("Thread locked", "The thread you tried to post in is currently locked.");
    }
    if (!thread->cycle && replies >= board->postLimit) {
      return error("Thread full", "The thread you tried to post in is full. Please start a new thread instead.");
    }
  } else {
    if (static_cast<int64_t>(comment.length()) < board->minimumThreadCharacters) {
      return error("Post too short",
                   absl::StrCat("Your post text was too short. On this board, threads require at least ",
                                board->minimumThreadCharacters, " characters."));
    }
    int64_t tpd = imageboard.tpd(boardString);
    if (tpd > board->maximumTPD) {
      return error("Thread limit reached", "The board you tried to post in has reached its daily thread limit.");
    }
    if (!requireCaptcha) {
      bool threadCaptcha = imageboard.flag("ThreadCaptcha");
      requireCaptcha = threadCaptcha;
    }
  }
  if (requireCaptcha) {
    if (!captchaSystem.validate(std::string(captcha))) {
      return error("Invalid captcha",
                   "The captcha you entered was incorrect. Go back, and refresh the page to get a new one.");
    }
  }
  std::string filename = "";
  if (!file.empty()) {
    if (file.length() > maxFileSize) {
      return error("Too high file size", "The file you uploaded was too large.");
    }
    std::optional<imageboard::File> savedFile = thumbnailerSystem.saveMedia(request.getFilename("file"), file, !parent);
    if (!savedFile) {
      return error("File error",
                   "There was a problem with the file you uploaded. Possible reasons include unsupported file type, or "
                   "incorrect file extension.");
    }
    if (savedFile->thumbWidth != -1 && savedFile->thumbHeight != -1) {
      imageboard.addFile(savedFile->filename, savedFile->thumbWidth, savedFile->thumbHeight);
    }
    filename = savedFile->filename;
  }
  std::optional<int64_t> number =
      imageboard.addPost(*board, parent, name, email, subject, comment, filename, board->bumpLimit, replies);
  if (!number) {
    return __500("Could not add post");
  }
  if (parent == 0) {
    // Thread
    return redirect(absl::StrCat("/", boardString, "/", *number));
  } else {
    // Reply
    return redirect(absl::StrCat("/", boardString, "/", parentString, "#post", *number));
  }
}

std::string Nanochan::recent(const scgi::Request& request) {
  int page = 1;
  std::string_view pageString = request.getURLParameter("page");
  std::from_chars(pageString.data(), pageString.data() + pageString.length(), page);
  bool sage = request.getURLParameter("sage") == "on";
  bool file = request.getURLParameter("file") == "on";
  bool parent = request.getURLParameter("parent") == "on";
  bool customName = request.getURLParameter("custname") == "on";
  bool tvolDelete = request.getURLParameter("tvoldelete") == "on";
  uint64_t postLimit = 128;
  std::string_view postLimitString = request.getURLParameter("postlimit");
  std::from_chars(postLimitString.data(), postLimitString.data() + postLimitString.length(), postLimit);
  postLimit = std::min(postLimit, 128uL);
  bool checkedPosts = request.getURLParameter("checkedposts") == "on";
  std::vector<imageboard::Board> boards = imageboard.boards();
  std::vector<imageboard::Post> posts = imageboard.recent(page, sage, file, parent, customName, tvolDelete, postLimit);
  return Recent(boards, posts, page, sage, file, parent, customName, tvolDelete, postLimit, checkedPosts).render();
}

std::string Nanochan::redirect(const std::string_view url) { return Redirect(url).render(); }

std::string Nanochan::stats([[maybe_unused]] const scgi::Request& request) {
  std::vector<imageboard::Board> boards = imageboard.boards();
  std::vector<imageboard::StatsEntry> entries = imageboard.stats();
  return Stats(boards, entries).render();
}

std::string Nanochan::thread(const std::string_view boardName, const std::string& postNumber) {
  int parent;
  try {
    parent = std::stoi(postNumber);
  } catch (...) {
    return __500("Could not parse thread number");
  }
  std::optional<int64_t> parentParent = imageboard.parent(boardName, parent);
  if (!parentParent || *parentParent) {
    return __404();
  }
  std::string announcement = imageboard.announcement();
  std::vector<imageboard::Board> boards = imageboard.boards();
  std::string banner = imageboard.banner(boardName);
  for (imageboard::Board& board : boards) {
    if (board.name == boardName) {
      bool cycle = imageboard.cycle(boardName, parent);
      std::vector<imageboard::Post> posts =
          imageboard.posts(boardName, parent, cycle ? board.postLimit : std::numeric_limits<int64_t>::max());
      std::vector<imageboard::Board>::size_type fileCount = 0;
      for (auto& post : posts) {
        if (post.file) {
          fileCount++;
        }
      }
      return Thread(announcement, boards, banner, board, posts, fileCount).render();
    }
  }
  return __500("Could not find board");
}

}  // namespace nanochan
