#ifndef NANOCHAN_NANOCHAN_H
#define NANOCHAN_NANOCHAN_H

#include "../imageboard/captcha_system.h"
#include "../imageboard/imageboard.h"
#include "../imageboard/thumbnailer_system.h"
#include "../scgi/handler.h"
#include "../util/posix_filesystem.h"

namespace nanochan {

/**
 * @brief Handles SCGI requests
 */
class Nanochan : public scgi::Handler {
 private:
  /**
   * @brief Image board to render off of
   */
  imageboard::ImageBoard imageboard;
  /**
   * @brief The captcha system to use for captchas
   */
  imageboard::CaptchaSystem captchaSystem;
  time_t harderCaptchaCooldown = 0;
  util::PosixFilesystem filesystem;
  imageboard::ThumbnailerSystem thumbnailerSystem;
  /**
   * @brief Stores handler functions for a specific route
   */
  std::unordered_map<std::string_view, std::string (Nanochan::*)(const scgi::Request&)> routes;

 public:
  Nanochan(const char* database);
  /**
   * @brief Updates various data
   */
  void update() override;
  /**
   * @brief Handles passing off the request to the correct handler
   *
   * @param request The SCGI request
   * @return The rendered page
   */
  std::string handle(const scgi::Request& request) override;
  /**
   * @brief Handles the 404 page
   *
   * @return The rendered page
   */
  std::string __404();
  /**
   * @brief Handles the 500 page
   *
   * @return The rendered page
   */
  std::string __500(const std::string_view reason);
  /**
   * @brief Handles displaying captchas
   *
   * @return Binary data for a JPG
   */
  std::string captcha(const scgi::Request& request);
  /**
   * @brief Handles the catalog page
   *
   * @param boardName The name of the board the catalog is for
   * @return The rendered page
   */
  std::string catalog(const std::string& boardName);
  /**
   * @brief Handles the error page
   * @param reason The reason for the error
   * @param description A description for what to do
   * @return The rendered page
   */
  std::string error(const std::string_view reason, const std::string_view description);
  /**
   * @brief Handles the index page
   *
   * @param request unused
   * @return The rendered page
   */
  std::string index(const scgi::Request& request);
  /**
   * @brief Handles the log page
   *
   * @param request The SCGI request
   * @return The rendered page
   */
  std::string log(const scgi::Request& request);
  /**
   * @brief Handles the overboard page
   *
   * @param request unused
   * @return The rendered page
   */
  std::string overboard(const scgi::Request& request);
  /**
   * @brief Handles the overboard all page
   *
   * @param request unused
   * @return The rendered page
   */
  std::string overboarda(const scgi::Request& request);
  /**
   * @brief Handles creating a new post
   *
   * @param request The SCGI request
   * @return The result
   */
  std::string post(const scgi::Request& request);
  /**
   * @brief Handles the recent page
   *
   * @param request The SCGI request
   * @return The rendered page
   */
  std::string recent(const scgi::Request& request);
  /**
   * @brief Handles the redirect page
   * @param url The url to redirect to
   * @return The rendered page
   */
  std::string redirect(const std::string_view url);
  /**
   * @brief Handles the stats page
   *
   * @param request unused
   * @return The rendered page
   */
  std::string stats(const scgi::Request& request);
  /**
   * @brief Handles the thread page
   *
   * @param boardName The board the thread is on
   * @param postNumber The post number of the thread
   * @return The rendered page
   */
  std::string thread(const std::string_view boardName, const std::string& postNumber);
};

}  // namespace nanochan

#endif
