#include "parser.h"

#include <cstring>

namespace scgi {

std::pair<std::unordered_map<std::string_view, std::string_view>,
          std::unordered_map<std::string_view, std::string_view>>
Parser::parseMultiPart(const std::string_view contents,
                       const std::string_view boundary) {
  std::string_view input(contents);
  std::unordered_map<std::string_view, std::string_view> parameterContents;
  std::unordered_map<std::string_view, std::string_view> parameterFiles;
  for (;;) {
    if (!input.starts_with("--")) {
      // Invalid boundary
      break;
    }
    input.remove_prefix(strlen("--"));
    if (!input.starts_with(boundary)) {
      // Invalid boundary
      break;
    }
    input.remove_prefix(boundary.length());
    if (input.starts_with("--")) {
      // Last boundary
      break;
    }
    if (!input.starts_with("\r\n")) {
      // Invalid boundary
      break;
    }
    input.remove_prefix(strlen("\r\n"));
    // Parse part headers
    std::unordered_map<std::string_view, std::string_view> partHeaders;
    bool fail = false;
    for (;;) {
      while (!input.empty() &&
             (input.front() == ' ' || input.front() == '\t')) {
        input.remove_prefix(1);
      }
      size_t split = std::string_view::npos;
      size_t sep = std::string_view::npos;
      for (size_t i = 0; i < input.length(); i++) {
        if (split == std::string_view::npos) {
          if (input[i] == ':' || input[i] == ' ' || input[i] == '\t') {
            split = i;
          }
        }
        if (input[i] == ':') {
          sep = i;
          break;
        }
        if (input.substr(i).starts_with(boundary)) {
          fail = true;
          break;
        }
      }
      if (fail || split == std::string_view::npos ||
          sep == std::string_view::npos) {
        fail = true;
        break;
      }
      std::string_view name = input.substr(0, split);
      input.remove_prefix(sep + strlen(":"));
      while (!input.empty() &&
             (input.front() == ' ' || input.front() == '\t')) {
        input.remove_prefix(1);
      }
      size_t lineSplit = input.find("\r\n");
      partHeaders.emplace(name, input.substr(0, lineSplit));
      input.remove_prefix(lineSplit + strlen("\r\n"));
      if (input.starts_with("\r\n")) {
        // Two \r\n in a row means headers are done
        input.remove_prefix(strlen("\r\n"));
        break;
      }
    }
    size_t end = input.find(boundary);
    // Parse part content
    if (end != std::string_view::npos) {
      if (end < 4) {
        // Invalid boundary
        break;
      }
      end -= 4;
    } else {
      end = input.length();
    }
    if (!fail) {
      auto contentDisposition = partHeaders.find("Content-Disposition");
      if (contentDisposition != partHeaders.end()) {
        std::unordered_map<std::string_view, std::string_view> disposition =
            Parser::parseContentDisposition(contentDisposition->second);
        auto name = disposition.find("name");
        if (name != disposition.end()) {
          parameterContents.emplace(name->second, input.substr(0, end));
          auto filename = disposition.find("filename");
          if (filename != disposition.end()) {
            parameterFiles.emplace(name->second, filename->second);
          }
        }
      }
    }
    input.remove_prefix(end);
    if (!input.starts_with("\r\n")) {
      // Invalid boundary
      break;
    }
    input.remove_prefix(strlen("\r\n"));
  }
  return {parameterContents, parameterFiles};
}

std::unordered_map<std::string_view, std::string_view>
Parser::parseContentDisposition(const std::string_view contentDisposition) {
  std::string_view input(contentDisposition);
  std::unordered_map<std::string_view, std::string_view> map;
  size_t formData = input.find("form-data;");
  if (formData == std::string_view::npos) {
    return map;
  }
  input.remove_prefix(formData + strlen("form-data;"));
  while (!input.empty()) {
    while (!input.empty() && (input.front() == ' ' || input.front() == '\t')) {
      input.remove_prefix(1);
    }
    size_t split = std::string_view::npos;
    size_t sep = std::string_view::npos;
    for (size_t i = 0; i < input.length(); i++) {
      if (split == std::string_view::npos) {
        if (input[i] == '=' || input[i] == ' ' || input[i] == '\t') {
          split = i;
        }
      }
      if (input[i] == '=') {
        sep = i;
        break;
      }
    }
    if (split == std::string_view::npos || sep == std::string_view::npos) {
      break;
    }
    std::string_view name = input.substr(0, split);
    input.remove_prefix(sep + strlen("="));
    while (!input.empty() && (input.front() == ' ' || input.front() == '\t')) {
      input.remove_prefix(1);
    }
    if (!input.starts_with('"')) {
      break;
    }
    input.remove_prefix(strlen("\""));
    size_t end = input.find('"');
    if (end == std::string_view::npos) {
      break;
    }
    std::string_view value = input.substr(0, end);
    input.remove_prefix(end + strlen("\""));
    map.emplace(name, value);
    while (!input.empty() && (input.front() == ' ' || input.front() == '\t')) {
      input.remove_prefix(1);
    }
    if (input.empty() || input.front() != ';') {
      break;
    }
    input.remove_prefix(strlen(";"));
  }
  return map;
}

std::unordered_map<std::string_view, std::string_view> Parser::urlDecode(
    const std::string_view& queryParams) {
  std::string_view input(queryParams);
  std::unordered_map<std::string_view, std::string_view> map;
  while (!input.empty()) {
    size_t split = input.find('=');
    if (split == std::string_view::npos) {
      break;
    }
    std::string_view key = input.substr(0, split);
    input.remove_prefix(split + strlen("="));
    size_t end = input.find('&');
    std::string_view value;
    if (end != std::string_view::npos) {
      value = input.substr(0, end);
      input.remove_prefix(end + strlen("&"));
    } else {
      value = input.substr(0, input.length());
      input.remove_prefix(input.length());
    }
    map.emplace(key, value);
  }
  return map;
}

}  // namespace scgi
