#include "request.h"

#include <absl/strings/str_cat.h>
#include <sys/socket.h>

#include <cassert>
#include <cstring>

#include "../util/util.h"
#include "parser.h"

namespace scgi {

constexpr ssize_t lengthPeek =
    16;  // One more than max number of digits in length
constexpr std::string_view formData = "multipart/form-data";

Request::Request(const int socket) : handle(socket) {}

Request::~Request() {
  delete[] headersBuffer;
  delete[] contentsBuffer;
}

Request::Request(Request&& other) noexcept {
  handle = other.handle;
  state = other.state;
  headersBuffer = std::exchange(other.headersBuffer, nullptr);
  headersLength = other.headersLength;
  contentsBuffer = std::exchange(other.contentsBuffer, nullptr);
  contentsLength = other.contentsLength;
  path = std::move(other.path);
  parts = std::move(other.parts);
  urlParameters = std::move(other.urlParameters);
  contentsParameters = std::move(other.contentsParameters);
  parameterFiles = std::move(other.parameterFiles);
}

Request& Request::operator=(Request&& other) noexcept {
  handle = other.handle;
  state = other.state;
  std::swap(headersBuffer, other.headersBuffer);
  std::swap(contentsBuffer, other.contentsBuffer);
  headersLength = other.headersLength;
  contentsLength = other.contentsLength;
  path = std::move(other.path);
  parts = std::move(other.parts);
  urlParameters = std::move(other.urlParameters);
  parameterFiles = std::move(other.parameterFiles);
  return *this;
}

/*
 * SCGI headers are sent as a netstring as defined in
 * https://cr.yp.to/proto/netstrings.txt In short a netstring is encoded as:
 * <length>:<string>,
 *
 * The length field is encoded in ASCII and has no defined upper bound.
 * This method limits the length by limiting it to a fixed number of digits.
 * To simplify implementation this method starts off by peaking a fixed amount
 * of bytes into the request and tries to parse the length of the netstring.
 * Using the length of the netstring, the entire netstring is read in and copies
 * to a buffer in memory.
 *
 * The netstring is encoded based off of section 3 of
 * https://python.ca/scgi/protocol.txt The netstring encodes header name value
 * pairs by alternating between putting C strings representing the names and
 * values. Pairs of pointers to thse C strings are gathered and stored in the
 * headers map. The headers map only stores pointers to strings meaning that
 * string data does not need to be copied around.
 *
 * The SCGI protocol guarentees that both a CONTENT_LENGTH and SCGI header
 * exist. This method ignores whether or not the SCGI header has been read. The
 * CONTENT_LENGTH header is parsed and stored to be used later when optionally
 * parsing the request's content.
 */
bool Request::parseHeaders() {
  assert(state ==
         ParseState::start);  // Check if this is the first time we are parsing
  assert(!headersBuffer);   // headersBuffer should not have been allocated yet
  assert(headers.empty());  // headers should not have been parsed yet
  // Peek into the request we are recieving to get the length of the headers
  // data
  char lengthBuffer[lengthPeek];
  ssize_t peaked = recv(handle, lengthBuffer, lengthPeek, MSG_PEEK);
  if (peaked <= 0) {
    // Fail if we cannot read any data
    return false;
  }
  // Parse what we read into the lengthBuffer from ASCII to an size_t
  ssize_t lengthLength = 0;
  state = ParseState::headers;
  for (; lengthBuffer[lengthLength] != ':'; lengthLength++) {
    if (lengthLength >= peaked) {
      // Fail if we run out of buffer without seeing a :
      return false;
    }
    if (lengthBuffer[lengthLength] < '0' || lengthBuffer[lengthLength] > '9') {
      // Fail if we encounter characters that are not digits
      return false;
    }
    // Add a decimal digit on the right
    headersLength *= 10;
    headersLength += lengthBuffer[lengthLength] - static_cast<int>('0');
  }
  lengthLength++;  // Extend length to include :
  headersLength +=
      lengthLength + 1;  // Extend length to be for the whole netstring
  // Allocate a buffer and read in the headers netstring into it
  headersBuffer = new char[headersLength];
  size_t total = 0;
  while (total < headersLength) {
    ssize_t bytesRead =
        recv(handle, headersBuffer + total, headersLength - total, 0);
    if (bytesRead <= 0) {
      // Fail if we cannot read the entire length
      return false;
    }
    total += bytesRead;
  }
  if (headersBuffer[headersLength - 1] != ',') {
    // Fail if the netstring does not end with a comma
    return false;
  }
  // Alternate between treating C strings as names and values
  char* name = headersBuffer + lengthLength;
  char* value;
  state = ParseState::name;
  for (size_t i = 0; i < headersLength; i++) {
    if (!headersBuffer[i]) {
      if (state == ParseState::name) {
        state = ParseState::value;
        value = headersBuffer + i + 1;
        // Save the name and value once we have read both
        headers.emplace(name, value);
      } else {
        state = ParseState::name;
        name = headersBuffer + i + 1;
      }
    }
  }
  if (state != ParseState::name) {
    // Fail if we had an incomplete name value pair
    return false;
  }
  // Parse the value of the CONTENT_LENGTH header from ASCII to an size_t
  auto contentLengthHeader = headers.find("CONTENT_LENGTH");
  if (contentLengthHeader == headers.end()) {
    // The SCGI standard guarentees CONTENT_LENGTH to exist
    return false;
  }
  char* content_length = contentLengthHeader->second;
  for (size_t i = 0; content_length[i]; i++) {
    if (content_length[i] < '0' || content_length[i] > '9') {
      // Fail if we encounter characters that are not digits
      return false;
    }
    // Add a decimal digit on the right
    contentsLength *= 10;
    contentsLength += content_length[i] - static_cast<int>('0');
  }
  // Get path from DOCUMENT_URI or fallback on to SCRIPT_NAME and PATH_INFO
  auto documentURI = headers.find("DOCUMENT_URI");
  if (documentURI != headers.end()) {
    path = documentURI->second;
  } else {
    auto scriptName = headers.find("SCRIPT_NAME");
    if (scriptName != headers.end()) {
      path = scriptName->second;
    }
    auto pathInfo = headers.find("PATH_INFO");
    if (pathInfo != headers.end()) {
      absl::StrAppend(&path, pathInfo->second);
    }
  }
  if (!path.empty()) {
    size_t start = 1;
    for (size_t i = 1; path[i]; i++) {
      if (path[i] == '/') {
        parts.push_back(path.substr(start, i - start));
        start = i + 1;
      } else if (!path[i + 1]) {
        parts.push_back(path.substr(start));
        break;
      }
    }
  }
  // Set the state to content to indicate that we are ready for the content to
  // now be read
  state = ParseState::contents;
  return true;
}

/*
 * If the request has contents read it into a buffer.
 * Then try to parse QUERY_STRING into parameters
 */
bool Request::parseContents() {
  assert(state ==
         ParseState::contents);  // Check if we are in the correct state
  assert(!contentsBuffer);  // contentsBuffer should not have been allocated yet
  // Skip if there are no contents
  if (contentsLength) {
    // Allocate a buffer and read in the contents section into it
    contentsBuffer = new char[contentsLength];
    size_t total = 0;
    while (total < contentsLength) {
      ssize_t bytesRead =
          recv(handle, contentsBuffer + total, contentsLength - total, 0);
      if (bytesRead <= 0) {
        // Fail if we cannot read the entire length
        return false;
      }
      total += bytesRead;
    }
  }
  // Set the state to finish to indicate that parsing has finished
  state = ParseState::finish;
  auto queryStringPair = headers.find("QUERY_STRING");
  if (queryStringPair == headers.end()) {
    return false;
  }
  urlParameters = Parser::urlDecode(std::string_view(queryStringPair->second));
  auto contentType = headers.find("CONTENT_TYPE");
  if (contentType == headers.end()) {
    return true;
  }
  if (!strcmp(contentType->second, "application/x-www-form-urlencoded")) {
    contentsParameters =
        Parser::urlDecode(std::string_view(contentsBuffer, contentsLength));
  } else if (!strncmp(contentType->second, formData.data(),
                      formData.length())) {
    std::string_view contents(contentsBuffer, contentsLength);
    std::string_view boundary =
        contentType->second + sizeof("multipart/form-data; boundary=") - 1;
    std::tie(contentsParameters, parameterFiles) =
        Parser::parseMultiPart(contents, boundary);
  }
  return true;
}

std::string_view Request::getURLParameter(
    const std::string_view parameter) const {
  return util::at_or(urlParameters, parameter, "");
}

std::string_view Request::getContentsParameter(
    const std::string_view parameter) const {
  return util::at_or(contentsParameters, parameter, "");
}

std::string_view Request::getFilename(const std::string_view parameter) const {
  return util::at_or(parameterFiles, parameter, "");
}

}  // namespace scgi
