#ifndef SCGI_REQUEST_H
#define SCGI_REQUEST_H

#include <string>
#include <unordered_map>
#include <vector>

namespace scgi {

/**
 * @brief Represents the states in the DFA for parsing a request
 */
enum class ParseState {
  /// Parsing has not begun
  start,
  /// The parser is ready to parse the headers section
  headers,
  /// The parser is ready to parse a names,
  name,
  /// The parser is ready to parse a value
  value,
  /// The parser is ready to parse the contents section
  contents,
  /// Parsing has finished,
  finish,
};

/**
 * @brief Handles reading and parsing a SCGI request
 */
class Request {
 private:
  /**
   * @brief A handle for the socket for the request
   */
  int handle;
  /**
   * @brief The current state of parsing the incoming request
   */
  ParseState state = ParseState::start;
  /**
   * @brief Stores string data for header names and values
   */
  char* headersBuffer = nullptr;
  /**
   * @brief Holds the length of #headersBuffer
   */
  size_t headersLength = 0;

 public:
  /**
   * @brief Creates a new request
   *
   * @param socket A socket that has not been converted into a request yet
   */
  Request(const int socket);
  ~Request();
  Request(const Request& other) = delete;
  Request& operator=(const Request& other) = delete;
  /**
   * @brief Move constructor
   *
   * @param other Other request
   */
  Request(Request&& other) noexcept;
  /**
   * @brief Move assignment
   *
   * @param other Other request
   */
  Request& operator=(Request&& other) noexcept;
  /**
   * @brief Parses the headers section of the request
   *
   * This method reads the contents section from the socket.
   * After reading the section it is parsed into #contentsBuffer.
   * This function parseContents() Request::headers
   * @note This function should be only called once
   * @note This function should be called before parseContents()
   * @return Status of parsing the headers section
   */
  bool parseHeaders();
  /**
   * @brief Parses the contents section of the request
   *
   * This method reads the contents part of the request.
   * It also parses the QUERY_STRING header into Request::parameters
   * @note This function should be only called once
   * @note This function should be called after parseHeaders()
   * @return Status of parsing the contents section
   */
  bool parseContents();
  std::string_view getURLParameter(const std::string_view parameter) const;
  std::string_view getContentsParameter(const std::string_view parameter) const;
  std::string_view getFilename(const std::string_view parameter) const;
  /**
   * @brief Holds header name value pairs
   */
  std::unordered_map<std::string_view, char*>
      headers;  // TODO: move value to string_view
  /**
   * @brief Holds the request's contents
   */
  char* contentsBuffer = nullptr;
  /**
   * @brief Holds the length of #contentsBuffer
   *
   * @warning May not include any null bytes
   */
  size_t contentsLength = 0;
  /**
   * @brief Holds the path that was requested
   *
   * This path will not include any query parameters.
   * If parseHeaders() can not parse the path it will default to "/".
   */
  std::string path = "/";
  /**
   * @brief Holds the parts of the path
   *
   * If parseHeaders() can not parse the path it will be empty
   */
  std::vector<std::string> parts;
  /**
   * @brief Holds parameter key value pairs from the URL
   */
  std::unordered_map<std::string_view, std::string_view> urlParameters;
  /**
   * @brief Holds parameter key value pairs from the contents
   */
  std::unordered_map<std::string_view, std::string_view> contentsParameters;
  /**
   * @brief Holds filename key value pairs from the contents
   */
  std::unordered_map<std::string_view, std::string_view> parameterFiles;
};

}  // namespace scgi

#endif
