#include "server.h"

#include <arpa/inet.h>
#include <fcntl.h>
#include <poll.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>

#include <algorithm>
#include <csignal>
#include <cstring>

#include "request.h"

namespace scgi {

static volatile sig_atomic_t shutdown = 0;

static void signalHandler([[maybe_unused]] int signal) { shutdown = 1; }

bool Server::start(Handler &handler) {
  // Create socket
  int serverSocket = socket(AF_INET, SOCK_STREAM, 0);
  if (serverSocket == -1) {
    perror("socket");
    return false;
  }
  int serverStatusFlags = fcntl(serverSocket, F_GETFL);
  if (serverStatusFlags < 0) {
    perror("fctnl");
    return false;
  }
  if (fcntl(serverSocket, F_SETFL, serverStatusFlags | O_NONBLOCK) == -1) {
    perror("fcntl");
    return false;
  }
  // Set REUSEADDR and REUSEPORT for convenience
  int value = 1;
  if (setsockopt(serverSocket, SOL_SOCKET, SO_REUSEADDR, &value, sizeof(int)) == -1) {
    perror("setsockopt");
    return false;
  }
  if (setsockopt(serverSocket, SOL_SOCKET, SO_REUSEPORT, &value, sizeof(int)) == -1) {
    perror("setsockopt");
    return false;
  }
  // Bind to 127.0.0.1:8080
  struct sockaddr_in serverAddress;
  memset(&serverAddress, 0, sizeof(serverAddress));
  serverAddress.sin_family = AF_INET;
  serverAddress.sin_addr.s_addr = inet_addr("127.0.0.1");
  serverAddress.sin_port = htons(8080);
  if (bind(serverSocket, (sockaddr *)&serverAddress, sizeof(serverAddress)) == -1) {
    perror("bind");
    return false;
  }
  // Listen on socket
  if (listen(serverSocket, 100) == -1) {
    perror("listen");
    return false;
  }
  // Setup signal handler
  signal(SIGINT, signalHandler);
  signal(SIGTERM, signalHandler);
  // Create event loop
  std::vector<struct pollfd> fds;
  fds.push_back({serverSocket, POLLIN, 0});
  while (!(shutdown && fds.size() == 1) && (poll(fds.data(), fds.size(), 1000) >= 0 || errno == EAGAIN)) {
    if (!shutdown && fds[0].revents & POLLIN) {
      struct sockaddr_in clientAddress;
      socklen_t clientAddressLength = sizeof(clientAddress);
      int clientSocket = accept(serverSocket, (sockaddr *)&clientAddress, &clientAddressLength);
      if (clientSocket != -1) {
        int clientStatusFlags = fcntl(serverSocket, F_GETFL);
        if (clientStatusFlags >= 0) {
          if (fcntl(serverSocket, F_SETFL, clientStatusFlags | O_NONBLOCK) != -1) {
            fds.push_back({clientSocket, POLLIN, 0});
          } else {
            perror("fcntl");
            close(clientSocket);
          }
        } else {
          perror("fcntl");
          close(clientSocket);
        }
      }
    }
    for (size_t i = 1; i < fds.size(); i++) {
      if (fds[i].revents & POLLIN) {
        // Handle request
        int clientSocket = fds[i].fd;
        Request request(clientSocket);
        if (request.parseHeaders() && request.parseContents()) {
          // Send response
          if (request.path.ends_with(".jpg")) {
            std::string jpg =
                "Status: 200 OK\r\n"
                "Content-Type: image/jpeg\r\n"
                "\r\n";
            send(clientSocket, jpg.c_str(), jpg.length(), 0);
          } else {
            std::string html =
                "Status: 200 OK\r\n"
                "Content-Type: text/html; charset=utf-8\r\n"
                "\r\n";
            send(clientSocket, html.c_str(), html.length(), 0);
          }
          std::string response = handler.handle(request);
          send(clientSocket, response.c_str(), response.length(), 0);
        }
        close(clientSocket);
        fds[i].revents &= POLLNVAL;
      }
    }
    handler.update();
    // Remove invalid fds after server socket
    auto end =
        std::remove_if(std::next(fds.begin()), fds.end(), [](struct pollfd fd) { return fd.revents & POLLNVAL; });
    fds.erase(end, fds.end());
  }
  signal(SIGINT, SIG_DFL);
  signal(SIGTERM, SIG_DFL);
  close(serverSocket);
  return false;
}

}  // namespace scgi
