#include "compiler.h"

#include <absl/strings/str_cat.h>
#include <unistd.h>

#include <iostream>

#include "../util/util.h"
#include "parser.h"

namespace templating {

/*
 * This functions goal is to generate source code for a render
 * function with a prototype of std::string render()
 * First we generate the body of the render function and gather
 * all of the strings that we need.
 * String literals that are able to be combined will be combined
 * with each other to improve performance (assuming no optimization).
 * Using that information we generate the entire file.
 * We take in a header file for the class this function belongs
 * to because we want to be able to access members defined in
 * the header. Using forward decleration could get rid of this
 * dependency, but we would need to know both the types and name
 * of all of the variables that we want to be able to use.
 * The namespace and class are passed into the function in order
 * to get the correct signature for the render function such that
 * the linker will be able to link our generated code to the rest
 * of the application.
 */
void Compiler::generateCode(util::Filesystem& filesystem, const std::string_view header,
                            const std::string_view ns, const std::string_view cl, const std::string& filename,
                            Result& result) {
  // Flatten the results into a single vector that is not self referential
  result.flatten();
  // Generate render function and gather strings
  std::vector<std::string> strings;
  std::string render;
  bool dirty = true;
  bool appending = false;
  size_t indent = 1;
  for (Token& token : result.root) {
    // Check if we should end appending
    switch (token.type) {
      case TokenType::forBegin:
      case TokenType::forEnd:
      case TokenType::ifBegin:
      case TokenType::elseBegin:
      case TokenType::ifEnd:
      case TokenType::let:
        if (appending) {
          absl::StrAppend(&render, ");\n");
        }
        appending = false;
        break;
      default:
        break;
    }
    switch (token.type) {
      case TokenType::text:
      case TokenType::vspace:
        // If dirty is not set we can just append to the last string
        if (!dirty) {
          absl::StrAppend(&strings.back(), token.data[0]);
          break;
        }
        // Else we need to start working on a new string
        strings.push_back(token.data[0]);
        dirty = false;
        // Add code for using this new string
        if (!appending) {
          absl::StrAppend(&render, std::string(indent * 2, ' '), "absl::StrAppend(&buffer, strings[",
                          strings.size() - 1, "]");
          appending = true;
        } else {
          absl::StrAppend(&render, ", strings[", strings.size() - 1, "]");
        }
        break;
      case TokenType::block:
        // These tokens were removed when flattening
        break;
      case TokenType::forBegin:
        // Create a for loop with an iterator
        absl::StrAppend(&render, std::string(indent * 2, ' '), "for (auto [it, i] = std::tuple{", token.data[1],
                        ".begin(), static_cast<size_t>(0)}; it != ", token.data[1], ".end(); it++, i++) {\n");
        indent++;
        absl::StrAppend(&render, std::string(indent * 2, ' '), "auto& ", token.data[0], " = *it;\n");
        // We need to start appending to a new string
        dirty = true;
        break;
      case TokenType::forEnd:
      case TokenType::ifEnd:
        indent--;
        // Close up block loop
        absl::StrAppend(&render, std::string(indent * 2, ' '), "}\n");
        // We need to start appending to a new string
        dirty = true;
        break;
      case TokenType::ifBegin:
        // Create if block
        absl::StrAppend(&render, std::string(indent * 2, ' '), "if (", token.data[0], ") {\n");
        indent++;
        // We need to start appending to a new string
        dirty = true;
        break;
      case TokenType::elseBegin:
        // Create else block
        absl::StrAppend(&render, std::string((indent - 1) * 2, ' '), "} else {\n");
        // We need to start appending to a new string
        dirty = true;
        break;
      case TokenType::let:
        absl::StrAppend(&render, std::string(indent * 2, ' '), "const auto ", token.data[0], " = ", token.data[1],
                        ";\n");
        break;
      case TokenType::expression:
        // Directly include expression into code
        if (!appending) {
          absl::StrAppend(&render, std::string(indent * 2, ' '), "absl::StrAppend(&buffer, ", token.data[0]);
          appending = true;
        } else {
          absl::StrAppend(&render, ", ", token.data[0]);
        }
        // We need to start appending to a new string
        dirty = true;
        break;
    }
  }
  absl::StrAppend(&render, ");\n");
  // Generate the complete file
  std::string code;
  // Include header defining class
  absl::StrAppend(&code, "#include \"", header,
                  "\"\n"
                  "\n"
                  "#include <absl/strings/str_cat.h>\n"
                  "\n"
                  "#include <tuple>\n"
                  "\n"
                  "namespace ",  // Enter the correct namespace
                  ns,
                  " {\n"
                  "\n"
                  "constexpr std::string_view strings[] = {\n");
  // Include all of the string constants we found before
  for (std::string& string : strings) {
    absl::StrAppend(&code, "    \"", util::escapeLiteral(string, indent * 4), "\",\n");
  }
  // Estimate the length of the rendered string to reduce the number of
  // reallocations
  absl::StrAppend(&code,
                  "};\n"
                  "\n"
                  "std::string ",
                  cl,
                  "::render() {\n"
                  "  std::string buffer;\n",
                  render,  // Use the render function body we generated before
                  "  return buffer;\n"
                  "}\n"
                  "\n"
                  "}  // namespace ",
                  ns, "\n");
  // Write the file to disk
  filesystem.writeFile(filename, code);
}

/*
 * In order for templates to be able to reference other template
 * files from within themselves the current directory must be changed
 * to the template directory.
 * Since we want the output path to be relative to the old working
 * directory we must save the old working directory and restore it
 * after parsing the templates.
 * Then we can change back to the old working directory and generate
 * the source code for the template.
 */
bool Compiler::compile(const std::string_view header, const std::string_view ns, const std::string_view cl,
                       const std::string_view in, const std::string& out) {
  // Split up in into the template directory and the template file
  std::string dir;
  std::string filename;
  size_t seperator = in.find_last_of('/');
  if (seperator == std::string::npos) {
    dir = ".";
    filename = in;
  } else {
    dir = in.substr(0, seperator + 1);
    filename = in.substr(seperator + 1);
  }
  // Save the old working directory
  char cwd[1024];
  if (!getcwd(cwd, sizeof(cwd))) {
    perror("getcwd");
    return false;
  }
  // Switch into the template directory
  if (chdir(dir.c_str()) == -1) {
    perror("chdir");
    return false;
  }
  // Parse the templates
  util::PosixFilesystem filesystem;
  Parser parser(filesystem);
  std::optional<Result> result = parser.templateParser(filename);
  // Switch back to the old working directory
  if (chdir(cwd) == -1) {
    perror("chdir");
    return false;
  }
  if (!result) {
    return false;
  }
  // Generate the source code for the templates
  generateCode(filesystem, header, ns, cl, out, *result);
  return true;
}

}  // namespace templating
