#include "parser.h"

#include <absl/strings/str_cat.h>

#include <iostream>

namespace templating {

/**
 * @brief Prints out a compile error
 *
 * @param reason The reason the compile failed
 * @param state The current parser state
 */
static void compileError(const std::string_view reason, const State& state) {
  std::cerr << "Error compiling " << state.file();
  std::cerr << " at " << state.line() << ":" << state.column();
  std::cerr << " " << reason << std::endl;
}

Parser::Parser(util::Filesystem& filesystem) : filesystem(filesystem) {}

/*
 * <template> ::= <include> <vspace> <blocks> | (<content> <vspace>)*
 * 1. Goto 2. if we can parse an extends block else goto 4.
 * 2. Parse vspace
 * 3. Parse blocks
 * 4. Succeed if we are at the end of the file
 * 5. Goto 7. if we can not parse content followed by a vspace
 * 6. Goto 5.
 * 7. Fail if we are not at the end of the file
 */
std::optional<Result> Parser::templateParser(const std::string& filename) {
  std::optional<std::string> contents = filesystem.readFile(filename);
  if (!contents) {
    return std::nullopt;
  }
  State state(filename, *contents);
  state.push();
  // 1.
  std::optional<Result> extends = includeParser(state);
  if (extends) {
    // 2.
    vspaceParser(state);
    // 3.
    Blocks blocks = blocksParser(state);
    // Merge the newly parsed blocks with the blocks from the base template
    std::move(extends->blocks.begin(), extends->blocks.end(), std::inserter(blocks, blocks.begin()));
    extends->blocks = std::move(blocks);
    // 4.
    if (!state.end()) {
      compileError("Unexpected end of content", state);
      return std::nullopt;
    }
    return *extends;
  }
  state.pop();
  Result result;
  for (;;) {
    std::optional<Result> content = contentParser(state);
    // 5.
    if (!content) {
      break;
    }
    // Merge the result from content with the total result
    std::move(content->root.begin(), content->root.end(), std::inserter(result.root, result.root.end()));
    std::move(result.blocks.begin(), result.blocks.end(), std::inserter(content->blocks, content->blocks.begin()));
    result.blocks = std::move(content->blocks);
    result.root.push_back(vspaceParser(state));
    // 6.
  }
  // 7.
  if (!state.end()) {
    compileError("Unexpected end of content", state);
    return std::nullopt;
  }
  return result;
}

/*
 * <blocks> ::= (<block> <vspace>)*
 * 1. Stop if we can not parse block followed by a vspace
 * 2. Goto 1.
 */
Blocks Parser::blocksParser(State& state) {
  Blocks result;
  for (;;) {
    std::optional<Result> blocks = blockParser(state);
    // 1.
    if (!blocks) {
      break;
    }
    // Merge block with the rest of the blocks
    std::move(blocks->blocks.begin(), blocks->blocks.end(), std::inserter(result, result.begin()));
    vspaceParser(state);
    // 2.
  }
  return result;
}

/*
 * <block> ::= "{% block " <name> " %}" <vspace> (<content> <vspace>)* "{% endblock %}"
 * 1. Fail if we do not start with "{{ block "
 * 2. Parse name successfully
 * 3. Fail if we do not continue with " %} "
 * 4. Parse vspace
 * 5. Goto 7. if we can not parse content followed by a vspace
 * 6. Goto 5.
 * 7. Fail if we do not end with "{% endfor %}"
 */
std::optional<Result> Parser::blockParser(State& state) {
  state.push();
  // 1.
  if (!match("{% block ", state)) {
    state.pop();
    return std::nullopt;
  }
  state.drop();
  // 2.
  std::optional<std::string> name = nameParser(state);
  if (!name) {
    compileError("Could not parse name", state);
    return std::nullopt;
  }
  // 3.
  if (!match(" %}", state)) {
    compileError("Missing %}", state);
    return std::nullopt;
  }
  // 4.
  vspaceParser(state);
  Result result;
  Token block(TokenType::block);
  block.data[0] = *name;
  result.root.push_back(std::move(block));
  std::vector<Token> tokens;
  for (;;) {
    std::optional<Result> content = contentParser(state);
    // 5.
    if (!content) {
      break;
    }
    // Merge the result from content with the total blocks
    std::move(content->root.begin(), content->root.end(), std::inserter(tokens, tokens.end()));
    std::move(result.blocks.begin(), result.blocks.end(), std::inserter(content->blocks, content->blocks.begin()));
    result.blocks = std::move(content->blocks);
    tokens.push_back(vspaceParser(state));
    // 6.
  }
  // Remove vertical space that precedes {% endblock %}
  if (!tokens.empty() && tokens.back().type == TokenType::vspace) {
    tokens.pop_back();
  }
  // Include the found block in the result
  result.blocks[std::move(*name)] = std::move(tokens);
  // 7.
  if (!match("{% endblock %}", state)) {
    compileError("Missing {% endblock %}", state);
    return std::nullopt;
  }
  return result;
}

/*
 * <for> ::= "{% for " <name> " in " ~("%}")* " %}" <vspace> (<content>
 * <vspace>)* "{% endfor %}"
 * 1. Fail if we do not start with "{% for "
 * 2. Parse name successfully
 * 3. Fail if we do not continue with " in "
 * 4. Goto 6. if the next 3 characters are " %}"
 * 5. Fail if the next 2 characters are "%}"
 * 6. Goto 4.
 * 7. Fail if we do not continue with " %} "
 * 8. Parse vspace
 * 9. Goto 11. if we can not parse content followed by a vspace
 * 10. Goto 9.
 * 11. Fail if we do not end with "{% endfor %}"
 */
std::optional<Result> Parser::forParser(State& state) {
  state.push();
  // 1.
  if (!match("{% for ", state)) {
    state.pop();
    return std::nullopt;
  }
  state.drop();
  // 2.
  Result result;
  Token forBegin(TokenType::forBegin);
  std::optional<std::string> item = nameParser(state);
  if (!item) {
    compileError("Could not parse item name", state);
    return std::nullopt;
  }
  forBegin.data[0] = std::move(*item);
  // 3.
  if (!match(" in ", state)) {
    compileError("Missing in", state);
    return std::nullopt;
  }
  for (;;) {
    std::optional<char> c1 = state.peek(0);
    std::optional<char> c2 = state.peek(1);
    std::optional<char> c3 = state.peek(2);
    if (!c1 || !c2 || !c3) {
      break;
    }
    // 4.
    if (*c1 == ' ' && *c2 == '%' && *c3 == '}') {
      break;
    }
    // 5.
    if (*c1 == '%' && *c2 == '}') {
      break;
    }
    forBegin.data[1].push_back(*c1);
    state.read();
    // 6.
  }
  // 7.
  if (!match(" %}", state)) {
    compileError("Missing %}", state);
    return std::nullopt;
  }
  result.root.push_back(std::move(forBegin));
  // 8.
  vspaceParser(state);
  for (;;) {
    std::optional<Result> content = contentParser(state);
    // 9.
    if (!content) {
      break;
    }
    // Merge the result from content with the total results
    std::move(content->root.begin(), content->root.end(), std::inserter(result.root, result.root.end()));
    std::move(result.blocks.begin(), result.blocks.end(), std::inserter(content->blocks, content->blocks.begin()));
    result.blocks = content->blocks;
    result.root.push_back(vspaceParser(state));
    // 10.
  }
  // 11.
  if (!match("{% endfor %}", state)) {
    compileError("Missing {% endfor %}", state);
    return std::nullopt;
  }
  Token forEnd(TokenType::forEnd);
  result.root.push_back(std::move(forEnd));
  return result;
}

/*
 * <if> ::= "{% if ~("%}")* " %}" <vspace> (<content><vspace>)* ["{% else %}<vspace>(<content> <vspace>)*] "{% endif %}"
 * 1. Fail if we do not start with "{% if "
 * 2. Goto 4. if the next 3 characters are " %}"
 * 3. Fail if the next 2 characters are "%}"
 * 4. Goto 2.
 * 5. Fail if we do not continue with " %} "
 * 6. Parse vspace
 * 7. Goto 9. if we can not parse content followed by a vspace
 * 8. Goto 7.
 * 9. Goto 10. if we continue with "{% else %}" else goto 13.
 * 10. Parse vspace
 * 11. Goto 13. if we can not parse content followed by a vspace
 * 12. Goto 11.
 * 13. Fail if we do not end with "{% endif %}"
 */
std::optional<Result> Parser::ifParser(State& state) {
  state.push();
  // 1.
  if (!match("{% if ", state)) {
    state.pop();
    return std::nullopt;
  }
  state.drop();
  Result result;
  Token ifBegin(TokenType::ifBegin);
  for (;;) {
    std::optional<char> c1 = state.peek(0);
    std::optional<char> c2 = state.peek(1);
    std::optional<char> c3 = state.peek(2);
    if (!c1 || !c2 || !c3) {
      break;
    }
    // 2.
    if (*c1 == ' ' && *c2 == '%' && *c3 == '}') {
      break;
    }
    // 3.
    if (*c1 == '%' && *c2 == '}') {
      break;
    }
    ifBegin.data[0].push_back(*c1);
    state.read();
    // 4.
  }
  // 5.
  if (!match(" %}", state)) {
    compileError("Missing %}", state);
    return std::nullopt;
  }
  result.root.push_back(std::move(ifBegin));
  // 6.
  vspaceParser(state);
  for (;;) {
    std::optional<Result> content = contentParser(state);
    // 7.
    if (!content) {
      break;
    }
    // Merge the result from content with the total results
    std::move(content->root.begin(), content->root.end(), std::inserter(result.root, result.root.end()));
    std::move(result.blocks.begin(), result.blocks.end(), std::inserter(content->blocks, content->blocks.begin()));
    result.blocks = content->blocks;
    result.root.push_back(vspaceParser(state));
    // 8.
  }
  state.push();
  // 9.
  if (match("{% else %}", state)) {
    Token elseBegin(TokenType::elseBegin);
    result.root.push_back(std::move(elseBegin));
    // 10.
    vspaceParser(state);
    for (;;) {
      std::optional<Result> content = contentParser(state);
      // 11.
      if (!content) {
        break;
      }
      // Merge the result from content with the total results
      std::move(content->root.begin(), content->root.end(), std::inserter(result.root, result.root.end()));
      std::move(result.blocks.begin(), result.blocks.end(), std::inserter(content->blocks, content->blocks.begin()));
      result.blocks = content->blocks;
      result.root.push_back(vspaceParser(state));
      // 12.
    }
    state.drop();
  } else {
    state.pop();
  }
  // 13.
  if (!match("{% endif %}", state)) {
    compileError("Missing {% endif %}", state);
    return std::nullopt;
  }
  Token ifEnd(TokenType::ifEnd);
  result.root.push_back(std::move(ifEnd));
  return result;
}

/*
 * <let> ::= "{% let " <name> " = " ~("%}")* " %}"
 * 1. Fail if we do not start with "{% for "
 * 2. Parse name successfully
 * 3. Fail if we do not continue with " in "
 * 4. Goto 6. if the next 3 characters are " %}"
 * 5. Fail if the next 2 characters are "%}"
 * 6. Goto 4.
 * 7. Fail if we do not continue with " %} "
 */
std::optional<Result> Parser::letParser(State& state) {
  state.push();
  // 1.
  if (!match("{% let ", state)) {
    state.pop();
    return std::nullopt;
  }
  state.drop();
  // 2.
  Result result;
  Token let(TokenType::let);
  std::optional<std::string> item = nameParser(state);
  if (!item) {
    compileError("Could not parse variable name", state);
    return std::nullopt;
  }
  let.data[0] = std::move(*item);
  // 3.
  if (!match(" = ", state)) {
    compileError("Missing =", state);
    return std::nullopt;
  }
  for (;;) {
    std::optional<char> c1 = state.peek(0);
    std::optional<char> c2 = state.peek(1);
    std::optional<char> c3 = state.peek(2);
    if (!c1 || !c2 || !c3) {
      break;
    }
    // 4.
    if (*c1 == ' ' && *c2 == '%' && *c3 == '}') {
      break;
    }
    // 5.
    if (*c1 == '%' && *c2 == '}') {
      break;
    }
    let.data[1].push_back(*c1);
    state.read();
    // 6.
  }
  // 7.
  if (!match(" %}", state)) {
    compileError("Missing %}", state);
    return std::nullopt;
  }
  result.root.push_back(std::move(let));
  return result;
}

/*
 * <content> ::= <text> | <block> | <for> | <if> | <include> | <expression>
 * 1. If we can parse a text block then finish
 * 2. If we can parse a block block then finish
 * 3. If we can parse a for block then finish
 * 4. If we can parse an if block then finish
 * 5. If we can parse an let block then finish
 * 6. If we can parse an include block then finish
 * 7. If we can parse an expression block then finish
 * 8. Fail
 */
std::optional<Result> Parser::contentParser(State& state) {
  state.push();
  // 1.
  std::optional<Result> text = textParser(state);
  if (text) {
    state.drop();
    return text;
  }
  state.pop();
  state.push();
  // 2.
  std::optional<Result> block = blockParser(state);
  if (block) {
    state.drop();
    return block;
  }
  state.pop();
  state.push();
  // 3.
  std::optional<Result> forResult = forParser(state);
  if (forResult) {
    state.drop();
    return forResult;
  }
  state.pop();
  state.push();
  // 4.
  std::optional<Result> ifResult = ifParser(state);
  if (ifResult) {
    state.drop();
    return ifResult;
  }
  state.pop();
  state.push();
  // 5.
  std::optional<Result> letResult = letParser(state);
  if (letResult) {
    state.drop();
    return letResult;
  }
  state.pop();
  state.push();
  // 6.
  std::optional<Result> include = includeParser(state);
  if (include) {
    state.drop();
    return include;
  }
  state.pop();
  state.push();
  // 7.
  std::optional<Result> expression = expressionParser(state);
  if (expression) {
    state.drop();
    return expression;
  }
  state.pop();
  // 8.
  return std::nullopt;
}

/*
 * <text> ::= (~(("{% ") | ("{{ ") | "\n"))+
 * 1. Stop if the next character is "\n"
 * 2. Stop if the next characters are "{%" or "{{"
 * 3. Goto 1.
 */
std::optional<Result> Parser::textParser(State& state) {
  Result result;
  Token text(TokenType::text);
  std::optional<char> c = state.peek();
  // 1.
  for (; c && *c != '\n'; c = state.peek()) {
    // 2.
    if (*c == '{') {
      std::optional<char> next = state.peek(1);
      if (next && (*next == '%' || *next == '{')) {
        break;
      }
    }
    text.data[0].push_back(*c);
    state.read();
    // 3.
  }
  if (text.data[0].empty()) {
    return std::nullopt;
  }
  result.root.push_back(std::move(text));
  return result;
}

/*
 * <include> ::= "{% include " <filename> " %}"
 * 1. Fail if we do not start with "{% include "
 * 2. Parse filename sucessfully
 * 3. Fail if we do not end with " %}"
 */
std::optional<Result> Parser::includeParser(State& state) {
  state.push();
  // 1.
  if (!match("{% include ", state)) {
    state.pop();
    return std::nullopt;
  }
  state.drop();
  // 2.
  std::optional<std::string> filename = filenameParser(state);
  if (!filename) {
    compileError("Could not parse filename", state);
    return std::nullopt;
  }
  // Parse the include
  std::optional<Result> result = templateParser(*filename);
  if (!result) {
    compileError("Could not parse include", state);
    return std::nullopt;
  }
  // 3.
  if (!match(" %}", state)) {
    compileError("Missing %}", state);
    return std::nullopt;
  }
  return result;
}

/*
 * <expression> ::= "{{ " ~("}}")* " }}"
 * 1. Fail if we do not start with "{{ "
 * 2. Goto 5. if the next 3 characters are " }}"
 * 3. Fail if the next 2 characters are "}}"
 * 4. Goto 2.
 * 5. Fail if we do not end with " }}"
 */
std::optional<Result> Parser::expressionParser(State& state) {
  state.push();
  // 1.
  if (!match("{{ ", state)) {
    state.pop();
    return std::nullopt;
  }
  state.drop();
  Result result;
  Token expression(TokenType::expression);
  for (;;) {
    std::optional<char> c1 = state.peek(0);
    std::optional<char> c2 = state.peek(1);
    std::optional<char> c3 = state.peek(2);
    if (!c1 || !c2 || !c3) {
      break;
    }
    // 2.
    if ((*c1 == ' ' || *c1 == '}') && *c2 == '}' && *c3 == '}') {
      break;
    }
    // 3.
    if (*c1 == '}' && *c2 == '}') {
      break;
    }
    expression.data[0].push_back(*c1);
    state.read();
    // 4.
  }
  // 5.
  if (!match(" }}", state)) {
    compileError("Missing }}", state);
    return std::nullopt;
  }
  result.root.push_back(std::move(expression));
  return result;
}

/*
 * <filename> ::= (~" ")+
 * 1. Fail if we start with a space
 * 2. Stop if the next character is a step
 * 3. Goto 2.
 */
std::optional<std::string> Parser::filenameParser(State& state) {
  std::string name;
  std::optional<char> c = state.peek();
  // 1.
  if (!c || *c == ' ') {
    return std::nullopt;
  }
  // 2.
  for (; c && *c != ' '; c = state.peek()) {
    name.push_back(*c);
    state.read();
    // 3.
  }
  return name;
}

/*
 * <name> ::= <letter> (<letter> | <number>)*
 * 1. Fail if we do not start with a letter
 * 2. Stop if the next character is not a letter or number
 * 3. Goto 2.
 * Since letterParser takes a const we do not have to push / pop state.
 */
std::optional<std::string> Parser::nameParser(State& state) {
  std::string name;
  std::optional<char> c = letterParser(state);
  // 1.
  if (!c) {
    return std::nullopt;
  }
  // 2.
  while (c) {
    name.push_back(*c);
    state.read();
    c = letterParser(state);
    if (!c) {
      c = numberParser(state);
    }
    // 3.
  }
  return name;
}

/*
 * <letter> ::= <upper> | <lower>
 * Since upperParser takes a const we do not have to push / pop state.
 */
std::optional<char> Parser::letterParser(const State& state) {
  std::optional<char> upper = upperParser(state);
  return upper ? upper : lowerParser(state);
}

/*
 * <upper> ::= A-Z
 */
std::optional<char> Parser::upperParser(const State& state) {
  std::optional<char> c = state.peek();
  if (c && (*c < 'A' || *c > 'Z')) {
    return std::nullopt;
  }
  return c;
}

/*
 * <lower> ::= a-z
 */
std::optional<char> Parser::lowerParser(const State& state) {
  std::optional<char> c = state.peek();
  if (c && (*c < 'a' || *c > 'z')) {
    return std::nullopt;
  }
  return c;
}

/*
 * <number> ::= 0-9
 */
std::optional<char> Parser::numberParser(const State& state) {
  std::optional<char> c = state.peek();
  if (c && (*c < '0' || *c > '9')) {
    return std::nullopt;
  }
  return c;
}

/*
 * <vspace> ::= ("\n" " "*)*
 * 1. Stop if we do not start with "\n"
 * 2. Read spaces
 * 3. Goto 1.
 */
Token Parser::vspaceParser(State& state) {
  Token vspace(TokenType::vspace);
  // 1.
  for (std::optional<char> c = state.peek(); c && *c == '\n'; c = state.peek()) {
    vspace.data[0].push_back(*c);
    state.read();
    // 2.
    for (c = state.peek(); c && *c == ' '; c = state.peek()) {
      vspace.data[0].push_back(*c);
      state.read();
    }
    // 3.
  }
  return vspace;
}

bool Parser::match(const std::string_view needle, State& state) {
  for (const char c1 : needle) {
    std::optional<char> c2 = state.read();
    if (!c2 || *c2 != c1) {
      return false;
    }
  }
  return true;
}

}  // namespace templating
