#ifndef TEMPLATING_PARSER_H
#define TEMPLATING_PARSER_H

#include <optional>
#include <string>

#include "../util/filesystem.h"
#include "result.h"
#include "state.h"

namespace templating {

/**
 * @brief Implements a parser for the following language
 *
 * @verbatim
 * <template> ::= <include> <vspace> <blocks> | (<content> <vspace>)*
 * <blocks> ::= (<block> <vspace>)*
 * <block> ::= "{% block " <name> " %}" <vspace> (<content> <vspace>)* "{% endblock %}"
 * <for> ::= "{% for " <name> " in " ~("%}")* " %}" <vspace> (<content> <vspace>)* "{% endfor %}"
 * <if> ::= "{% if ~("%}")* " %}" <vspace> (<content><vspace>)* ["{% else %}<vspace>(<content> <vspace>)*] "{% endif %}"
 * <let> ::= "{% let " <name> " = " ~("%}")* " %}"
 * <content> ::= <text> | <block> | <for> | <if> | <let> | <include> | <expression>
 * <text> ::= (~(("{%") | ("{{") | "\n"))+
 * <include> ::= "{% include " <filename> " %}"
 * <expression> ::= "{{ " ~("}}")* " }}"
 * <filename> ::= (~" ")+
 * <name> ::= <letter> (<letter> | <number>)*
 * <letter> ::= <upper> | <lower>
 * <upper> ::= A-Z
 * <lower> ::= a-z
 * <number> ::= 0-9
 * <vspace> ::= ("\n" " "*)*
 * @endverbatim
 */
class Parser {
 private:
  util::Filesystem& filesystem;

 public:
  Parser(util::Filesystem& filesystem);
  /**
   * @brief Parses a template
   *
   * The current directory should be changed to the template folder root.
   * The variable filename should refer to a file that can be found in the
   * current directory.
   *
   * @param filename The file to parse
   * @return Parsing result
   */
  std::optional<Result> templateParser(const std::string& filename);
  /**
   * @brief Parses a series of block blocks
   *
   * @param filesystem The filesystem templates are read from
   * @param state The current parser state
   * @return A list of parsed blocks
   */
  Blocks blocksParser(State& state);
  /**
   * @brief Parses a block block
   *
   * @param filesystem The filesystem templates are read from
   * @param state The current parser state
   * @return Parsing result for the block block
   */
  std::optional<Result> blockParser(State& state);
  /**
   * @brief Parses a for block
   *
   * @param state The current parser state
   * @return Parsing result for the for block
   */
  std::optional<Result> forParser(State& state);
  /**
   * @brief Parses an if block
   *
   * @param state The current parser state
   * @return Parsing result for the if block
   */
  std::optional<Result> ifParser(State& state);
  static std::optional<Result> letParser(State& state);
  /**
   * @brief Parses content
   *
   * @param filesystem The filesystem templates are read from
   * @param state The current parser state
   * @return Parsing result of content
   */
  std::optional<Result> contentParser(State& state);
  /**
   * @brief Parses text
   *
   * @param state The current parser state
   * @return Token containing parsed text
   */
  static std::optional<Result> textParser(State& state);
  /**
   * @brief Parses an include block
   *
   * @param filesystem The filesystem templates are read from
   * @param state The current parser state
   * @return Parsing result for included template
   */
  std::optional<Result> includeParser(State& state);
  /**
   * @brief Parses an expression block
   *
   * @param state The current parser state
   * @return Token containing parsed expression
   */
  static std::optional<Result> expressionParser(State& state);
  /**
   * @brief Parses a filename
   *
   * @param state The current parser state
   * @return The parsed filename
   */
  static std::optional<std::string> filenameParser(State& state);
  /**
   * @brief Parses a name
   *
   * @param state The current parser state
   * @return The parsed name
   */
  static std::optional<std::string> nameParser(State& state);
  /**
   * @brief Parses a letter
   *
   * @param state The current parser state
   * @return The parsed letter
   */
  static std::optional<char> letterParser(const State& state);
  /**
   * @brief Parses an uppercase letter
   *
   * @param state The current parser state
   * @return The parsed uppercase letter
   */
  static std::optional<char> upperParser(const State& state);
  /**
   * @brief Parses a lowercase letter
   *
   * @param state The current parser state
   * @return The parsed lowercase letter
   */
  static std::optional<char> lowerParser(const State& state);
  /**
   * @brief Parses a number
   *
   * @param state The current parser state
   * @return The parsed number
   */
  static std::optional<char> numberParser(const State& state);
  /**
   * @brief Parses verticle space
   *
   * @param state The current parser state
   * @return Token containing verticle space
   */
  static Token vspaceParser(State& state);
  /**
   * @brief Checks if what comes next in state matches needle
   *
   * @param needle The string to check for
   * @param state The current parser state
   * @retval true On match
   * @retval false On no match
   */
  static bool match(std::string_view needle, State& state);
};

}  // namespace templating

#endif
