#include "parser.h"

#include <gtest/gtest.h>

#include "../util/fake_filesystem.h"

TEST(templatingParser, blockParser) {
  util::FakeFilesystem fakeFilesystem;
  templating::Parser parser(fakeFilesystem);
  {
    templating::State state("", "");
    std::optional<templating::Result> block = parser.blockParser(state);
    ASSERT_FALSE(block.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<templating::Result> block = parser.blockParser(state);
    ASSERT_FALSE(block.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% block ");
    std::optional<templating::Result> block = parser.blockParser(state);
    ASSERT_FALSE(block.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 10);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% block a");
    std::optional<templating::Result> block = parser.blockParser(state);
    ASSERT_FALSE(block.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 11);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% block a %}");
    std::optional<templating::Result> block = parser.blockParser(state);
    ASSERT_FALSE(block.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 14);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% block a %}b");
    std::optional<templating::Result> block = parser.blockParser(state);
    ASSERT_FALSE(block.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 15);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% block a %}b{% endblock %}");
    std::optional<templating::Result> block = parser.blockParser(state);
    ASSERT_TRUE(block.has_value());
    ASSERT_EQ(block->root.size(), 1);
    EXPECT_EQ(block->root.front().type, templating::TokenType::block);
    EXPECT_EQ(block->root.front().data[0], "a");
    auto it = block->blocks.find("a");
    ASSERT_NE(it, block->blocks.cend());
    EXPECT_EQ(it->second.size(), 1);
    EXPECT_EQ(it->second.front().type, templating::TokenType::text);
    EXPECT_EQ(it->second.front().data[0], "b");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 29);
    std::cerr.rdbuf(oldBuffer);
  }
}

TEST(templatingParser, forParser) {
  util::FakeFilesystem fakeFilesystem;
  templating::Parser parser(fakeFilesystem);
  {
    templating::State state("", "");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_FALSE(forStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_FALSE(forStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% for ");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_FALSE(forStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 8);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% for x");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_FALSE(forStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 9);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% for x in ");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_FALSE(forStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 13);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% for x in a");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_FALSE(forStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 14);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% for x in a %}");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_FALSE(forStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 17);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% for x in a %}b");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_FALSE(forStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 18);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% for x in a %}b{% endfor %}");
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_TRUE(forStatement.has_value());
    ASSERT_EQ(forStatement->root.size(), 4);
    EXPECT_EQ(forStatement->root.at(0).type, templating::TokenType::forBegin);
    EXPECT_EQ(forStatement->root.at(0).data[0], "x");
    EXPECT_EQ(forStatement->root.at(0).data[1], "a");
    EXPECT_EQ(forStatement->root.at(1).type, templating::TokenType::text);
    EXPECT_EQ(forStatement->root.at(1).data[0], "b");
    EXPECT_EQ(forStatement->root.at(2).type, templating::TokenType::vspace);
    EXPECT_EQ(forStatement->root.at(2).data[0], "");
    EXPECT_EQ(forStatement->root.at(3).type, templating::TokenType::forEnd);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 30);
    std::cerr.rdbuf(oldBuffer);
  }
  for (int i = 0; i < 10; i++) {
    std::string value(i, 'a');
    std::string content = absl::StrCat("{% for x in ", value, " %}{% endfor %}");
    templating::State state("", content);
    std::optional<templating::Result> forStatement = parser.forParser(state);
    ASSERT_TRUE(forStatement.has_value());
    ASSERT_EQ(forStatement->root.size(), 2);
    EXPECT_EQ(forStatement->root.at(0).type, templating::TokenType::forBegin);
    EXPECT_EQ(forStatement->root.at(0).data[0], "x");
    EXPECT_EQ(forStatement->root.at(0).data[1], value);
    EXPECT_EQ(forStatement->root.at(1).type, templating::TokenType::forEnd);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 28 + i);
  }
}

TEST(templatingParser, ifParser) {
  util::FakeFilesystem fakeFilesystem;
  templating::Parser parser(fakeFilesystem);
  {
    templating::State state("", "");
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_FALSE(ifStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_FALSE(ifStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% if ");
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_FALSE(ifStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 7);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% if a");
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_FALSE(ifStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 8);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% if a %}");
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_FALSE(ifStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 11);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% if a %}b");
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_FALSE(ifStatement.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 12);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% if a %}b{% endif %}");
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_TRUE(ifStatement.has_value());
    ASSERT_EQ(ifStatement->root.size(), 4);
    EXPECT_EQ(ifStatement->root.at(0).type, templating::TokenType::ifBegin);
    EXPECT_EQ(ifStatement->root.at(0).data[0], "a");
    EXPECT_EQ(ifStatement->root.at(1).type, templating::TokenType::text);
    EXPECT_EQ(ifStatement->root.at(1).data[0], "b");
    EXPECT_EQ(ifStatement->root.at(2).type, templating::TokenType::vspace);
    EXPECT_EQ(ifStatement->root.at(2).data[0], "");
    EXPECT_EQ(ifStatement->root.at(3).type, templating::TokenType::ifEnd);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 23);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% if a %}b{% else %}c{% endif %}");
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_TRUE(ifStatement.has_value());
    ASSERT_EQ(ifStatement->root.size(), 7);
    EXPECT_EQ(ifStatement->root.at(0).type, templating::TokenType::ifBegin);
    EXPECT_EQ(ifStatement->root.at(0).data[0], "a");
    EXPECT_EQ(ifStatement->root.at(1).type, templating::TokenType::text);
    EXPECT_EQ(ifStatement->root.at(1).data[0], "b");
    EXPECT_EQ(ifStatement->root.at(2).type, templating::TokenType::vspace);
    EXPECT_EQ(ifStatement->root.at(2).data[0], "");
    EXPECT_EQ(ifStatement->root.at(3).type, templating::TokenType::elseBegin);
    EXPECT_EQ(ifStatement->root.at(4).type, templating::TokenType::text);
    EXPECT_EQ(ifStatement->root.at(4).data[0], "c");
    EXPECT_EQ(ifStatement->root.at(5).type, templating::TokenType::vspace);
    EXPECT_EQ(ifStatement->root.at(5).data[0], "");
    EXPECT_EQ(ifStatement->root.at(6).type, templating::TokenType::ifEnd);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 34);
    std::cerr.rdbuf(oldBuffer);
  }
  for (int i = 0; i < 10; i++) {
    std::string value(i, 'a');
    std::string content = absl::StrCat("{% if ", value, " %}{% endif %}");
    templating::State state("", content);
    std::optional<templating::Result> ifStatement = parser.ifParser(state);
    ASSERT_TRUE(ifStatement.has_value());
    ASSERT_EQ(ifStatement->root.size(), 2);
    EXPECT_EQ(ifStatement->root.at(0).type, templating::TokenType::ifBegin);
    EXPECT_EQ(ifStatement->root.at(0).data[0], value);
    EXPECT_EQ(ifStatement->root.at(1).type, templating::TokenType::ifEnd);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 21 + i);
  }
}

TEST(templatingParser, letParser) {
  {
    templating::State state("", "");
    std::optional<templating::Result> let = templating::Parser::letParser(state);
    ASSERT_FALSE(let.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<templating::Result> let = templating::Parser::letParser(state);
    ASSERT_FALSE(let.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% let ");
    std::optional<templating::Result> let = templating::Parser::letParser(state);
    ASSERT_FALSE(let.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 8);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% let x");
    std::optional<templating::Result> let = templating::Parser::letParser(state);
    ASSERT_FALSE(let.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 9);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% let x = ");
    std::optional<templating::Result> let = templating::Parser::letParser(state);
    ASSERT_FALSE(let.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 12);
    std::cerr.rdbuf(oldBuffer);
  }
  for (int i = 0; i < 10; i++) {
    std::string value(i, 'a');
    std::string content = absl::StrCat("{% let x = ", value, " %}");
    templating::State state("", content);
    std::optional<templating::Result> let = templating::Parser::letParser(state);
    ASSERT_TRUE(let.has_value());
    ASSERT_EQ(let->root.size(), 1);
    EXPECT_EQ(let->root.front().type, templating::TokenType::let);
    EXPECT_EQ(let->root.front().data[0], "x");
    EXPECT_EQ(let->root.front().data[1], value);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 15 + i);
  }
}

TEST(templatingParser, textParser) {
  {
    templating::State state("", "");
    std::optional<templating::Result> text = templating::Parser::textParser(state);
    ASSERT_FALSE(text.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "{% ");
    std::optional<templating::Result> text = templating::Parser::textParser(state);
    ASSERT_FALSE(text.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "{{ ");
    std::optional<templating::Result> text = templating::Parser::textParser(state);
    ASSERT_FALSE(text.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "\n");
    std::optional<templating::Result> text = templating::Parser::textParser(state);
    ASSERT_FALSE(text.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "  {% ");
    std::optional<templating::Result> text = templating::Parser::textParser(state);
    ASSERT_TRUE(text.has_value());
    ASSERT_EQ(text->root.size(), 1);
    EXPECT_EQ(text->root.front().type, templating::TokenType::text);
    EXPECT_EQ(text->root.front().data[0], "  ");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 3);
  }
  {
    templating::State state("", "  {{ ");
    std::optional<templating::Result> text = templating::Parser::textParser(state);
    ASSERT_TRUE(text.has_value());
    ASSERT_EQ(text->root.size(), 1);
    EXPECT_EQ(text->root.front().type, templating::TokenType::text);
    EXPECT_EQ(text->root.front().data[0], "  ");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 3);
  }
  {
    templating::State state("", "  \n");
    std::optional<templating::Result> text = templating::Parser::textParser(state);
    ASSERT_TRUE(text.has_value());
    ASSERT_EQ(text->root.size(), 1);
    EXPECT_EQ(text->root.front().type, templating::TokenType::text);
    EXPECT_EQ(text->root.front().data[0], "  ");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 3);
  }
}

TEST(templatingParser, contentParser) {
  util::FakeFilesystem fakeFilesystem;
  templating::Parser parser(fakeFilesystem);
  {
    templating::State state("", "");
    std::optional<templating::Result> content = parser.contentParser(state);
    ASSERT_FALSE(content.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<templating::Result> content = parser.contentParser(state);
    ASSERT_TRUE(content.has_value());
    ASSERT_EQ(content->root.size(), 1);
    EXPECT_EQ(content->root.front().type, templating::TokenType::text);
    EXPECT_EQ(content->root.front().data[0], " ");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 2);
  }
  {
    templating::State state("", "{% block a %}{% endblock %}");
    std::optional<templating::Result> content = parser.contentParser(state);
    ASSERT_TRUE(content.has_value());
    ASSERT_EQ(content->root.size(), 1);
    EXPECT_EQ(content->root.front().type, templating::TokenType::block);
    EXPECT_EQ(content->root.front().data[0], "a");
    auto it = content->blocks.find("a");
    ASSERT_NE(it, content->blocks.cend());
    ASSERT_TRUE(it->second.empty());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 28);
  }
  {
    templating::State state("", "{% for a in b %}{% endfor %}");
    std::optional<templating::Result> content = parser.contentParser(state);
    ASSERT_TRUE(content.has_value());
    ASSERT_EQ(content->root.size(), 2);
    EXPECT_EQ(content->root.at(0).type, templating::TokenType::forBegin);
    EXPECT_EQ(content->root.at(0).data[0], "a");
    EXPECT_EQ(content->root.at(0).data[1], "b");
    EXPECT_EQ(content->root.at(1).type, templating::TokenType::forEnd);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 29);
  }
  {
    templating::State state("", "{% if a %}{% endif %}");
    std::optional<templating::Result> content = parser.contentParser(state);
    ASSERT_TRUE(content.has_value());
    ASSERT_EQ(content->root.size(), 2);
    EXPECT_EQ(content->root.at(0).type, templating::TokenType::ifBegin);
    EXPECT_EQ(content->root.at(0).data[0], "a");
    EXPECT_EQ(content->root.at(1).type, templating::TokenType::ifEnd);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 22);
  }
  {
    fakeFilesystem.writeFile("a", "a");
    templating::State state("", "{% include a %}");
    std::optional<templating::Result> content = parser.contentParser(state);
    ASSERT_TRUE(content.has_value());
    ASSERT_EQ(content->root.size(), 2);
    EXPECT_EQ(content->root.at(0).type, templating::TokenType::text);
    EXPECT_EQ(content->root.at(0).data[0], "a");
    EXPECT_EQ(content->root.at(1).type, templating::TokenType::vspace);
    EXPECT_EQ(content->root.at(1).data[0], "");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 16);
    fakeFilesystem.reset();
  }
  {
    templating::State state("", "{{ a }}");
    std::optional<templating::Result> content = parser.contentParser(state);
    ASSERT_TRUE(content.has_value());
    ASSERT_EQ(content->root.size(), 1);
    EXPECT_EQ(content->root.front().type, templating::TokenType::expression);
    EXPECT_EQ(content->root.front().data[0], "a");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 8);
  }
}

TEST(templatingParser, includeParser) {
  util::FakeFilesystem fakeFilesystem;
  templating::Parser parser(fakeFilesystem);
  {
    templating::State state("", "");
    std::optional<templating::Result> include = parser.includeParser(state);
    ASSERT_FALSE(include.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<templating::Result> include = parser.includeParser(state);
    ASSERT_FALSE(include.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% include ");
    std::optional<templating::Result> include = parser.includeParser(state);
    ASSERT_FALSE(include.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 12);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% include a");
    std::optional<templating::Result> include = parser.includeParser(state);
    ASSERT_FALSE(include.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 13);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{% include a %}");
    std::optional<templating::Result> include = parser.includeParser(state);
    ASSERT_FALSE(include.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 13);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    fakeFilesystem.writeFile("a", "a");
    templating::State state("", "{% include a %}");
    std::optional<templating::Result> include = parser.includeParser(state);
    ASSERT_TRUE(include.has_value());
    ASSERT_EQ(include->root.size(), 2);
    EXPECT_EQ(include->root.at(0).type, templating::TokenType::text);
    EXPECT_EQ(include->root.at(0).data[0], "a");
    EXPECT_EQ(include->root.at(1).type, templating::TokenType::vspace);
    EXPECT_EQ(include->root.at(1).data[0], "");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 16);
    fakeFilesystem.reset();
  }
}

TEST(templatingParser, expressionParser) {
  {
    templating::State state("", "");
    std::optional<templating::Result> expression = templating::Parser::expressionParser(state);
    ASSERT_FALSE(expression.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<templating::Result> expression = templating::Parser::expressionParser(state);
    ASSERT_FALSE(expression.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{{ ");
    std::optional<templating::Result> expression = templating::Parser::expressionParser(state);
    ASSERT_FALSE(expression.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 4);
    std::cerr.rdbuf(oldBuffer);
  }
  {
    std::stringstream buffer;
    std::streambuf *oldBuffer = std::cerr.rdbuf(buffer.rdbuf());
    templating::State state("", "{{  ");
    std::optional<templating::Result> expression = templating::Parser::expressionParser(state);
    ASSERT_FALSE(expression.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 5);
    std::cerr.rdbuf(oldBuffer);
  }
  for (int i = 0; i < 10; i++) {
    std::string exp(i, 'a');
    std::string content = absl::StrCat("{{ ", exp, " }}");
    templating::State state("", content);
    std::optional<templating::Result> expression = templating::Parser::expressionParser(state);
    ASSERT_TRUE(expression.has_value());
    ASSERT_EQ(expression->root.size(), 1);
    EXPECT_EQ(expression->root.front().type, templating::TokenType::expression);
    EXPECT_EQ(expression->root.front().data[0], exp);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 7 + i);
  }
}

TEST(templatingParser, filenameParser) {
  {
    templating::State state("", "");
    std::optional<std::string> filename = templating::Parser::filenameParser(state);
    ASSERT_FALSE(filename.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<std::string> filename = templating::Parser::filenameParser(state);
    ASSERT_FALSE(filename.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "a");
    std::optional<std::string> filename = templating::Parser::filenameParser(state);
    ASSERT_TRUE(filename.has_value());
    EXPECT_EQ(*filename, "a");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 2);
  }
  {
    templating::State state("", "a ");
    std::optional<std::string> filename = templating::Parser::filenameParser(state);
    ASSERT_TRUE(filename.has_value());
    EXPECT_EQ(*filename, "a");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 2);
  }
  {
    templating::State state("", "aa ");
    std::optional<std::string> filename = templating::Parser::filenameParser(state);
    ASSERT_TRUE(filename.has_value());
    EXPECT_EQ(*filename, "aa");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 3);
  }
}

TEST(templatingParser, nameParser) {
  {
    templating::State state("", "");
    std::optional<std::string> name = templating::Parser::nameParser(state);
    ASSERT_FALSE(name.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<std::string> name = templating::Parser::nameParser(state);
    ASSERT_FALSE(name.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "1");
    std::optional<std::string> name = templating::Parser::nameParser(state);
    ASSERT_FALSE(name.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "a");
    std::optional<std::string> name = templating::Parser::nameParser(state);
    ASSERT_TRUE(name.has_value());
    EXPECT_EQ(*name, "a");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 2);
  }
  {
    templating::State state("", "a ");
    std::optional<std::string> name = templating::Parser::nameParser(state);
    ASSERT_TRUE(name.has_value());
    EXPECT_EQ(*name, "a");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 2);
  }
  {
    templating::State state("", "a1");
    std::optional<std::string> name = templating::Parser::nameParser(state);
    ASSERT_TRUE(name.has_value());
    EXPECT_EQ(*name, "a1");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 3);
  }
  {
    templating::State state("", "a1 ");
    std::optional<std::string> name = templating::Parser::nameParser(state);
    ASSERT_TRUE(name.has_value());
    EXPECT_EQ(*name, "a1");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 3);
  }
}

TEST(templatingParser, letterParser) {
  {
    templating::State state("", "");
    std::optional<char> number = templating::Parser::letterParser(state);
    ASSERT_FALSE(number.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<char> number = templating::Parser::letterParser(state);
    ASSERT_FALSE(number.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 26; i++) {
    std::string content(1, 'A' + i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::letterParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, 'A' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 26; i++) {
    std::string content(1, 'A' + i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::letterParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, 'A' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 26; i++) {
    std::string content(1, 'a' + i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::letterParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, 'a' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 26; i++) {
    std::string content(1, 'a' + i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::letterParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, 'a' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
}

TEST(templatingParser, upperParser) {
  {
    templating::State state("", "");
    std::optional<char> number = templating::Parser::upperParser(state);
    ASSERT_FALSE(number.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<char> number = templating::Parser::upperParser(state);
    ASSERT_FALSE(number.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 26; i++) {
    std::string content(1, 'A' + i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::upperParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, 'A' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 26; i++) {
    std::string content(1, 'A' + i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::upperParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, 'A' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
}

TEST(templatingParser, lowerParser) {
  {
    templating::State state("", "");
    std::optional<char> number = templating::Parser::lowerParser(state);
    ASSERT_FALSE(number.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<char> number = templating::Parser::lowerParser(state);
    ASSERT_FALSE(number.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 26; i++) {
    std::string content(1, 'a' + i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::lowerParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, 'a' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 26; i++) {
    std::string content(1, 'a' + i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::lowerParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, 'a' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
}

TEST(templatingParser, numberParser) {
  {
    templating::State state("", "");
    std::optional<char> number = templating::Parser::numberParser(state);
    ASSERT_FALSE(number.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    std::optional<char> number = templating::Parser::numberParser(state);
    ASSERT_FALSE(number.has_value());
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 10; i++) {
    std::string content = absl::StrCat(i);
    templating::State state("", content);
    std::optional<char> number = templating::Parser::numberParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, '0' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  for (int i = 0; i < 10; i++) {
    std::string content = absl::StrCat(i, " ");
    templating::State state("", content);
    std::optional<char> number = templating::Parser::numberParser(state);
    ASSERT_TRUE(number.has_value());
    EXPECT_EQ(*number, '0' + i);
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
}

TEST(templatingParser, vspaceParser) {
  {
    templating::State state("", "");
    templating::Token token = templating::Parser::vspaceParser(state);
    EXPECT_EQ(token.type, templating::TokenType::vspace);
    EXPECT_EQ(token.data[0], "");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", " ");
    templating::Token token = templating::Parser::vspaceParser(state);
    EXPECT_EQ(token.type, templating::TokenType::vspace);
    EXPECT_EQ(token.data[0], "");
    EXPECT_EQ(state.line(), 1);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "\na");
    templating::Token token = templating::Parser::vspaceParser(state);
    EXPECT_EQ(token.type, templating::TokenType::vspace);
    EXPECT_EQ(token.data[0], "\n");
    EXPECT_EQ(state.line(), 2);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "\n\na");
    templating::Token token = templating::Parser::vspaceParser(state);
    EXPECT_EQ(token.type, templating::TokenType::vspace);
    EXPECT_EQ(token.data[0], "\n\n");
    EXPECT_EQ(state.line(), 3);
    EXPECT_EQ(state.column(), 1);
  }
  {
    templating::State state("", "\n    \n    a");
    templating::Token token = templating::Parser::vspaceParser(state);
    EXPECT_EQ(token.type, templating::TokenType::vspace);
    EXPECT_EQ(token.data[0], "\n    \n    ");
    EXPECT_EQ(state.line(), 3);
    EXPECT_EQ(state.column(), 5);
  }
}

TEST(templatingParser, match) {
  {
    templating::State state("", "needle");
    EXPECT_TRUE(templating::Parser::match("needle", state));
  }
  {
    templating::State state("", "needle ");
    EXPECT_TRUE(templating::Parser::match("needle", state));
  }
  {
    templating::State state("", " needle");
    EXPECT_FALSE(templating::Parser::match("needle", state));
  }
  {
    templating::State state("", " ");
    EXPECT_FALSE(templating::Parser::match("needle", state));
  }
  {
    templating::State state("", "");
    EXPECT_FALSE(templating::Parser::match("needle", state));
  }
  {
    templating::State state("", " ");
    EXPECT_TRUE(templating::Parser::match("", state));
  }
  {
    templating::State state("", "");
    EXPECT_TRUE(templating::Parser::match("", state));
  }
}