#ifndef TEMPLATING_STATE_H
#define TEMPLATING_STATE_H

#include <optional>
#include <string>
#include <vector>

namespace templating {

/**
 * @brief Locaton in a file
 */
struct Location {
  /// The number of characters into the file
  size_t cursor;
  /// The line number that the cursor is on
  size_t line;
  /// The column number that the cursor is on
  size_t column;
};

/**
 * @brief Holds the state needed for parsing
 */
class State {
 private:
  /**
   * @brief The file that is being parsed
   */
  const std::string_view filename;
  /**
   * @brief The contents of the file being parsed
   */
  const std::string_view contents;
  /**
   * @brief The current location in the file being parsed
   */
  Location location;
  /**
   * @brief An undo history for locations being parsed
   *
   * This variable is used as a stack.
   */
  std::vector<Location> history;

 public:
  /**
   * @brief Creates a new state for parsing
   *
   * Parsing starts at the beginning of the file.
   * The filename that is passed in is used purely for display.
   * @param filename The filename of the file being parsed
   * @param contents The contents of the file being parsed
   */
  State(const std::string_view filename, const std::string_view contents);
  /**
   * @brief Peeks forward n characters
   *
   * @param n The number of characters to skip over
   * @retval std::nullopt If the end of the file was reached
   * @return The requested character
   */
  std::optional<char> peek(const size_t n = 0) const;
  /**
   * @brief Reads a character
   *
   * @retval std::nullopt If the end of the file was reached
   * @return The requested character
   */
  std::optional<char> read();
  /**
   * @brief Checks if the end of the file has been reached
   *
   * @return If the end of the file has been reached
   */
  bool end() const;
  /**
   * @brief Pushes the current location on the history stack
   */
  void push();
  /**
   * @brief Pops a location off the top of the stack and uses it
   *
   * This function invokes undefined behavior if the stack is empty.
   */
  void pop();
  /**
   * @brief Pops a location off the top of the stack
   *
   * This function should be used if the history pushed onto the stack
   * is no longer needed.
   */
  void drop();
  /**
   * @brief The filename of the file being parsed
   * @return The filename of the file being parsed
   */
  std::string_view file() const;
  /**
   * @brief The current line being parsed
   * @return The current line being parsed
   */
  size_t line() const;
  /**
   * @brief The current column being parsed
   * @return The current column being parsed
   */
  size_t column() const;
};

}  // namespace templating

#endif
