#include "util.h"

#include <absl/strings/str_cat.h>

namespace util {

std::string escapeLiteral(const std::string_view raw, const size_t indent) {
  std::string result;
  for (char c : raw) {
    switch (c) {
      case '\a':
        absl::StrAppend(&result, "\\a");
        break;
      case '\b':
        absl::StrAppend(&result, "\\b");
        break;
      case '\t':
        absl::StrAppend(&result, "\\t");
        break;
      case '\n':
        absl::StrAppend(&result, "\\n\"\n", std::string(indent, ' '), "\"");
        break;
      case '\v':
        absl::StrAppend(&result, "\\v");
        break;
      case '\f':
        absl::StrAppend(&result, "\\f");
        break;
      case '\r':
        absl::StrAppend(&result, "\\r");
        break;
      case '\x1B':
        absl::StrAppend(&result, "\\x1B");
        break;
      case '"':
        absl::StrAppend(&result, "\\\"");
        break;
      case '\\':
        absl::StrAppend(&result, "\\\\");
        break;
      default:
        result += c;
        break;
    }
  }
  return result;
}

std::string_view limit(const std::string_view string, const size_t length) {
  if (string.length() <= length) {
    return string;
  }
  return string.substr(0, length);
}

std::string stripHTML(const std::string_view string) {
  if (string.find('<') == std::string::npos) {
    return std::string(string);
  }
  std::string stripped;
  stripped.reserve(string.length());
  int depth = 0;
  for (char c : string) {
    if (c == '<') {
      depth++;
    } else if (depth == 0) {
      stripped += c;
    } else if (c == '>') {
      depth--;
    }
  }
  return stripped;
}

bool writeBinaryFile(const std::string& path, const std::span<uint8_t> buffer) {
  FILE* file = fopen(path.c_str(), "wb");
  if (!file) {
    return false;
  }
  size_t written = fwrite(buffer.data(), buffer.size(), 1, file);
  fclose(file);
  return written == 1;
}

std::optional<std::string> detectExtension(const std::string_view filename, const std::string_view fileContents) {
  if (fileContents.empty()) {
    return std::nullopt;
  }
  if (fileContents.starts_with("\x89PNG\x0D\x0A\x1A\x0A")) {
    return "png";
  }
  if (fileContents.starts_with("\xFF\xD8\xFF")) {
    return "jpg";
  }
  if (fileContents.starts_with("GIF87a") || fileContents.starts_with("GIF89a")) {
    return "gif";
  }
  if (fileContents.starts_with("\x1A\x45\xDF\xA3")) {
    return "webm";
  }
  if (fileContents.length() >= 4 &&
      (fileContents.substr(4).starts_with("ftypmp42") || fileContents.substr(4).starts_with("ftypisom"))) {
    return "mp4";
  }
  if (fileContents.starts_with("\xFF\xFB") || fileContents.starts_with("ID3")) {
    return "mp3";
  }
  if (fileContents.starts_with("OggS")) {
    return "ogg";
  }
  if (fileContents.starts_with("fLaC")) {
    return "flac";
  }
  if (fileContents.starts_with("%PDF")) {
    return "pdf";
  }
  if (fileContents.starts_with("PK\x03\x04") ||
      (fileContents.length() >= 30 && fileContents.substr(30).starts_with("mimetypeapplication/epub+zip"))) {
    return "epub";
  }
  if (filename.ends_with(".txt")) {
    return "txt";
  }
  if (fileContents.starts_with("RIFF") && fileContents.length() >= 8 && fileContents.substr(8).starts_with("WEBP")) {
    return "webp";
  }
  if (filename.ends_with(".svg") && (fileContents.find("DOCTYPE svg") != std::string_view::npos ||
                                     fileContents.find("<svg") != std::string_view::npos)) {
    return "svg";
  }
  return std::nullopt;
}

FileClass fileClass(const std::string_view extension) {
  if (extension == "png") {
    return FileClass::image;
  } else if (extension == "jpg") {
    return FileClass::image;
  } else if (extension == "gif") {
    return FileClass::image;
  } else if (extension == "webm") {
    return FileClass::video;
  } else if (extension == "mp4") {
    return FileClass::video;
  } else if (extension == "mp3") {
    return FileClass::audio;
  } else if (extension == "ogg") {
    return FileClass::audio;
  } else if (extension == "flac") {
    return FileClass::audio;
  } else if (extension == "pdf") {
    return FileClass::pdf;
  } else if (extension == "epub") {
    return FileClass::document;
  } else if (extension == "svg") {
    return FileClass::image;
  } else if (extension == "txt") {
    return FileClass::text;
  } else if (extension == "webp") {
    return FileClass::image;
  }
  return FileClass::unknown;
}

std::string_view extension(const std::string_view filename) {
  std::string_view name(filename);
  size_t dot = filename.rfind('.');
  if (dot != std::string_view::npos) {
    name.remove_prefix(dot + 1);
  }
  return name;
}

}  // namespace util
