#ifndef UTIL_NANOCHAN_H
#define UTIL_NANOCHAN_H

#include <optional>
#include <span>
#include <string>
#include <unordered_map>

namespace util {

enum class FileClass {
  audio,
  document,
  image,
  pdf,
  text,
  unknown,
  video,
};

/**
 * @brief Escapes a string to be included in a C++ string literal
 *
 * @warning Only control codes, double quotes, and backslashes are escaped
 * @param raw The string to escape
 * @param indent The amount of spaces to use for indentation
 * @return The escaped string
 */
std::string escapeLiteral(const std::string_view raw, const size_t indent);
/**
 * @brief Limits the length of a string
 *
 * @param string The string to limit
 * @param length The maximum length of the string
 * @return The string
 */
std::string_view limit(const std::string_view string, const size_t length);
/**
 * @brief Strips HTML tags
 *
 * A tag is opened with a < and close with a >.
 * @warning Angle brackets must only be used for tags and not in attributes
 * @param string The string to strip
 * @return The stripped string
 */
std::string stripHTML(const std::string_view string);
bool writeBinaryFile(const std::string& path, const std::span<uint8_t> buffer);
std::optional<std::string> detectExtension(  // TODO: String view?
    const std::string_view filename, const std::string_view fileContents);
FileClass fileClass(const std::string_view extension);
std::string_view extension(const std::string_view filename);
template <typename K, typename V, typename... Args>
V at_or(const std::unordered_map<K, V>& map, const K& key, Args&&... args) {
  const auto pair = map.find(key);
  if (pair == map.end()) {
    return V(std::forward<Args>(args)...);
  }
  return pair->second;
}

}  // namespace util

#endif
